  /*-----------------------------------------------------------*
   |                         FINDER.c                          |
   |                                                           |
   | This is the top file for FINDER, except for Fmain.c where |
   | the function main() lives in solitary splendour. When     |
   | FINDER is used as an external utility by other programs,  |
   | this file will be linked in but Fmain.c will not.         |
   |                                                           |
   | This is the main file where FINDER's external variables   |
   | are officially declared.                                  |
   *-----------------------------------------------------------*/

#define MAINFILE
#define IN_SCOTT

#include <sys/time.h>

#include "FINDER.h"
#ifdef SCOTT
#include "called_by_finder.h" // protos shared by SCOTT and FINDER
#endif

// search satistics
static boolean optimal[MAXISET_ARRAY];        // true iff the model is optimal
static boolean old_optimal[MAXISET_ARRAY];    // true iff the model was optimal
static int clauses_modelled[MAXISET_ARRAY];   // clauses true in each model
static int searches[MAXISET_ARRAY];           // attempts at a new model
static int failed_searches[MAXISET_ARRAY];    // failed try for new model
static int successful_searches[MAXISET_ARRAY];// successful try for new model
static int tests[MAXISET_ARRAY];              // clauses tested in each theory

// variables that manage the approximation to a decison procedure
static struct itimerval time_buffer;
static int min_time_limit;                   // lower bound on time_limit 
static int max_time_limit;                   // upper bound on time_limit 

static int time_limit;                       // global search time limit
static int time_limit_resets[MAXISET_ARRAY]; // number of resets
static int max_time;                         // max time in a successful search
static int max_time_in[MAXISET_ARRAY];       // ... in a particular theory
static int max_backtracks_in[MAXISET_ARRAY]; // likewise max backtracks


static int time_outs;                        // jobs killed by alarm clock
static int time_outs_in_test;                // jobs killed by alarm clock in test()
static float tolerance;                      // time_limit = max_time * tolerance
static float min_prob_extn;                  // limit on development of cset

  /*-----------------------------------------------------------*
   | The last parameter gives us a reasonable threshold for    |
   | deciding consistency based on the current most difficult, |
   | yet successful, search.                                   |
   *-----------------------------------------------------------*/

void inc_time_limit() 
  /*---------------------------------------------------------*
   | Called by cover_new_clauses().                          |
   |                                                         |
   | Increases the limit on the search time by the tolerance |
   | factor.                                                 |
   *---------------------------------------------------------*/
{
  if (time_limit*tolerance < max_time_limit){
    int last_time_limit;

    last_time_limit = time_limit;
    time_limit *= tolerance;
    
    message(2,"");
    message(2,"----> Resetting search time limit to %d = %d*%.1f", 
	    time_limit, last_time_limit, tolerance);
    message(2,"");
  }
  else if (time_limit< max_time_limit) {
    time_limit = max_time_limit;
    message(2,"");
    message(2,"----> Resetting search time limit to the maximum %d", 
	    time_limit);
    message(2,"");
  }

} /*--inc_time_limit()--*/


void reset_optimal()
  /*-----------------------------------------------------------*
   | Called by check_failed_clauses() and execute_strategy().  |
   |                                                           |
   | Resets statistics governing whether a model is considered |
   | optimal.                                                  |
   *-----------------------------------------------------------*/
{
  int theory;

  for (theory = 1; theory < MAXISET_ARRAY; theory++){
    if (searches[theory] > 10 ) searches[theory] -= 10;
    else searches[theory] = 0;
    optimal[theory] = false;
  }

} /*--reset_optimal()--*/


void inc_cardinality()
  /*-------------------------------------------------------------*
   | Called by cover_new_clauses(). Increase the cardinality of  |
   | each sort iff we have not reached the limit for that sort.  |
   *-------------------------------------------------------------*/
{
  static int calls = 0;
  int i;

  if (calls > 4) return;

  calls++;
  message(2,"");
  message(3,"----> Increasing minimum cardinality of all sorts");
  for (i = 0; sort[i].sname; i++ ) {
    if (sort[i].sizmin < sort[i].sizmax) {
      message(3,"<---- Values for %s > %d", sort[i].sname, sort[i].sizmin);
      sort[i].sizmin++;
    }
  }
  message(2,"");

} /*--inc_cardinality()--*/


void init_theory(int theory)
  /*---------------------------------------------------------*
   | Called by init_finder(), set_up_last_model(), and SCOTT |
   | fcn discard().  Initialises statistics for the theory   |
   | index.                                                  |
   *---------------------------------------------------------*/
{
  size[theory] = 
    tests[theory] = 
    updates[theory] = 
    searches[theory] = 
    failed_searches[theory] = 
    successful_searches[theory] = 
    clauses_modelled[theory] = 
    max_backtracks_in[theory] = 
    max_time_in[theory] = 
    time_limit_resets[theory] = 0;  
  
  optimal[theory] = false;

} /*--init_theory()--*/


boolean init_finder(char filename[])
  /*------------------------------------------------------------*
   | Called by init_semantics() to initialises theory space     |
   | data.                                                      |
   |                                                            |
   | Reads the input file and searches for a model of the input |
   | theory.  This tests the soundness of the input file.  If a |
   | model is found, this is set as the base model for all      |
   | theories                                                   |
   |                                                            |
   | Returns true iff a model of the input theory is found.     |
   *------------------------------------------------------------*/
{
 int i; 
 FILE *initial_finder_file;

  /* global inits not handeled by initialize() */
  piping = false;
  pipe_open = false;
  negative = false;
  bug_out = false;
  time_outs = time_outs_in_test = 0;
  
  initialize();

  if (!declare_otter_symbols()) return false;

  if (filename[0] != '\0') {

    record_event("read_finder_file\n");

    /*---read input file (may conatin a base theory)--*/
    printf("\n------------> reading finder input file.\n");  
    if ( !(initial_finder_file = fopen(filename,"r")) )
      inerror("Unable to open file",filename);
    read_job(initial_finder_file);
    fclose(initial_finder_file);
  }
  
  if (over_argmax) return false;  // functors too big to model

  tr_parameters.reps[NULL_BACKTRACKS].i_report = 0;
  tr_parameters.reps[SURJ_BACKTRACKS].i_report = 0;
  tr_parameters.reps[SOLUTIONS].i_report = 0;
  tr_parameters.reps[BAD_TESTS].i_report = 0;
  tr_parameters.vlength = 0;
  tr_parameters.done = 0; // Prime trans_ref 
  single_model = true;    // Generate at most one model for each theory
  
  // Run the timer a hundred times so it behaves itself.
  for (i=0; i<100; i++) {
    start_alarm_clock(); 
    stop_alarm_clock();
  }
  
  max_time = 0;
  setting.maxtime = time_limit;

  current_theory = 0;
  go(0);

  if (updates[0] > 0) return true;
  else return false;

} /*--init_finder()--*/



void set_up_last_model()
  /*-----------------------------------------------------------*
   | Called by model().  Sets up the last model of the current |
   | theory for testing.                                       |
   *-----------------------------------------------------------*/
{
  int i;
  functor f;  
  clause cl; /*loop indices*/
  char cl_str[CLAUSE_STRING_ARRAY];
  
  if (size[current_theory] == 0){ // Its a new theory!!

    /*----------------------------------------------------------*
     | We are starting up a new a new theory so intialise its   |
     | model to that of the base theory (supplied by the finder |
     | input file). Note: only the very first call to model()   |
     | will set up the model of the base theory.                |
     *----------------------------------------------------------*/

    last_vlength[current_theory] = last_vlength[0];
    for (i = 0; i < last_vlength[0]; i++)
      last_model[current_theory][i] = last_model[0][i];
    for ( i = 0; sort[i].sname; i++ ) {
      sort[i].siz = last_sizes[current_theory][i] = last_sizes[0][i];
      sort[i].sizmask = last_sizmasks[current_theory][i] = last_sizmasks[0][i];
    }
    for (f = FN; f; f = f->nextfunct) {
      f->fcc = f->old_fcc[current_theory] = f->old_fcc[0];
      set_fats(f);
    }

    // reset TH to point at the front of theory (this is real important)
    TH[current_theory] = theory[current_theory];
    TH[current_theory]->nextclause = TH[current_theory]->lastclause = 0;

    init_theory(current_theory);

    // if there's a base theory copy it to the new theory
    if (size[0] > 0){
      for (cl = TH[0]; cl; cl = cl->nextclause) {
	cl_str[0] ='\0';    // clear the line
	clause_to_string(cl,cl_str);
	add_clause(cl_str); // read clause as tho' input
	if (size[current_theory] < CL_MAX-1) add_to_theory();
      }      
    }

  }
  else {
    for ( i = 0; sort[i].sname; i++ ) {
      sort[i].siz = last_sizes[current_theory][i];
      sort[i].sizmask = last_sizmasks[current_theory][i];
    }
    for ( f = FN; f; f = f->nextfunct ) {
      f->fcc = f->old_fcc[current_theory];
      set_fats(f);
    }
  }

  set_table(FN);

  setting.max_backtracks = 0; // no backtrack limit
  setting.maxtime = time_limit;

  tr_parameters.reps[NULL_BACKTRACKS].i_report = 0;
  tr_parameters.reps[SURJ_BACKTRACKS].i_report = 0;
  tr_parameters.reps[SOLUTIONS].i_report = 0;
  tr_parameters.reps[BAD_TESTS].i_report = 0;
  tr_parameters.vlength = last_vlength[current_theory];
  tr_parameters.done = 0;

} /*--set_up_last_model()--*/


int get_last_backtracks()
  /*----------------------------------------------------------*
   | Called by model(). Returns the number of backtracks that |
   | the most recent search employed.                         |
   *----------------------------------------------------------*/
{
  return  (tr_parameters.reps[NULL_BACKTRACKS].i_report +
	   tr_parameters.reps[SURJ_BACKTRACKS].i_report +
	   tr_parameters.reps[SOLUTIONS].i_report +
	   tr_parameters.reps[BAD_TESTS].i_report);

} /*--get_last_backtracks()--*/


void short_timesup()
  /*--------------------------------------------------------*
   | Called by model().  What to do if an alarm signal gets |
   | caught.                                                |
   *--------------------------------------------------------*/
{
  job_done( &tr_parameters, "time limit" );
  bug_out = true;
  time_outs++;
} /*--short_timesup()--*/


void start_alarm_clock()
    /*----------------------------------------------------------*
     | Called by model().  Sets a temporary limit on CPU time.  |
     *----------------------------------------------------------*/
{
  bug_out = false;    // process flag for alarm call

  // when we reach the limit turn off the timer
  time_buffer.it_interval.tv_sec = 0;
  time_buffer.it_interval.tv_usec = 0;
  
  // set the time limit
  time_buffer.it_value.tv_sec = setting.maxtime/1000;
  time_buffer.it_value.tv_usec = (setting.maxtime % 1000) *1000;
  
  signal( SIGVTALRM, short_timesup );
  if (setitimer(ITIMER_VIRTUAL, &(time_buffer), 0) == -1) {
    printf("\nERROR: could not set alarm clock.\n");
    exit(1);
  }
} /*--start_alarm_clock()--*/

int stop_alarm_clock()
    /*-------------------------------------------------------*
     | Called by model().  Unsets the temporary limit on CPU |
     | time.                                                 |
     *-------------------------------------------------------*/
{
  int time_left;

  bug_out = false;    // process flag for alarm call

  // turn off alarm clocks
  if (getitimer(ITIMER_VIRTUAL, &(time_buffer)) == -1) {
    printf("\nERROR: could not get alarm clock value.\n");
    exit(1);
  } 
  
  time_left = time_buffer.it_value.tv_sec*1000 + 
    time_buffer.it_value.tv_usec/(float)1000;
  
  // unset the timer
  time_buffer.it_value.tv_sec = 0;
  time_buffer.it_value.tv_usec = 0;
  if (setitimer(ITIMER_VIRTUAL, &(time_buffer), 0) == -1) {
    printf("\nERROR: could not unset alarm clock.\n");
    exit(1);
  } 

  // return timeleft in milliseconds
  return time_left;
  
} /*--stop_alarm_clock()--*/

boolean model(char clause[], int in_theory, 
	      boolean test_only, 
	      boolean add_if_true)
  /*------------------------------------------------------------*
   | Called by SCOTT fcn's: model_clause() test_clause(),       |
   | false_in_model().                                          |
   |                                                            |
   | Tries to determine whether the clause is consistent with   |
   | the current theory.  First, test it in the current model.  |
   | If satisfied the given clause is added to the theory.  If  |
   | not satisfied, and if the model has not been marked as     |
   | optimal, search for a new model of the current theory plus |
   | the given clause.  If a new model is found, then that      |
   | becomes the current model and the given clause is added to |
   | the theory.                                                |
   |                                                            |
   | Returns true iff the given clause is modelled by the       |
   | (possibly new) current model.                              |
   |                                                            |
   | Note that the timer is reset on every call. This allows a  |
   | time limit to be set for each individual search.           |
   *------------------------------------------------------------*/
{
  boolean satisfied = false;

  current_theory = in_theory;  // set global variable 

  // test in previous model first
  set_up_last_model();  
  add_clause(clause);
  set_matrix(TH[current_theory]);
  set_matrix(&latest_clause);

  tests[current_theory]++;
  if (test(last_model[current_theory], 0)) {
    satisfied = true;
  }

  // search for a new model
  if (size[current_theory] < CL_MAX &&
      !satisfied && 
      !test_only && 
      !already_optimal()){

    int last_num_updates, last_backtracks, last_time = 0;

    last_num_updates = updates[current_theory];
    searches[current_theory]++;

    bug_out = false;    // process flag for alarm call
    start_alarm_clock();
    go(0);

    last_time = setting.maxtime - stop_alarm_clock();
    last_backtracks = get_last_backtracks();      

    if (updates[current_theory] > last_num_updates) {
      satisfied = true;
      successful_searches[current_theory]++;

      // record new max backtracks needed
      if (max_backtracks_in[current_theory] < last_backtracks)
	max_backtracks_in[current_theory] = last_backtracks;
      
      // record new max time needed
      if (last_time > min_time_limit && last_time < max_time_limit) {
	
	if (max_time_in[current_theory] < last_time)
	  max_time_in[current_theory] = last_time;
	if (max_time < last_time) max_time = last_time;	

	if (last_time*tolerance > time_limit) {
	  if (max_time_limit >= last_time*tolerance) {
	    time_limit = last_time * tolerance;
	    message(2,"Resetting search time limit to %d = %d*%.1f",
		    time_limit, last_time, tolerance);
	  }
	  else {
	    time_limit = max_time_limit;
	    message(2,"Resetting search time limit to the maximum %d", 
		    time_limit);
	  }
	  time_limit_resets[current_theory]++;
	}
      }
      
    }
    else failed_searches[current_theory]++;
  }  
  
  if (satisfied && add_if_true) {
    clauses_modelled[current_theory]++;
    if (size[current_theory] < CL_MAX-1 
	&& !optimal[current_theory]) {
      message(4, "Adding following clause to theory %d:\n%s\n", current_theory, clause);
      add_to_theory(); 
    }
    if (size[current_theory] >= CL_MAX-1) {
      abort_scott("CL_MAX exceeded.  You will need to increase this value\n"
		  "(defined in finder_dir/defs.h) and recompile SCOTT.");
    }
  }

  return satisfied;
  
} /*--model()--*/


boolean already_optimal()
  /*-----------------------------------------------------------*
   | Called by model().                                        |
   |                                                           |
   | The idea is that as the theory becomes more constrained,  |
   | it becomes more diffciult to show that an extension is    |
   | consistent.  Here we try to model the probability that we |
   | have reached the limit of small finite models.            |
   *-----------------------------------------------------------*/
{
  
  if (optimal[current_theory]) return true;

  if (searches[current_theory] < 10 ||
      !successful_searches[current_theory]) return false;

  if ((successful_searches[current_theory]/
       (float)(searches[current_theory])) < min_prob_extn){ 
    optimal[current_theory] = true;
    return true;
  }
  
  return false;

} /*--already_optimal()--*/

void go(int x)
  /*-----------------------------------------------------------*
   | Called by go(), find_first_model(), and model().          |
   |                                                           |
   | Main work loop of FINDER.  Calls transref to search every |
   | search space, defined by setting the cardinalities of the |
   | sorts, within the ranges stipulated.                      |
   *-----------------------------------------------------------*/
{
  if (bug_out) return;

  if ( negative && !clauses_fail ) return;
  if ( !sort[x].sname ) {
    if (set_up_cc()) {
      firstmodel = true; // perhaps need to change this?
      transref(&tr_parameters);
    }
    if (bug_out) return;
  }
  else if ( sort[x].piped && !tr_parameters.done ) go(x+1);
  else for ( sort[x].siz = sort[x].sizmin;
	     sort[x].siz <= sort[x].sizmax && !tr_parameters.done;
	     sort[x].siz++ ) {
    go(x+1);
  }
} /*--go()--*/


void output_the_model(int new_theory)
  /*-----------------------------------------------------------*
   | Called by check_for_model().  Prints the current model of |
   | the given theory.                                         |
   *-----------------------------------------------------------*/
{
  current_theory = new_theory;
  print_the_model();

} /*--exit_with_model()--*/

boolean equality_is_present()
  /*----------------------------------------------------*
   | Called by check_for_model().  Returns true iff the |
   | function symbol "equal" is defined.                |
   *----------------------------------------------------*/
{
  int i;
  static boolean already_checked = false;
  static boolean answer = false;

  if (!already_checked) { // don't check twice
    for (i = 0; i < FN_MAX && fn[i].fname; i++) {
      if (strcmp(fn[i].fname,"equal") == 0) {
	return true;
      }
    }
  }
  
  return answer;
  
} /*--equality_is_present()--*/


void print_theory(int i)
  /*------------------------------------------------------------*
   | Debugging tool. Prints all clauses in the current theory.  |
   *------------------------------------------------------------*/
{  
  clause cl;
  char cl_str[CLAUSE_STRING_ARRAY];

  printf("\n%%Printing theory %d",i);
  for (cl = TH[i]; cl; cl = cl->nextclause){
    cl_str[0] ='\0';    // clear the line
    clause_to_string(cl,cl_str);
    printf("\n%s",cl_str);
  }
  printf("\n");

} /*--print_theory()--*/


boolean check_theory(int a_theory)
  /*--------------------------------------------------------*
   | Debugging tool. Test each theory clause in its current |
   | model.                                                 |
   *--------------------------------------------------------*/
{  
  clause cl;
  char cl_str[CLAUSE_STRING_ARRAY];
  int old_current_theory;

  old_current_theory = current_theory;
  current_theory = a_theory;  // set global variable 

  //  printf("\n%%Checking theory %d",a_theory);
  for (cl = TH[a_theory]; cl; cl = cl->nextclause){
    cl_str[0] ='\0';    // clear the line
    clause_to_string(cl,cl_str);

    set_up_last_model();  
    add_clause(cl_str);
    set_matrix(TH[current_theory]);
    set_matrix(&latest_clause);

    tests[current_theory]++;
    if (!test(last_model[current_theory], 0)) {
      printf("\nFollowing clause fails in the current model of "
	     "theory %d:\n%s\n\n", a_theory, cl_str);
      current_theory = old_current_theory;
      return false;
    }
  }
  current_theory = old_current_theory;
  return true;

} /*--print_theory()--*/


void clause_to_string(clause cl, char *str)
  /*------------------------------------------------------------*
   | Called by set_up_last_mode().  Converts a FINDER clause to |
   | its output string - a kludge to copy a cl (The existing    |
   | clause is reparsed and put into a different theory).       |
   *------------------------------------------------------------*/
{
  wff_stack wf;
  
  for ( wf = cl->prem; wf; wf = wf->nxt ) {
    if ( wf != cl->prem ) strcat(str,", ");
    wff_to_str(wf->w,str);
  }
  //  if ( !cl->prem ) strcat(str,"$true"); // supperfluous
  strcat(str,"  ->  ");
  for ( wf = cl->conc; wf; wf = wf->nxt ) {
    if ( wf != cl->conc ) strcat(str,", ");
    wff_to_str(wf->w,str);
    }
  //  if ( !cl->conc ) strcat(str,"$false"); // supperfluous
  strcat(str,".\0"); // I think I'm becomming paranoid...

} /*--clause_to_string()--*/


void wff_to_str(wff w, char *str)
  /*-----------------------------------------------------*
   | Called by clause_to_string().  Recursively copies a |
   | formula in Polish to a string.                      |
   *-----------------------------------------------------*/
{
    int i;

    if ( !w ) return;
    strcat(str,w->sym->fname);
    if ( w->sym->adicity ) {
	strcat(str,"(");
	for (i=0; i<w->sym->adicity; i++) {
	    if (i) strcat(str,",");
	    wff_to_str(w->sb[i],str);
	}
	strcat(str,")");
    }
} /*--wff_to_str()--*/


/*==========================================================\
| The following fcns encapsulate data referenced outside of |
| this file.                                                |
\==========================================================*/

void set_min_prob_extn(float prob){min_prob_extn = prob;}
void set_min_time_limit(int limit){min_time_limit = limit;}
void set_max_time_limit(int limit){max_time_limit = limit;}
void set_tolerance(float limit){tolerance = limit;}
void set_time_limit(int limit){time_limit = limit;}

boolean get_old_optimal(int theory)
{
  return old_optimal[theory];
}
int get_clauses_modelled(int theory)
{
  return clauses_modelled[theory];
}
int get_failed_searches(int theory){
  return failed_searches[theory];
}
int get_last_default_size(int theory){
  return last_sizes[theory][2];
}
int get_max_backtracks_in(int theory){
  return max_backtracks_in[theory];
}

int get_time_limit(){
  return time_limit;
}
int get_max_time_in(int theory){
  return max_time_in[theory];
}
int get_searches(int theory) {
  return searches[theory];
}
int get_size(int theory){
  return size[theory];
}
int get_successful_searches(int theory){
  return successful_searches[theory];
}
int get_tests(int theory){
  return tests[theory];
}
int get_time_outs(){
  return time_outs;
}
int get_time_outs_in_test(){
  return time_outs_in_test;
}
int get_updates(int theory){
  return updates[theory];
}
int get_time_limit_resets(int theory){
  return time_limit_resets[theory];
}

/*========================== END =============================*/
