/*------------------------------------------------------------------------*/
/* Main routines for obtaining finder input */
/*------------------------------------------------------------------------*/

#include "FINDER.h"

static char job_line[JL_MAX]; /* Line of input text		   */
static int  current_line = 0; /* Current line number		   */
static int  input_point =0;   /*Current input point on th job_line */
static FILE *source;	      /* Current input stream              */


/*------------------------------------------------------------------------*/
char next_char()

     /*Called by get_a_char() and get_word().
      *Returns the next character from the job_line[] (or cmd_line[]),
      *Skipping over comments and if job_line[] is exhausted, reads
      *the next input line.

      *The input line may be a job_line (if input is coming from stdin) or 
      *a cmd_line (if input is given using the -c command line option).*/

{
  if ( cmd_line_input ) return command_line[input_point++];
  
  while (job_line[input_point] == '%' ||  /* skip comments    */
	 job_line[input_point] == '\0') { /* skip empty lines */
    
    if (source &&                       /* if reading from a file */
	!fgets(job_line,JL_MAX,source)) /* refill the job_line */
      return '\0';                 /* patch for end_of_file */
    input_point = 0;               /* restart at the begining of a line */
    current_line++;                /* have moved to the next line */
#ifdef DEBUG
    fputs(job_line,stderr);
#endif
  }	
  
  return job_line[input_point++]; /* next charater on the job_line */

} /* next_char() */



/*------------------------------------------------------------------------*/
char get_a_char()

     /*Called by get_word().
      *Returns next meaningful character from the job_line*/
     
{
  unsigned char c;
  
  while ( isspace(c = next_char()) ) ;
  return c;

} /* get_a_char() */



/*------------------------------------------------------------------------*/
void get_word( unsigned char word[] )
     /*Called by get_key_word().
      *most fcn's in this file and by fcn's in parser.c;
      
      *Gets next meaningful word from the job_line*/
{
  int i;
  
  word[0] = get_a_char();
  
  if ( strchr("=><+,:.%)(}{",word[0]) )  /* check for leading punctuation */
    word[1] = '\0';  

  else {  /* get following charaters */
    
    for (i=1; !(isspace(word[i] = next_char())); i++ ){ /* scan word  */
      
      if (strchr("=<+,:.%)(}{",word[i])){ /* simple word terminator */
	input_point--;                    /* unread the terminator  */
	break;
      }

      if (word[i] == '>' && word[i-1] == '-' ){

	if (i > 1){                  /* "->" terminates preceding word */
	  input_point -= 2;          /* unread the terminator          */
	  word[i-1] = '\0';          /* rubout the termintor           */
	  break;
	}
	else {                  /* "->" is the word to be returned ...  */
	  i++;                  /* so don't overwrite the the delimiter */
	  break;
	}
      }
    } /* end for */

    word[i] = '\0'; /*rubout trailing space or delimiter */

  } /* end else */

} /* get_word() */



/*------------------------------------------------------------------------*/
statement_type get_key_word()

     /*Called by read_job().
      *Gets next key word from the job_line and determines its type.*/
     
{
  unsigned char wd[20];
  
  get_word(wd);
  if ( !strcmp(wd,".") || !strcmp(wd,"end") ) return nul;
  if ( !strcmp(wd,"sort") ) return srt;
  if ( !strcmp(wd,"function") ) return fcn;
  if ( !strcmp(wd,"constant") ) return fcn0;
  if ( !strcmp(wd,"clause") ) return cls;
  if ( !strcmp(wd,"setting") ) return prm;
  if ( !strcmp(wd,"include") ) return inc;
  inwarning
    (wd,"found where key word (sort/function/clause/setting/end) expected");
  return nul;
} /* get_key_word() */



/*------------------------------------------------------------------------*/
void read_job( FILE *infil )
     
     /*Called by FINDER fcn main() and SCOTT fcn model_finder_file() .
      *Loops continuously over input file until get_key_word() returns null.
      *Integrates the object according to its type*/
{
  *job_line = 0;
  input_point = 0;
  source = infil;
  for (;;)
    switch( get_key_word() ) {
    case nul: return;
    case srt: get_item(get_sort,0,0); break;
    case fcn: get_item(get_func,0,0); break;
    case fcn0: get_item(get_const,0,0); break;
    case cls: get_item(get_input_cl,0,0); break;
    case prm: get_item(get_set,0,0); break;
    case inc: get_incl_file(); break;
    }
} /* read_job */



/*------------------------------------------------------------------------*/
void get_incl_file()

     /*Called by
      *Get_incl causes a file to be read as though it were
      *included at the current point. Note that it must end
      *with the word `end'. The file name is enclosed in 
      *(round) parentheses. Since the routine get_a_char() 
      *is used, the name may have white space such as line
      *breaks inserted in it: these will be ignored.
      */
     
     
{
  FILE *oldsource, *infil;
  char include_filename[MAX_FILE_NAME], oldjobline[JL_MAX];
  int oldinputpoint, i;
  
  get_word( include_filename );
  if ( *include_filename != '(' )
    inerror("Included filename must be in parentheses","");
  
  i = 0;
  while ( (include_filename[i] = get_a_char()) != ')' ) i++;
  include_filename[i] = '\0';
  
  oldsource = source;
  strcpy(oldjobline, job_line);
  oldinputpoint = input_point;
  if ( !(infil = fopen(include_filename,"r")) )
    inerror("Cannot open file",include_filename);
  read_job( infil );
  fclose(source);
  source = oldsource;
  strcpy(job_line, oldjobline);
  input_point = oldinputpoint;

} /* get_incl_file() */


/*------------------------------------------------------------------------*/
int get_int()

     /*Called by get_set(), get_item(), sort_spec(), func_spec(), 
      *and got_function().
      *Gets an expected unsigned integer charater from the input line.
      *Exits in error if one is not found.*/

{
  unsigned char s[LEXLENGTH];
  int rc, i;
  
  do get_word(s); while ( *s == ':' );
  rc = 0;
  for ( i = 0; s[i]; i++ ) {
    if ( !isdigit(s[i]) )
      inerror(s,"found where integer expected");
    rc *= 10;
    rc += (s[i] - '0');
  }
  return rc;

} /*get_int()*/

/*------------------------------------------------------------------------*/
float get_float()

     /*Called by get_set().
      *Gets an expected float from the input line.
      *Exits in error if one is not found.*/

{
  unsigned char word[LEXLENGTH];
  int digit;
  float sum, decimal;

  do get_word(word); while ( *word == ':' );
  sum = 0;
  for ( digit = 0; word[digit]; digit++ ) {
    if ( word[digit] == '.') break;
    if ( !isdigit(word[digit]) )
      inerror(word,"found where integer expected");
    sum *= 10;
    sum += (word[digit] - '0');
  }

  do get_word(word); while ( *word == '.' );
  decimal = 1;
  for (digit = 0; word[digit]; digit++ ) {
    if ( !isdigit(word[digit]) )
      inerror(word,"found where integer expected");
    decimal *= 10;
    sum += ((word[digit] - '0') / decimal);
  }

  return sum;

} /*get_float()*/


/*------------------------------------------------------------------------*/
void get_item( void (*item)(), kind k, functor f )

     /*Called by read_job(), sort_spec(), get_func(), get_cons(), 
      *and get_set().

      *Calls the appropriate function (*item)() to get either
      *one item of the given species (e.g. one clause) or a set of them.
      *A set, naturally, is specified by listing its members within a 
      *pair of braces.
      *
      *Some items like clauses have an optional integer parameter given
      *in parentheses after the key word, so get_item must search for 
      *this first.
      *
      *Functions called by get_item have all four of the parameters it 
      *issues, but generally don't use most of them.  Although the above
      *function read_job does not make use of the parameters k and f, 
      *functions (*item)() may do so when they themselves call get_item 
      *for their sub-items.
      *All calls to (*item)() are pased the first word of the item.*/

{
  unsigned char s[LEXLENGTH];
  int arg = 0;

  get_word(s);

  if ( s[0] == '(' ) {  /* initial arg of item is an int in parentheses */
    arg = get_int();
    get_word(s); /*get closing parentheses */
    if ( s[0] != ')' ) inerror("Missing right parenthesis","");
    get_word(s); /*get string following integer argument */
  }

  if ( s[0] == '{' ) {  /* list of items in braces */
    get_word(s);
    while ( s[0] != '}' ) {
      if ( s[0] != '.' ) (*item)(k,f,s,arg);
      get_word(s);  /* get next item */
    }
  }
  else if ( s[0] != '.' ) (*item)(k,f,s,arg);   /* single item */
}



/*------------------------------------------------------------------------*/
void get_sort( kind k, functor f, unsigned char name[], int x )

     /*Called by get_item().
      *Creates a new sort if "name" does'nt already exist.
      *Gets a (possibly) revised specification for that sort.*/
{	
  get_item( sort_spec, newkind(name), 0 );

} /* get_sort() */



/*------------------------------------------------------------------------*/
void sort_spec( kind k, functor f, unsigned char spec[], int x )

     /* Called by get_item().
      *	Revises the specifications (if any) for the kind k.*/

{
  int canon_count;

  if ( !strcmp(spec,"cardinality") || !strcmp(spec,"card") ) {
    if ( k->canonical ) {
      inwarning(k->sname,"an enumerated sort - can't set cardinality");
      return;
    }
    get_word(spec);
    switch (spec[0]) {
    case '=': k->sizmax = k->sizmin = get_int(); return;
    case '<': k->sizmax = get_int()-1;
      if ( k->sizmax < k->sizmin ) k->sizmin = k->sizmax;
      return;
    case '>': k->sizmin = get_int() + 1;
      if ( k->sizmax < k->sizmin ) k->sizmax = k->sizmin;
      return;
    }
    inerror("","Statement of cardinality needs =, < or >");		
  }
  if (!strcmp(spec,"enum") ) {
    de_enumerate( k );
    for ( canon_count = 0; canon_count < SZ; canon_count++ ) {
      get_word(spec);
      if ( spec[0] == ':' ) get_word(spec);
      canon( k, spec );
      get_word(spec);
      if ( spec[0] == '.' ) return;
      if ( spec[0] != ',' )
	inerror(spec,"found where comma expected");
    }
  }
  inerror(spec,"found where key word \"cardinality\" or \"enum\" expected");

} /* sort_spec() */



/*------------------------------------------------------------------------*/
void de_enumerate( kind k )

     /*Called by sort_spec().
      *Removes the existing enumeration (if any) of kind k. */
{
  free_canon( k->canonical );
  k->canonical = 0;

} /* de_enumerate() */


/*------------------------------------------------------------------------*/
void free_canon( name_list lp )

     /*Called by de_emumerate().
      *Recusively frees pointers to the cannical names in the 
      *enumeration list. */
{
  if ( !lp ) return;
  free_canon( lp->nextname );
  free( lp->nextname );
  //  printf("free name %x\n");
  lp->nextname = 0;

} /* free_canon() */



/*------------------------------------------------------------------------*/
void get_func( kind k, functor f, unsigned char name[], int x )
     
     /*Called by get_item(), and get_cons().
      *Scans for the argument and value types of the (new) functor. 
      *If the functor all ready exists then this is redefined.
      *Since newfuntor() is called and sets the print mode to null
      *(which it needs to do because it usually sets up variables, etc) 
      *so this function resets it to the default full.
      *Then calls func_spec() to scan any function specifications.
 
      *Parameters k, f and x are dummies for get_func.  The string s
      *holds the name of the function. */
     
{
  int i;
  int arity=0;
  boolean got_arg_types=false;
  boolean got_val_type=false;
  unsigned char type_str[LEXLENGTH], punctuation[LEXLENGTH];
  unsigned char fcn_types[ARGMAX+1][LEXLENGTH];
  
  for ( i = 0; i <= ARGMAX; i++ ) fcn_types[i][0] = '\0';
  
  get_word(type_str);
  
  if ( type_str[0] == ':' ) get_word(type_str); /* skip ':' */

  /* get the types of the functor's arguments, and then its value type */
  do {
#ifdef SCOTT
    if (arity > ARGMAX){
      inwarning("Adicity too great for function",name);
      over_argmax = true; 
      return;
    }
#else
    if (arity > ARGMAX) inerror("Adicity too great for function",name);
#endif
    strcpy(fcn_types[arity++],type_str); /*record the next type*/
    if (got_arg_types) {
      got_val_type = true;
      arity--; /* the functions values is stored as the last type */
    }
    else {
      get_word(punctuation);

      if (!strchr(".,-",punctuation[0]))
	inerror(type_str,"found where punctuation ( , -> or . ) expected");

      switch(punctuation[0]) {
      case '.':
	arity--;
	got_val_type = true;
	break;
      case ',': 
	if (got_arg_types) inerror(name,"can only have ONE value type");
	get_word(type_str); /* get next argument type */
	break;
      case '-':
	if ( punctuation[1] != '>' )
	  inerror(punctuation,"found where \"->\" expected");
	got_arg_types = true;
	get_word(type_str); /* get value type */
	break;
      }
    }
  }
  while (!got_val_type);

  f = newfunctor( name, Func, arity);
  f->val_type = set_up_type(fcn_types[arity]);
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(fcn_types[i]);
  f->pmode = full;
  get_item( func_spec, 0, f );
  
} /* get_func */



/*------------------------------------------------------------------------*/
void get_const( kind k, functor f, unsigned char name[], int x )
     
     /*Called by get_item().
      *Calls get_func() for 0-adic function symbol. */

{
    unsigned char const_type[LEXLENGTH];

    f = newfunctor( name, Func, 0 );

    do	get_word(const_type);
    while ( const_type[0] == ':' );

    f->val_type = set_up_type(const_type);
    f->pmode = full;
    get_item( func_spec, 0, f );

} /* get_const() */




/*------------------------------------------------------------------------*/
kind set_up_type( unsigned char type[] )

     /*Called by get_func() and get_const().
      *Returns a pointer to the kind coresponding to the given "type".*/
{
  int i;
  
  for ( i = 0; sort[i].sname; i++ ) 
    if ( !strcmp(sort[i].sname,type) ) return( &sort[i] );
  inwarning("Non-existent sort",type);
  return 0;
} /* set_up_type() */




/*------------------------------------------------------------------------*/
void func_spec( kind k, functor f, unsigned char key_word[], int x )

     /* Called by get_func() and get_const().
      *Sets up the various properties of the function that can be 
      *specified in the section following its declaration. 
      *(The paramters x and f are just dummies). */
     
{
  unsigned char specs[LEXLENGTH];
  
  if ( !strcmp(key_word,"total") ) f->total = true;
  else if ( !strcmp(key_word,"partial") ) f->total = false;
  else if ( !strcmp(key_word,"injective") ) f->injective = true;
  else if ( !strcmp(key_word,"left_injective") || 
	   !strcmp(key_word,"column_injective") )
    f->leftin = (f->adicity == 2);
  else if ( !strcmp(key_word,"right_injective") || 
	   !strcmp(key_word,"row_injective") )
    f->rightin = (f->adicity == 2);
  else if ( !strcmp(key_word,"surjective") ) f->surjective = true;
  else if ( !strcmp(key_word,"left_surjective") || 
	   !strcmp(key_word,"column_surjective") )
    f->leftsur = (f->adicity == 2);
  else if ( !strcmp(key_word,"right_surjective") || 
	   !strcmp(key_word,"row_surjective") )
    f->rightsur = (f->adicity == 2);
  else if ( !strcmp(key_word,"bijective") )
    f->surjective = f->injective = true;
  else if ( !strcmp(key_word,"left_bijective") || 
	   !strcmp(key_word,"column_bijective") )
    f->leftsur = f->leftin = (f->adicity == 2);
  else if ( !strcmp(key_word,"right_bijective") || 
	   !strcmp(key_word,"row_bijective") )
    f->rightsur = f->rightin = (f->adicity == 2);
  else if ( !strcmp(key_word,"commutative") ) f->comm = (f->adicity == 2);
  else if ( !strcmp(key_word,"cut") || key_word[0] == '!' ) f->cut_it = true;
  else if ( !strcmp(key_word,"no_priority") ) f->nop = true;
  else if ( !strcmp(key_word,"print") ) {
    get_word(specs);
    if ( specs[0] == ':' ) get_word(specs);
    if ( !strcmp(specs,"none") ) f->pmode = null;
    else if ( !strcmp(specs,"brief") ) f->pmode = brief;
    else if ( !strcmp(specs,"full") ) f->pmode = full;
    else inwarning(specs,"found where print mode expected");
    return;
  }
  else if ( !strcmp(key_word,"scope") ) {
    f->scope = get_int();
    if ( f->scope > 10 ) f->scope = 10;
    return;
  }
  else inwarning(key_word,"found where function specification expected");

} /* func_spec() */


/*------------------------------------------------------------------------*/
boolean add_clause(unsigned char clause_string[])

     /*Called by model() and find_theories().
      *Parse clause_string as tho' it was next on the input line. 
      *Returns TRUE iff clause is non-trivial*/

{
    unsigned char lead_word[LEXLENGTH];

    strcpy(job_line,clause_string);
    input_point = 0;

    get_word(lead_word);
    if (get_cl(lead_word,0))
      return true;
    else
      inwarning("\nPassed clause is trivially satisfiable.","");
    return false;

}/* add_clause() */


void add_declaration(char declaration[])
  /*----------------------------------------------------------------------*\
    Called by SCOTT fcn declare_otter_symbols().
    Parse "declaration" as tho' it was next on the input line. 
  \*----------------------------------------------------------------------*/
{
    strcpy(job_line,declaration);
    input_point = 0;
    source = NULL; // not reading from a file

    switch( get_key_word() ) {
    case nul: return;
    case srt: get_item(get_sort,0,0); break;
    case fcn: get_item(get_func,0,0); break;
    case fcn0: get_item(get_const,0,0); break;
    case cls: get_item(get_input_cl,0,0); break;
    case prm: get_item(get_set,0,0); break;
    case inc: get_incl_file(); break;
    }

}/*--add_declaration()----------------------------------------------------*/


/*------------------------------------------------------------------------*/
void get_input_cl( kind k, functor f, unsigned char first_word[], int priority)

     /*Called by get_item(). 
      *Get the clause off the input line and, provided it's non-trivial,
      *add it to the current theory.*/
     
{
  if ( get_cl(first_word,priority) ) 
    add_to_theory();
  else 
    inwarning("Input clause is trivially satisfiable.","");
}


/*------------------------------------------------------------------------*/
boolean get_cl( unsigned char lead_word[], int priority )

     /*Called by get_input_cl() and add_clause().
      *Parse the next clause on the input line into latest_clause.`
      *Returns TRUE iff the clause is non-trivial.*/ 

{
  wff_stack *ins = 0;
  wff tempwff = 0 ;
  boolean parsed_premises =false;
  latest_clause.priority = priority;
  begin_parse();
  
  do {
    if ( lead_word[0] != '-' ) {
      tempwff = parse( lead_word );
      if (parsed_premises)
	ins = &(latest_clause.conc);
      else
	ins = &(latest_clause.prem);
    }

    if ( !strchr(",-.",lead_word[0]) )
      inerror(lead_word,"found in clause where punctuation expected");

    switch( lead_word[0] ) {
    case '.':	
      ins = &(latest_clause.conc);
      break;
    case '-':	
      if ( lead_word[1] != '>' ) 
	inerror(lead_word,"found where \"->\" expected");
      parsed_premises = true;
    case ',':	
      get_word(lead_word); /*get next predicate name*/
    }

    if ( tempwff ) {
      while ( *ins ) ins = &((*ins)->nxt);
      *ins = (wff_stack) malloc(sizeof(WFF_STACK));
      (*ins)->w = tempwff;
      (*ins)->nxt = 0;
    }
  }
  while ( lead_word[0] != '.' ); /* all clauses terminated with "." */

  typecheck();
 /* if ( trivial_cl() ) return false; */
  end_parse();
  get_weight();

  return true;

} /* get_cl() */



/*------------------------------------------------------------------------*/
boolean isint( unsigned char *s )
     
     /*Called by one line in get_set(). This could be dodgy. */
     
{
  do if ( !isdigit(*s) ) return false;
  while ( *(++s) );
  return true;
}/* isint() */


/*------------------------------------------------------------------------*/
void get_set( kind k, functor f, unsigned char set[], int x )
     
     /*Called by get_item().
      *Extracts setting information.*/
     
{
  if ( !strcmp(set,"solutions") ) setting.max_models = get_int();
  else if ( !strcmp(set,"backtracks") ) 
    setting.max_backtracks = get_int();
  else if ( !strcmp(set,"current_priority") )
    setting.latest_priority = get_int();
  else if ( !strcmp(set,"stack") ) {
    get_word(set);
    if ( set[0] == ':' ) get_word(set);
    if ( !strcmp(set,"itsy") ) setting.batches = S_ITSY;
    else if ( !strcmp(set,"bitsy") ) setting.batches = S_BITSY;
    else if ( !strcmp(set,"tiny") ) setting.batches = S_TINY;
    else if ( !strcmp(set,"small") ) setting.batches = S_SMALL;
    else if ( !strcmp(set,"medium") ) setting.batches = S_MEDIUM;
    else if ( !strcmp(set,"large") ) setting.batches = S_LARGE;
    else if ( !strcmp(set,"LARGE") ) setting.batches = S_VERYLARGE;
    else if ( !strcmp(set,"huge") ) setting.batches = S_HUGE;
    else if ( !strcmp(set,"maximal") ) setting.batches = S_MAXIMAL;
    else if ( isint(set) ) sscanf(set,"%d",&(setting.batches));
    else inwarning(set,"found where stack size expected");
    return;
  }
  else if ( !strcmp(set,"stack_margin") ) 
    setting.extra_batches = get_int();
  else if ( !strcmp(set,"verbosity") ) get_item( verb_spec, 0, 0 );
  else if ( !strcmp(set,"ref_length") ) setting.maxref = get_int();
  else if ( !strcmp(set,"pre_test") ) setting.pretest_limit = get_int();
  else if ( !strcmp(set,"test_ref_length") ) 
    setting.test_maxref = get_int();
  else if ( !strcmp(set,"time_limit") ) setting.maxtime = get_int();
  
  else if ( !strcmp(set,"primary") ) setting.Nosecs = true;
  else if ( !strcmp(set,"secondary") ) setting.Nosecs = false;
  
  else if ( !strcmp(set,"noisy") ) setting.noisy = true;
  else if ( !strcmp(set,"quiet") ) setting.noisy = false;
  
  else if ( !strcmp(set,"time_stamp") ) setting.timestamps = true;
  else if ( !strcmp(set,"no_time_stamp") ) setting.timestamps = false;
  
  else if ( !strcmp(set,"report_breaks") ) setting.report_breaks = true;
  else if ( !strcmp(set,"silent_breaks") ) setting.report_breaks = false;
  
  else if ( !strcmp(set,"tops_off") ) setting.topsoff = true;
  else if ( !strcmp(set,"tops_on") ) setting.topsoff = false;
  
  else if ( !strcmp(set,"reverse_test") ) setting.invert_change = true;
  else if ( !strcmp(set,"obverse_test") ) setting.invert_change = false;
  
  else if ( !strcmp(set,"test_by_vars") ) setting.vars_over_wt = true;
  else if ( !strcmp(set,"test_by_weight") ) setting.vars_over_wt = false;
  
  else if ( !strcmp(set,"pretty") ) setting.pm.pretty = true;
  else if ( !strcmp(set,"ugly") ) {
    setting.pm.pretty = false;
    setting.pm.verbosity = pugly;
  }
  else inwarning("Unrecognosed setting option",set);

} /* get_set() */



/*------------------------------------------------------------------------*/
void verb_spec( kind k, functor f, unsigned char s[], int x )

     /*Called by get_set() only.*/
{
  unsigned vb;
  
  vb = setting.pm.verbosity;
  if ( !strcmp(s,"models") ) vb = verb_spec_spec(vb,pmodels);
  else if ( !strcmp(s,"job") ) vb = verb_spec_spec(vb,pjob);
  else if ( !strcmp(s,"stats") ) vb = verb_spec_spec(vb,pstats);
  else {
    inwarning("Unrecognosed verbosity option",s);
    verb_spec_spec(0,0);
  }
  setting.pm.verbosity = vb;

} /*verb_spec()*/



/*------------------------------------------------------------------------*/
unsigned verb_spec_spec( unsigned vb, unsigned mask )

/*Called by verb_spec() only. */

{
    unsigned char s[LEXLENGTH];

    get_word(s);
    if ( *s == ':' ) get_word(s);
    vb &= ~mask;
    if ( !strcmp(s,"none") ) return( vb );
    if ( !strcmp(s,"brief") ) return( vb | (mask & (mask >> 1)) );
    if ( !strcmp(s,"full") ) return( vb | (mask & (mask << 1)) );
    inwarning(s, "found where level (none, brief or full) expected");
    return vb;

} /* verb_spec_spec() */



/*------------------------------------------------------------------------*/
boolean got_partial()

/*Called by FINDER fcn main.c. -- not relvant to SCOTT*/

{
/*    char s[LEXLENGTH];

    source = stdin;
    if ( !piping ) return false;
    get_word(s);
    if ( !strcmp(s,"end") ) return false;
    if ( strcmp(s,"model") ) {
	fprintf(stderr,"\n\"%s\" found where \"model\" expected\n",s);
	return false;
    }
    while ( got_function() );
*/
    return true;
}/* got_partial() */



/*------------------------------------------------------------------------*/
boolean got_function()

/* Called by got_partial(). -- not relvant to SCOTT*/

{
/*    char s[LEXLENGTH];
    int i, j;
    functor f;
    kind k;

    get_word(s);
    if ( *s == '.' ) return false;
    if ( !strcmp(s,"size") ) {
	get_word(s);
	if ( (i = lex_pos(s)) != -1)
	  if (lexicon[i].iskind )
	    k = sort+lexicon[i].offset;
	  else skipout("Incorrect sort name transfer",1);
	k->siz = get_int();
	return true;
    }
    if ( (i = lex_pos(s)) != -1)
      if (!lexicon[i].iskind ) 
	f = fn+lexicon[i].offset;
      else skipout("Incorrect function name transfer",1);
    switch( f->adicity ) {
    case 0:	*f->table = get_int();
	break;
      case 1:	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	  *(f->table+i) = get_int();
	*(f->table+i) = f->val_type->siz;
	break;
      case 2:	for ( i = 0; i <= f->arg_type[0]->siz; i++ )
	  for ( j = 0; j <= f->arg_type[1]->siz; j++ )
	      if ( i < f->arg_type[0]->siz && j < f->arg_type[1]->siz )
		  *(f->table+(SZ*i)+j) = get_int();
	      else
		  *(f->table+(SZ*i)+j) = f->val_type->siz;
    }
*/
    return true;

} /* got_function() */


/*------------------------------------------------------------------------*/
void inerror(char *s1, char *s2)

     /*Called by numerous functions in files:
      *FINDER.c, initial.c, input.c, parser.c, and prepare.c.
      
      *Print input error and exit with status set to current line number*/
{
  fprintf(stderr,"\nInput error on line %d: %s\n", current_line, job_line);
  fprintf(stderr,"\n%s %s\n\n",s1,s2);
  exit(current_line);
}


/*------------------------------------------------------------------------*/
void inwarning(char *s1, char *s2)

     /*Called by verb_spec(), verb_spec_spec(), get_set(), func_spec(), 
      *set_up_type(), sort_spec(), and get_key_word() (all in this file).
      
      *Print input warning but continue jobs.*/
{
  fprintf(stderr,"\nWarning: possible error on line %d: %s\n", 
	  current_line, job_line);
  fprintf(stderr,"%s %s\n\n\n",s1,s2);
}

/*================================ END ===================================*/
