/*--------------------------------------------------------------------------*/
/*Routines for testing clauses against existing models.*/
/*--------------------------------------------------------------------------*/

#include "FINDER.h"
//#define DEBUG

static unsigned thisvector[V_LENGTH];  /* Copy of the test vector   */


/*--------------------------------------------------------------------------*/
int test( unsigned info[], TRS *T )
     
     /*Called by model() and find_theory(), purely as a tester -- just
      *ends up callin is_good() -- and by transref for the serious search.

      *Decodes the vector info[] (the current model) into testable data 
      *structures. 
      *Tests any clauses of weight greater than 2 against this model.
      *Uses the external function Ref to record a refutation if one is found
      *If the model passes all the tests, sends the it to the printup routine.
      * 
      *Returns TRUE if all tests are passed
      *and FALSE if some refutation occurs.
      *
      *If the global flag negative flag is set then we want to keep 
      *structures that come through the whole generate-and-test procedure 
      *without ever passing the tests. 
      *So instead of calling the print function, model simply records that 
      *the clause was found to fail.  
      *This flag is set in the fcn main but nowhere else. */
     
{
  clause c;
  boolean cc_is_good = TRUE;
  
  vect_into_model(info);
  
  /* search for a bad guy successful? */
  if ( negative && !clauses_fail ) return FALSE;
  
  /*Initial call sets T=0; this calls is_good and sets up the modeling
    structure T such that initial latest_clause will be satisfied*/

  if ( !T ) { /* just perform a straight test */
    cc_is_good = is_good(&latest_clause);

    /* if searching for a bad guy this reverses the sense of the search */
    if ( negative ) clauses_fail = !cc_is_good;
    return (cc_badguy != cc_is_good); /* skip out with the result */
  }
  
  /* Get to here only if called by tranref and supplied with a T */

  /* Do the following iff SCOTT is operating */
  if ( latest_clause.prem || latest_clause.conc ) {
    /*Perhaps test each clause in the existing theory first*/
    for ( c = TH[current_theory]; c; c = c->nextclause )
      if (c->weight > setting.pretest_limit 
	  && is_after( &latest_clause, c ))
	
	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;           /* refutation found */
	}

    /*then test the new clause */
    cc_is_good = is_good(&latest_clause);
    if ( !cc_is_good && !cc_badguy ) {
      record_refutation( &latest_clause, T );
      return FALSE;               /* refutation found */
    }

    /*Perhaps test each clause in the existing theory after instead*/
    for ( c = TH[current_theory]; c; c = c->nextclause )
      if (c->weight > setting.pretest_limit
	  && is_after( c, &latest_clause ))

	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;           /* refutation found */
	}

  } /*if latest clause is undefined do old FINDER */
  else {
    for ( c = TH[current_theory]; c; c = c->nextclause ) {
      if ( c->weight > setting.pretest_limit )
	
	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;             /* refutation found */
	}
    }
  }

  /*Skip out if doing a negative search; else record some stats */  
  if ( negative ) clauses_fail = 0;
  else if ( cc_badguy && cc_is_good ) return TRUE;
  else {

    updates[current_theory]++; /* ehh? why here */

    if ( T->firstchange < 0 || T->firstchange >= tr_parameters.vlength )
      first_change = tr_parameters.vlength;
    else first_change = T->inorder[T->firstchange];

    set_last_model();
    print_model();
    if ((updates[current_theory] && 
	 updates[current_theory]==setting.max_models) ||
	single_model)
      job_done( &tr_parameters, "limit on number of solutions" );
  }

  return TRUE;
} /* test() */



/*--------------------------------------------------------------------------*/
void vect_into_model( unsigned vector[] )

     /*Called by test().
      *Translates vector[] (contains the values assigned to functions) 
      *into testable data structures.

      *The array thisvector[] is filled with a copy of info[] for purposes 
      *of deciding what should be printed, and of recording the current 
      *model where FINDER is used as a TESTER.
      *(Really just another copy of last_model[])

      *Note that piped functions do not feature in the vector, though they 
      *are in the list of testable functions. */

{
  functor f;
  int i, offset;
  int argument[ARGMAX];

  /* Store a copy of vector[] in case we need to record a new good model*/
  for ( i = 0; i < tr_parameters.vlength; i++ )
    thisvector[i] = vector[i];
  
  /* Translate to testable data structure */
  for ( f = FN; f; f = f->nextfunct ) 
    if ( !f->piped ) {
      if (f->adicity) f->junk = vector[f->fcc-1];
      for (i=0; i<f->adicity; i++) argument[i] = 0;
      do {
	offset = getoffset(f,argument);
	if (offset >= 0) *(f->table+offset) = vector[vec_pos(f,argument)];
      }
      while (another_argument(f,argument));
    }
} /* vect_into_model() */



/*--------------------------------------------------------------------------*/
void set_last_model()
     
     /*Called by test().
      *If all the tests are passed, stores the most recent good model of 
      *the current theory in last_model[] and last_sizes[], and olf_fcc[] 
      *fields of the function symbols. 
      *These are later used by model() for testing new clauses*/

{
  int i;
  functor f;
  
  for ( i = 0; i < tr_parameters.vlength; i++ )
    last_model[current_theory][i] = thisvector[i];

  for ( i = 0; sort[i].sname; i++ ) {
    last_sizes[current_theory][i] = sort[i].siz;
    last_sizmasks[current_theory][i] = sort[i].sizmask;
  }
  
  for ( f = FN; f; f = f->nextfunct ) 
    f->old_fcc[current_theory] = f->fcc;
  
  last_vlength[current_theory] = tr_parameters.vlength;
  
} /*  set_last_model() */



/*--------------------------------------------------------------------------*/
boolean is_good( clause cl )

     /*Called by test().
      *Returns TRUE iff the clause is satisfied by the current model.*/
     
{
  wff_stack wf;
  
  /* if too many variables, search for a refutation instead of testing*/
  if ( cl->varcount > VCMAX ) return tisgood(cl);
  
  for ( wf = cl->varlist; wf; wf = wf->nxt ){
    /* set the first assignment to variables */
    /* note variables should never have a val_type->siz=0 */
    wf->w->val = (setting.invert_change? 0:wf->w->sym->val_type->siz-1 );
#ifdef DEBUG
    if ( wf->w->val < 0) {
      inerror(wf->w->sym->fname,
	      ": gets negative initial assignment in Ftest.c:is_good()");
    }
#endif
    
  }
  do
    if (clause_fails(cl) )
      return false;
  while (next_assignment(cl));

  return true;
  
} /* is_good() */



/*--------------------------------------------------------------------------*/
void sanity_check(wff w)

     /*Called by clause_fails() .
      *Check that formula is well formed */
{
  int i;
  boolean sane=true;

  if (w->matrix != w->sym->table) {
    report_insanity("Incorrect table pointer",w,-1);
    sane = false;
  }
  if (w->val > w->sym->val_type->siz || /* value assigned is greather than 
					   the number of possible values*/
      w->val < 0 || /* value assigned is negative */
      
      ((w->val == w->sym->val_type->siz) /* either no assignment */
       && w->sym->total)) {
      report_insanity("Value out of range",w,-1);
      sane = false;

    }
  
  for (i=0; i < w->sym->adicity ; i++)
    if (w->sbval[i] != &(w->sb[i]->val)) {
      if (!sane) {
	report_insanity("Mismatch of subvalues",0,i);
      }
      else {
	report_insanity("Mismatch of subvalues",w,i);
      }
      sane = false;
    }
  if (!sane) exit(3);
} /* sanity_check() */



/*--------------------------------------------------------------------------*/
void report_insanity(char error_message[], wff w, int code)

     /* Called by  sanity_check() */

{
  if (w) {
    fprintf(stderr,"Insane formula ");
    outwff(stderr,w);
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"%s",error_message);
  if (code>=0) fprintf(stderr," (%d)",code);
  fprintf(stderr,"\n");
} /* insane() */


/*--------------------------------------------------------------------------*/
boolean clause_fails( clause c )

     /*Called by is_good().
      *This is the actual test. 
      *The value of each subformula is calculated and assigned to its 
      *_val field. Then the premises and conclusions of the clause are 
      *checked to see whether it has been refuted.
      *Returns TRUE iff the clause is true in the current model.*/
     
{
  /*
  wff_stack wf;
  int i, offset, arity;
  int subvalue[ARGMAX];
  int **sbval;
  int *fats;

  // assign values

  boolean failure=true;

  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    for (i=0; i< wf->w->sym->adicity; i++){
#ifdef DEBUG
      if (*(wf->w->sbval[i]) == -1) {
	inerror(wf->w->sym->fname,": unassigned in Ftest.c:clause_fails()");
      }
#endif
      subvalue[i] = *(wf->w->sbval[i]);
    }
    wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalue));
#ifdef DEBUG
    sanity_check(wf->w); 
#endif
  }
  
  // Check for refutation
  failure = true;
  for ( wf = c->prem; wf; wf = wf->nxt )
    if ( !wf->w->val ) failure = false;
  for ( wf = c->conc; wf; wf = wf->nxt )
    if ( wf->w->val ) failure = false;

  return failure;
  */

  /* optimisation fix 1. hack in get_offset() */
  /*
  wff_stack wf;
  int i, offset;
  int subvalue[ARGMAX];

   for ( wf = c->testlist; wf; wf = wf->nxt ) {
    offset = 0;
    for (i=0; i < wf->w->sym->adicity; i++){
      subvalue[i] = *(wf->w->sbval[i]);
      offset += (subvalue[i] *  wf->w->sym->fats[i]);
    }
    wf->w->val = *(wf->w->matrix + offset);
  }
  */

  /* optimisation fix 2.  */
  /*
  wff_stack wf;
  int i, offset, arity;
  int **sbval;
  int *fats;
  int subvalue[ARGMAX];

  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    offset = 0;
    arity =  wf->w->sym->adicity;
    sbval = wf->w->sbval;
    fats = wf->w->sym->fats;
    for (i=0; i < arity; i++, fats++, sbval++){
      subvalue[i] = *(*sbval);
      offset += (subvalue[i] * (*fats));
    }
    wf->w->val = *(wf->w->matrix + offset);
  }
  */

  /* optimisation fix 3.  */
  wff_stack wf;
  int i, offset, arity;
  int **sbval;
  int *fats;

  for ( wf = c->testlist; wf; wf = wf->nxt ) {
#ifdef DEBUG
      if (*(wf->w->sbval[i]) == -1) {
	inerror(wf->w->sym->fname,": unassigned in Ftest.c:clause_fails()");
      }
#endif
    offset = 0;
    arity =  wf->w->sym->adicity;
    sbval = wf->w->sbval;
    fats = wf->w->sym->fats;
    for (i=0; i < arity; i++, fats++, sbval++)
      offset += ( (*(*sbval)) * (*fats));
    wf->w->val = *(wf->w->matrix + offset);
#ifdef DEBUG
    sanity_check(wf->w); 
#endif
  }

  /* optimisation fix 0. */
  for ( wf = c->conc; wf; wf = wf->nxt ) if ( wf->w->val ) return false;
  for ( wf = c->prem; wf; wf = wf->nxt ) if ( !wf->w->val ) return false;
  return true;
  
} /* clause_fails() */



/*--------------------------------------------------------------------------*/
boolean next_assignment( clause c )
     
     /*Called by is_good().
      *Gets the next assignment of values to the variables in the clause c.
      *Returns FALSE iff all possible values have been exhausted.*/

{
  wff_stack wf;
  boolean not_invert_change;
  int pretest_limit;

  // rather than calculate inside the for loop
  not_invert_change = !setting.invert_change; 
  pretest_limit = setting.pretest_limit;

  /* foreach variable ... */
  for ( wf = c->varlist; wf; wf = wf->nxt )
    if ( c->weight <= pretest_limit  || not_invert_change ) {
      if ( !wf->w->val ) wf->w->val = wf->w->sym->val_type->siz-1;
      else {
	wf->w->val--;
	return true;
      }
    }
    else {
      if ( wf->w->val == wf->w->sym->val_type->siz-1 ) wf->w->val = 0;
      else {
	wf->w->val++;
	return true;
      }
    }
  return false;
} /* next_assignment() */



/*--------------------------------------------------------------------------*/
void record_refutation( clause c, TRS *T )

     /*Called by test().
      *Describes the new refutaion to transref.*/
     
{
  wff_stack wf;
  int argument[ARGMAX];
  int i;
  
  for ( wf = c->testlist; wf; wf = wf->nxt )
    if ( is_open(wf->w->sym) ) {
      for (i=0; i<wf->w->sym->adicity; i++)
	argument[i] = *(wf->w->sbval[i]);
      Ref( vec_pos(wf->w->sym,argument), T );
    }
} /* record_refutation() */


/*------------------------------------------------------------------------*/
void lift_clause(clause c)

     /*Called by test().
      *Lifts clauses that fail up the order of testing, 
      *so that the clauses doing the work congregate at the top.*/
     
{
  clause d,e;
  
  if (c==TH[current_theory]) return;
  for (d=TH[current_theory]; d->nextclause != c; d = d->nextclause);
  if (d==TH[current_theory]) {
    d->nextclause = c->nextclause;
    c->nextclause = d;
    TH[current_theory] = c;
  }
  else {
    for (e=TH[current_theory]; e->nextclause != d; e = e->nextclause);
    d->nextclause = c->nextclause;
    c->nextclause = d;
    e->nextclause = c;
  }
} /* lift_clause() */

/*=============================== END ====================================*/




