/*========================================\
| FILE: auto.c: SCOTT's autonomous mode.  |
\========================================*/

#include "scott.h"

#define MAX_PROPERTIES 20

#define NUM_CL         0
#define NUM_POS_CL     1
#define NUM_NEG_CL     2
#define NUM_MIX_CL     3
#define NUM_EQ_CL      4
#define NUM_UNIT_CL    5
#define NUM_GRD_CL     6
#define NUM_HORN_CL    7
#define MAX_LITS       8
#define MIN_LITS       9
#define MAX_WT        10
#define MIN_WT        11
#define MAX_DIST_VARS 12
#define MIN_DIST_VARS 13
    
static int property[MAX_PROPERTIES];

boolean all_clauses(int (*property)()) 
/*---------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns TRUE iff every clause on |
 | the USABLE, SOS, and INPUT_FIRTS lists have the given property.     |
 *---------------------------------------------------------------------*/
{
    struct clause *c;
    boolean test = true;
    
    for (c=Usable->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    for (c=Sos->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    for (c=First->first_cl; c && test; c=c->next_cl)
      test = (*property)(c);
    return test;

} /*--all_clauses()--*/

boolean at_least_one_clause(int (*property)())
/*------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns TRUE iff at least one |
 | clause on the USABLE, SOS, and INPUT_FIRTS lists has the given   |
 | property.                                                        |
 *------------------------------------------------------------------*/
{
    struct clause *c;
    boolean test = false;
    
    for (c=Usable->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    for (c=Sos->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    for (c=First->first_cl; c && !test; c=c->next_cl)
      test = (*property)(c);
    return test;

} /*--at_least_one_clause()--*/

int get_max(int (*property)())
/*---------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns maximum value across the |
 | USABLE, SOS, and INPUT_FIRST lists of the given property.           |
 *---------------------------------------------------------------------*/
{    
  struct clause *c;
  int next, max = 0;

  for (c=Usable->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  for (c=Sos->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  for (c=First->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    max = (next > max ? next : max);
  }
  return max;

}/*--get_max()--*/

int get_min(int (*property)())
/*---------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns minimum value across the |
 | USABLE, SOS, and INPUT_FIRST lists of the given property.           |
 *---------------------------------------------------------------------*/
{    
  struct clause *c;
  int next, min = MAX_INT;

  for (c=Usable->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  for (c=Sos->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  for (c=First->first_cl; c; c=c->next_cl) {
    next = (*property)(c);
    min = (next < min ? next : min);
  }
  return min;

}/*--get_min()--*/

int get_num(int (*property)())
/*---------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns number of clauses on the |
 | USABLE, SOS, and INPUT_FIRST lists of the given property.           |
 *---------------------------------------------------------------------*/
{    
  struct clause *c;
  int num = 0;

  for (c=Usable->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  for (c=Sos->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  for (c=First->first_cl; c; c=c->next_cl) 
    if ((*property)(c)) num++;

  return num;

}/*--get_num()--*/


int clause_weight(struct clause *c)
{ 
  return (weight_cl(c, Weight_pick_given_index));
}

/*----------------------------------------------*
 | functions that identify various clause types |
 *----------------------------------------------*/

int any_clause(struct clause *c)
{ 
  return 1;
}

int nonground_clause(struct clause *c)
{ 
  return (!ground_clause(c));
}

int mix_clause(struct clause *c)
{ 
  return (!(pos_clause(c) || neg_clause(c)));
}

int mix_ground_clause(struct clause *c)
{ 
  return (!(pos_clause(c) || neg_clause(c)) && ground_clause(c));
}

int mix_nonground_clause(struct clause *c)
{ 
  return (!(pos_clause(c) || neg_clause(c)) && !ground_clause(c));
}


int pos_ground_clause(struct clause *c)
{ 
  return (pos_clause(c) && ground_clause(c));
}


int pos_ground_unit_clause(struct clause *c)
{ 
  return (pos_clause(c) && ground_clause(c) && unit_clause(c));
}

int pos_ground_nonunit_clause(struct clause *c)
{ 
  return (pos_clause(c) && ground_clause(c) && !unit_clause(c));
}

int pos_nonground_clause(struct clause *c)
{ 
  return (pos_clause(c) && !ground_clause(c));}

int pos_nonground_unit_clause(struct clause *c)
{ 
  return (pos_clause(c) && !ground_clause(c) && unit_clause(c));
}

int pos_nonground_nonunit_clause(struct clause *c)
{ 
  return (pos_clause(c) && !ground_clause(c) && !unit_clause(c));
}


int neg_ground_clause(struct clause *c)
{ 
  return (neg_clause(c) && ground_clause(c));}


int neg_ground_unit_clause(struct clause *c)
{ 
  return (neg_clause(c) && ground_clause(c) && unit_clause(c));
}

int neg_ground_nonunit_clause(struct clause *c)
{ 
  return (neg_clause(c) && ground_clause(c) && !unit_clause(c));
}

int neg_nonground_clause(struct clause *c)
{ 
  return (neg_clause(c) && !ground_clause(c));}

int neg_nonground_unit_clause(struct clause *c)
{ 
  return (neg_clause(c) && !ground_clause(c) && unit_clause(c));
}

int neg_nonground_nonunit_clause(struct clause *c)
{ 
  return (neg_clause(c) && !ground_clause(c) && !unit_clause(c));
}


int equality_axiom(struct clause *c)
/*-----------------------------------------------------------------------*
 | Called by auto_set_conjecture() Returns true iff "c" is a symmetry or |
 | reflexivity clasue for equality.                                      |
 *-----------------------------------------------------------------------*/
{
  if (symmetry_clause(c)) return 1;
  if (num_literals(c) != 1) return 0;
  if (!eq_lit(c->first_lit)) return 0;
  else {
    struct term *atom;
    
    atom = c->first_lit->atom;
    
    if (!atom->farg->argval->type == VARIABLE)
      return 0;
    
    if (!atom->farg->narg->argval->type == VARIABLE)
      return 0;
    
    if (!term_ident(atom->farg->argval,
		    atom->farg->narg->argval))
      return 0;
  }
  return 1;

}/*--equality_axiom()--*/

boolean has_pos_nonground(struct list *lptr);

boolean has_pos_nonground(struct list *lptr)
/*----------------------------------------------------------------------*
 | Called by automatic_3_settings().  Returns true iff given list has a |
 | positive non-ground clause                                           |
 *----------------------------------------------------------------------*/
{
    struct clause *c;

    for (c = lptr->first_cl; c; c = c->next_cl) {
      if (pos_clause(c) && !ground_clause(c))
	return(1);
    }
    return (0);

} /*--has_pos_nonground()--*/

boolean has_neg_nonground(struct list *lptr);

boolean has_neg_nonground(struct list *lptr)
  /*----------------------------------------------------------------------*
   | Called by automatic_3_settings().  Returns true iff given list has a |
   | negative non-ground clause                                           |
   *----------------------------------------------------------------------*/
{
    struct clause *c;

    for (c = lptr->first_cl; c; c = c->next_cl) {
      if (neg_clause(c) && !ground_clause(c))
	return(1);
    }
    return (0);

} /*--has_neg_nonground()--*/

void automatic_3_settings()
  /*-----------------------------------*
   | Called by read_all_input().       |
   |                                   |
   | SCOTTs very own autonomous mode.  |
   *-----------------------------------*/
{
  boolean propositional, horn, unit, equality, pure_eq, symmetry;

  if (Passive->first_cl)
    abend("Passive list not accepted in auto3 mode.");
  else if (Demodulators->first_cl)
    abend("Demodulators list not accepted in auto3 mode.");
  else if (Weight_pick_given || Weight_purge_gen || Weight_terms)
    abend("Weight lists not accepted in auto3 mode.");

  /*-------------------------------------------------.
  | Don't use initial INPUT_FIRST list in auto3 mode |
  `-------------------------------------------------*/

  if (First->first_cl) {
    move_clauses(any_clause, First, Sos);

  }
  
  printf("\n------------> automatic settings\n");
  
  /* Find out some basic properties. */
  
  propositional = all_clauses(propositional_clause);
  equality = at_least_one_clause(equality_clause);
  symmetry = at_least_one_clause(symmetry_clause);
  horn = all_clauses(horn_clause);
  unit = all_clauses(unit_clause);
  pure_eq = all_clauses(equality_clause);

  printf("\nInput clauses form a %s%s set%s%s.\n\n",
	 (propositional ? "propositional, " : ""), 
	 (unit ? "unit" : (horn ? "horn" : "non-horn")), 
	 (pure_eq ? " all with equality" :
	  (equality ? " with some equality clauses" : 
	   " without any equality clauses")),
	 (symmetry ? " and with a symmetry clause" : "")
	 );
  
  set_properties();

//   /*-----------------------------------------------.
//   | for AUTO1 all positive clause are moved to SOS |
//   `-----------------------------------------------*/
//
//  if (has_pos_nonground(Sos) || has_pos_nonground(First))
//     printf("\nSOS has positive nonground clause; therefore it is not changed.\n");
//   else {
//      /*-----------------------------------------------------.
//      | this helps to guard againts a search based purely on |
//      | ground clauses                                       |
//      `-----------------------------------------------------*/
//     printf("\nEvery positive clause in SOS is ground (or SOS is empty);\n");
//     printf("therefore we move all positive usable clauses to SOS.\n");
//     move_clauses(pos_clause, Usable, Sos);
//   }

  // opt for AUTO1 behaviour
  move_clauses(pos_clause, Usable, Sos);

 
  if (propositional) {
    printf("\nAll clauses are propositional; therefore we set the\n");
    printf("propositional flag and use ordered hyperresolution.\n");
    auto_change_flag(stdout, PROPOSITIONAL, 1);
    auto_change_flag(stdout, HYPER_RES, 1);
    return; // saves on a nested conditionals
  }
  
  /* nonpropositional case */
  
  if (property[MAX_LITS] == 1) {
    printf("\nSetting pick_given_ratio to 2, ");
    printf("because all clauses are units.\n");
    auto_change_parm(stdout, PICK_GIVEN_RATIO, 2);
  }
  else {
    /* Nonunit */
    printf("\nSetting hyper_res, because there are nonunits.\n");
    auto_change_flag(stdout, HYPER_RES, 1);
    
    if (equality || !horn) {
	 /*---------------------------------------------------.
	 | main difference between AUTO1 and AUTO2 is setting |
	 | this flag                                          |
	 `---------------------------------------------------*/
      printf("\nSetting ur_res, because this is a nonunit set containing\n");
      printf("either equality literals or non-Horn clauses.\n");
      auto_change_flag(stdout, UR_RES, 1);
   }
     
    if (horn) {
      printf("\nClearing order_hyper, because all clauses are Horn.\n");
      auto_change_flag(stdout, ORDER_HYPER, 0);
    }
  
    if (!horn) {
      /* Non-Horn */
      printf("\nSetting factor and  unit_deletion, ");
      printf("because there are non-Horn clauses.\n");
      auto_change_flag(stdout, FACTOR, 1);
      auto_change_flag(stdout, UNIT_DELETION, 1);
    }
  }

  if (equality) {
    printf("\nEquality is present, so we set the knuth_bendix flag.\n");
    auto_change_flag(stdout, KNUTH_BENDIX, 1);
    if (property[MAX_LITS] > 1) {
      printf("\nAs an incomplete heuristic, we paramodulate with units only.\n");
      auto_change_flag(stdout, PARA_FROM_UNITS_ONLY, 1);
      auto_change_flag(stdout, PARA_INTO_UNITS_ONLY, 1);
    }
    if (symmetry) {
      printf("\nThere is a clause for symmetry of equality, so it is\n");
      printf("assumed that equality is fully axiomatized; therefore,\n");
      printf("paramodulation is disabled.\n\n");
      auto_change_flag(stdout, PARA_FROM, 0);
      auto_change_flag(stdout, PARA_INTO, 0);
    }
  }
  
  
}/*--automatic_3_settings()--*/


void auto_set_sos()
/*------------------------------------------------------------------------*
 |    Called by init_semantics() *after* check_conjecture_list().         |
 |                                                                        |
 | Moves SOS clauses to the USABLE iff we can show that the resulting     |
 | list is consistent and does not interfere with the completeness of any |
 | inference rules.                                                       |
 *------------------------------------------------------------------------*/
{
  struct clause *cl, *next;
  char finder_cl[CLAUSE_STRING_ARRAY];
  int new_time_limit = 0;
  boolean first;
  
  // sos strategy conflicts with pure paramodulation
  if (property[NUM_UNIT_CL] == property[NUM_EQ_CL]
      && Flags[KNUTH_BENDIX].val) return;

  message(2,"");
  message(2,"------------> checking set of support strategy:");
  message(2,"");

  // Use maxiset position 1 for this testing
  // and set maximum time limit for modelling input clauses
  set_time_limit(tParms[MAX_SEARCH_TIME].val);

  // model the Usable
  first = true;
  cl=Usable->first_cl; 
  while (cl) { 
    next = cl->next_cl;
    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (!model(finder_cl, 1, false, true)) {
      if (first) {
	message(2,"");
	message(2,"====> Moving the following USABLE clauses to the SOS");
	message(2,"");
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      append_cl(Sos,cl);
      tStats[SOS_STRATEGY_REFINEMENTS]++;
    }
    cl = next;
  }

  // now model the SOS
  first = true;
  cl=Sos->first_cl; 
  while (cl) { 
    next = cl->next_cl;
    
    // dont move positive clauses if using positive hyper_res
    if (pos_clause(cl) && Flags[HYPER_RES].val){
      cl = next;
      continue;
    }
    // don't move negative clauses if using negative hyper_res
    if (neg_clause(cl) && Flags[NEG_HYPER_RES].val){
      cl = next;
      continue;
    }    

    // don't move unit clauses if using UR-res
    if (unit_clause(cl) && Flags[UR_RES].val){
      cl = next;
      continue;
    }

    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (model(finder_cl, 1, false, true)) {
      if (first) {
	message(2,"");
	message(2,"====> Moving the following SOS clauses to the USABLE list");
	message(2,"");
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      first = false;
      rem_from_list(cl);
      append_cl(Usable,cl);
      tStats[SOS_STRATEGY_REFINEMENTS]++;
    }
    cl = next;
  }
   
  // this should give us an appropriate starting time_limit
  new_time_limit = get_max_time_in(1);
  if (new_time_limit < tParms[MIN_SEARCH_TIME].val)
    set_time_limit(tParms[MIN_SEARCH_TIME].val);
  else {
    message(2, "Setting initial search time limit to %d", new_time_limit);
    set_time_limit(new_time_limit);
  }

  // reinit tempory maxiset position 1
  init_sizes(1);
  init_theory(1);

  message(2,"");
  
}/*--auto_set_sos()--*/


int_function auto_set_conjecture()
/*------------------------------------------------------------------*
 |               Called by init_conjecture_list().                  |
 |                                                                  |
 | Determines the most likely properties of a conjecture clause and |
 | ensures that the input first list is set appropriately. Set the  |
 | external list "conjecture".                                      |
 *------------------------------------------------------------------*/
{
  int (*conj_clause_type)() = NULL;
  struct clause *cl, *next;
  boolean first;
  char type_name[100] = "";

  conj_clause_type = rare_clause_type(type_name);

  if (conj_clause_type == &any_clause) {
    message(2,"");
    message(2, "Could not establish a suitable conjecture clause type.", type_name);
    message(2,"");
    return NULL;
  }

  message(2,"");
  message(2, "So the conjecture is probably expressed by %s clauses.", type_name);
  message(2,"");

  first = true;
  cl=First->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if ((*conj_clause_type)(cl)) {
      if (first) {
	message(2,"");
	message(2,"Adding the following INPUT_FIRST clauses to the conjecture list:");
	message(2,"");
	first = false;
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      cl->stype = Conj;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[CONJECTURE_SIZE]++;
    }
    cl = next;
  }

  first = true;
  cl=Sos->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if ((*conj_clause_type)(cl)) {
      if (first) {
	message(2,"");
	message(2,"Adding the following SOS clauses to the conjecture list:");
	message(2,"");
	first = false;
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      cl->stype = Conj;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[CONJECTURE_SIZE]++;
    }
    cl = next;
  }
  
  first = true;
  cl=Usable->first_cl; 
  while (cl) {
    next = cl->next_cl;
    if ((*conj_clause_type)(cl)) {
      if (first) {
	message(2,"");
	message(2,"Adding the following USABLE clauses to the conjecture list:");
	message(2,"");
	first = false;
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      cl->stype = Conj;
      append_cl_ext(cl,&conjecture); 
      tStats[CONJECTURE_SIZE]++;      
    }
    cl = next;
  } 

  return conj_clause_type;

}/*--auto_set_conjecture()-- */


int_function rare_clause_type(char *type_message)
  /*--------------------------------------------------------------------*
   | Called by auto_set_conjecture().  Returns a pointer to a function  |
   | that identifies clauses that least often within the clause space.  |
   *--------------------------------------------------------------------*/
{
  int (*clause_type)() = &any_clause;
  int smallest = MAX_INT; 
  int current = 0;

  if (at_least_one_clause(ground_clause) && 
      at_least_one_clause(nonground_clause)) {

    if (at_least_one_clause(unit_clause)) { // the full mix

      // ground units
      current = get_num(neg_ground_unit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative ground unit");
	smallest = current;
	clause_type = &neg_ground_unit_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_ground_unit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive ground unit");
	smallest = current;
	clause_type = &pos_ground_unit_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      // nonground units
      current = get_num(neg_nonground_unit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative non-ground unit");
	smallest = current;
	clause_type = &neg_nonground_unit_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_nonground_unit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive non-ground unit");
	smallest = current;
	clause_type = &pos_nonground_unit_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      // ground non-units
      current = get_num(neg_ground_nonunit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative ground non-unit");
	smallest = current;
	clause_type = &neg_ground_nonunit_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_ground_nonunit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive ground non-unit");
	smallest = current;
	clause_type = &pos_ground_nonunit_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(mix_ground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "mixed and ground");
	smallest = current;
	clause_type = &mix_ground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      // nonground non-units
      current = get_num(neg_nonground_nonunit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative non-ground non-unit");
	smallest = current;
	clause_type = &neg_nonground_nonunit_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_nonground_nonunit_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive non-ground non-unit");
	smallest = current;
	clause_type = &pos_nonground_nonunit_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(mix_nonground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "mixed non-ground");
	smallest = current;
	clause_type = &mix_nonground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
    }
    else { // all non-units
      // ground
      current = get_num(neg_ground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative ground");
	smallest = current;
	clause_type = &neg_ground_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_ground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive ground");
	smallest = current;
	clause_type = &pos_ground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(mix_ground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "mixed ground");
	smallest = current;
	clause_type = &mix_ground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      // non ground
      current = get_num(neg_nonground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "negative non-ground");
	smallest = current;
	clause_type = &neg_nonground_clause;
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(pos_nonground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "positive non-ground");
	smallest = current;
	clause_type = &pos_nonground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
      current = get_num(mix_nonground_clause);
      if (current > 0 && smallest > current) {
	strcpy(type_message, "mixed non- ground");
	smallest = current;
	clause_type = &mix_nonground_clause; 
	message(3,"%s clauses: %d.", type_message, smallest);
      }
    }
  }
  else { // all ground or all nonground...
    current = get_num(neg_clause);
    if (current > 0 && smallest > current) {
      strcpy(type_message, "negative");
      smallest = current;
      clause_type = & neg_clause;
      message(3,"%s clauses: %d.", type_message, smallest);
    }
    current = get_num(pos_clause);
    if (current > 0 && smallest > current) {
      strcpy(type_message, "positive");
      smallest = current;
      clause_type = &pos_clause; 
      message(3,"%s clauses: %d.", type_message, smallest);
    }
    current = get_num(mix_clause);
    if (current > 0 && smallest > current) {
      strcpy(type_message, "mixed");
      smallest = current;
      clause_type = &mix_clause; 
      message(3,"%s clauses: %d.", type_message, smallest);
    }
  }

  return clause_type;

}/*--rare_clause_type()--*/


void set_properties()
/*------------------------------------------------------------------*
 | Called by automatic_3_settings().  Set clause space properties.  |
 *------------------------------------------------------------------*/
{
  property[MAX_LITS] = get_max(num_literals);
  property[MIN_LITS] = get_min(num_literals);
  property[MAX_WT] = get_max(clause_weight);
  property[MIN_WT] = get_min(clause_weight);
  property[MAX_DIST_VARS] = get_max(distinct_vars);
  property[MIN_DIST_VARS] = get_min(distinct_vars);
  
  property[NUM_CL] = Stats[SOS_SIZE] +Stats[USABLE_SIZE];
  property[NUM_POS_CL] = get_num(pos_clause);
  property[NUM_NEG_CL] = get_num(neg_clause);
  property[NUM_MIX_CL] = get_num(mix_clause);
  property[NUM_EQ_CL] = get_num(equality_clause);
  property[NUM_HORN_CL] = get_num(horn_clause);
  property[NUM_UNIT_CL] = get_num(unit_clause);
  property[NUM_GRD_CL] = get_num(ground_clause);
  
  printf("There are %3d clauses of which ...\n"
	 "          %3d are positive,\n"
	 "          %3d are negative,\n"
	 "          %3d are mixed,\n"  
	 "          %3d are horn,\n"
	 "          %3d are unit,\n"
	 "          %3d are ground,\n"	   
	 "          %3d are equality.\n\n"
	 "Clause  weight ranges from %d to %d.\n"  
	 "Variable count ranges from %d to %d.\n" 
	 "Literal  count ranges from %d to %d.\n",
	 property[NUM_CL], 
	 property[NUM_POS_CL], 
	 property[NUM_NEG_CL], 
	 property[NUM_MIX_CL], 
	 property[NUM_HORN_CL], 
	 property[NUM_UNIT_CL], 
	 property[NUM_GRD_CL], 
	 property[NUM_EQ_CL],
	 property[MIN_WT], property[MAX_WT],
	 property[MIN_DIST_VARS], property[MAX_DIST_VARS],
	 property[MIN_LITS], property[MAX_LITS]);
  
}/*--set_properties()--*/

void check_eq_settings()
{
  if (at_least_one_clause(equality_clause) 
      && property[MAX_LITS] > 1){
    printf("\nAs an incomplete heuristic, we paramodulate with units only.\n");
    auto_change_flag(stdout, PARA_FROM_UNITS_ONLY, 1);
    auto_change_flag(stdout, PARA_INTO_UNITS_ONLY, 1);
  }
}
