/*=================================================\
|               FILE: clause_space.c               |
|                                                  |
| SCOTT functions that operate on the clause space |
\=================================================*/

#include "scott.h"

#if  (MAX_VARS < (VARMAX-2))
#define SCOTT_MAX_VARS MAX_VARS   // OTTER maximum
#else
#define SCOTT_MAX_VARS (VARMAX-2) // FINDER maximum
#endif

static struct clause_dptr *clause_dptr_avail;
static long clause_dptr_gets, clause_dptr_frees, clause_dptr_avails;

void init_clause_dptr()  
/*-------------------------------------------------------------*
 | Called by init_clause_space().  Initialise dynamic memory.  |
 *-------------------------------------------------------------*/
{
  clause_dptr_frees = 0;
  clause_dptr_avails = 0;
  clause_dptr_gets = 0;
  clause_dptr_avail = NULL;

}/*--init_clause_space()--*/

struct clause_dptr *get_clause_dptr()
/*------------------------------------------------------------------------*
 | Called by prepend_cl_ext() and append_cl_ext().  Allocates appropriate |
 | memory block and returns a pointer to the block.                       |
 *------------------------------------------------------------------------*/
{
    struct clause_dptr *p;

    clause_dptr_gets++;
    if (clause_dptr_avail == NULL)
	p = (struct clause_dptr *) tp_alloc((int) sizeof(struct clause_dptr));
    else {
	clause_dptr_avails--;
	p = clause_dptr_avail;
	clause_dptr_avail = clause_dptr_avail->next;
	}
    p->next = NULL;
    p->prev = NULL;
    p->c = NULL;
    return(p);

}/*--get_clause_dptr()--*/


void free_clause_dptr(struct clause_dptr *p)
/*------------------------------------------------------------------------*
 | Called by delete_cl_ext().  Frees the memory associated with the given |
 | pointer.                                                               |
 *------------------------------------------------------------------------*/
{
    clause_dptr_frees++;
    clause_dptr_avails++;
    p->next = clause_dptr_avail;
    clause_dptr_avail = p;

}/*--free_clause_dptr()--*/


void delete_cl_ext(struct clause *c, struct external_list *elist)
/*-----------------------------------------------------------.
| Called by numerous functions.  Deletes the pointer to c in |
| the given doubly linked list                               |
`-----------------------------------------------------------*/
{
  struct clause_dptr *cp, *prev_cp;

  cp = elist->first_cp;
  prev_cp = NULL; 

  while (cp && cp->c != c){
    prev_cp = cp;   // remember the predecessor
    cp = cp->next;
  }
  
  if (!cp || cp->c != c) 
    abort_scott("delete_cl_ext() could not find clause "
		"in the given list of pointers.");

  if (!prev_cp) // first in the list
    elist->first_cp = cp->next;
  else 
    prev_cp->next = cp->next; // skip forward over cp

  if (!cp->next)  // last in the list
    elist->last_cp = prev_cp;
  else 
    cp->next->prev = prev_cp; // skip backward over cp

  free_clause_dptr(cp);
  
}/*--delete_cl_ext()--*/

void prepend_cl_ext(struct clause *c, struct external_list *elist)
/*-----------------------------------------------------------------------*
 | Called by start_scan(), scan_sos(), scan_base(), sos_scan_insert()    |
 |                                                                       |
 | Establishes a new pointer to c at or near the begining of the given   |
 | pointer list.  If the clause has only one model an attempt is made to |
 | ensure that models alternate.                                         |
 *-----------------------------------------------------------------------*/
{
  struct clause_dptr *new_cp;
  
  new_cp = get_clause_dptr();
  new_cp->c = c;

  // list is empty
  if(!elist->last_cp) {
    elist->first_cp = elist->last_cp = new_cp;
    return;
  }
  
  if (c->num_sets == 1) { // alternate maxisets

    struct clause_dptr *prev_cp, *next_cp;

    prev_cp = NULL;
    next_cp = elist->first_cp;
    while (next_cp) {
      // ensure next clause has different models
      if (strncmp(next_cp->c->in_set,c->in_set,MAXISET_ARRAY) != 0){
	if (prev_cp) prev_cp->next = new_cp;
	else elist->first_cp = new_cp;
	next_cp->prev = new_cp;
	new_cp->next = next_cp;
	new_cp->prev = prev_cp;
	return;
      }
      prev_cp = next_cp;
      next_cp = next_cp->next;
    }
  }
  // otherwise, or if this fails, stick it in the front
  new_cp->next = elist->first_cp;
  elist->first_cp->prev = new_cp;
  elist->first_cp = new_cp;

}/*--prepend_cl_ext()--*/


void append_cl_ext(struct clause *c, struct external_list *elist)
/*-----------------------------------------------------------------------*
 | Called by start_scan(), scan_sos(), scan_base(), sos_scan_insert(),   |
 | init_clause_space().                                                  |
 |                                                                       |
 | Establishes a new pointer to c at or near the end ofthe given pointer |
 | list.  If the clause has only one model, an attempt is made to ensure |
 | that models alternate.                                                |
 *-----------------------------------------------------------------------*/
{
  struct clause_dptr *new_cp;
  
  new_cp = get_clause_dptr();
  new_cp->c = c;

  // list is empty
  if(!elist->last_cp) {
    elist->first_cp = elist->last_cp = new_cp;
    return;
  }

  if (c->num_sets == 1) { // alternate maxisets

    struct clause_dptr *prev_cp, *next_cp;

    //scan backward
    next_cp = NULL;
    prev_cp = elist->last_cp;
    while (prev_cp) {
      // ensure prev clause has different models
      if (strncmp(prev_cp->c->in_set,c->in_set,MAXISET_ARRAY) != 0){
	if (next_cp) next_cp->prev = new_cp;
	else elist->last_cp = new_cp;
	prev_cp->next = new_cp;
	new_cp->prev = prev_cp;
	new_cp->next = next_cp;
	return;
      }
      next_cp = prev_cp;
      prev_cp = prev_cp->prev;
    }
  }

  // otherwise or if this fails, stick it in the back
  new_cp->prev = elist->last_cp;
  elist->last_cp->next = new_cp;
  elist->last_cp = new_cp;
  
}/*--append_cl_ext()--*/



void sos_scan_insert(struct clause *c)
/*-----------------------------------------------------------------------*
 | Called by scott_insert() adn init_clause_space().                     |
 |                                                                       |
 | Insert the given clause into the appropriate external lists.  New     |
 | clauses are placed at the start of the list so that the scan order is |
 | always changing.                                                      |
 *-----------------------------------------------------------------------*/
{
  int maxiset;
  
  if (!develop_semantics) return; // this structure is no longer needed
  
  prepend_cl_ext(c,&sos_scan[c->num_sets]);
  
  // insert in failed_clauses[] too!
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset] && !c->in_set[maxiset]) 
      prepend_cl_ext(c,&failed_clauses[maxiset]);
  
}/*--sos_scan_insert()--*/


void sos_scan_delete(struct clause *c)
/*--------------------------------------------------------------------*
 | Called by scott_delete().                                          |
 |                                                                    |
 | Removes the given clause from any external lists that contain it.  |
 *--------------------------------------------------------------------*/
{
  int maxiset;

  if (!develop_semantics) return; // this structure is no longer needed
  
  delete_cl_ext(c,&sos_scan[c->num_sets]);

  // remove from failed_clauses[] too!
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset] && !c->in_set[maxiset]) 
      delete_cl_ext(c,&failed_clauses[maxiset]);

}/*--sos_scan_delete()--*/


void declare_soft_equality()  
/*----------------------------------------------------------------------*
 | Called by init_clause_space().                                       |
 |                                                                      |
 | Something experimental.  adjust the base maxiset to force either     |
 | anti-symmetric or non-symmetric.  This assumes that we are using a   |
 | special equality definition that does not contain the symmetry axiom |
 | for equality.                                                        |
 *----------------------------------------------------------------------*/
{
  //---- anti-symmetric equality
  // model("equal(A,B), equal(B,A) -> $false.", 0, false, true);
  
  //---- non-symmetric equality
  add_declaration("constant scott_a default.");
  add_declaration("constant scott_b default.");
  model("equal(scott_a,scott_b), equal(scott_b,scott_a) -> $false.", 
	0, false, true);
  
}/*--declare_soft_equality()--*/


void declare_hard_equality()  
/*-----------------------------------------------------------------------.
| Called by init_clause_space().  Insist that all models of equality are |
| the identity reklation                                                 |
`-----------------------------------------------------------------------*/
{
  init_theory(0); // hack to get rid of any existing equality definition
  model("equal(A,B) = (A = B).",0,false, true);

}/*--declare_hard_equality()--*/


void init_clause_space()
/*----------------------------------------------------------------------.
| Called by init_semantics() and read_all_input().                      |
|                                                                       |
| Initialises external clause lists sos_scan[], failed_clauses[],       |
| semantic_base, and conjecture.  Since FINDER was not active during    |
| input processing, also ensures clause lengths have been recorded.     |
|                                                                       |
| This must come BEFORE auto_set_conjecture() and auto_set_sos(), since |
| these require clauses to have been integrated into there respective   |
| lists.  This will also be called AFTER incase any list have changed.  |
`----------------------------------------------------------------------*/
{
  int maxiset;
  struct clause *cl;

  init_clause_dptr();

  for (maxiset = 0; maxiset <= last_set_index; maxiset++){

    sos_scan[maxiset].first_cp = 
      sos_scan[maxiset].last_cp = 
      failed_clauses[maxiset].first_cp = 
      failed_clauses[maxiset].last_cp = NULL;
  }

  // set weights of initial clause space
  for (cl = Usable->first_cl; cl; cl=cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }
  
  for (cl=Sos->first_cl; cl; cl=cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
    sos_scan_insert(cl); // initial scan
  }

  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }

}/*--init_clause_space() ---------------------------------------------*/


void init_conjecture_list()

/*-------------------------------------------------------.
| Called by init_semantics().  Setup an initial list of  |
| pointers to the default conjecture clauses.  Then call |
| check_conjecture_list() to ensure conjecture is easliy |
| satisfiable.                                           |
`-------------------------------------------------------*/

{
  struct clause *cl, *next;
  boolean first;
  int (*conj_clause_type)() = NULL;
    
  message(2,"");
  message(2,"------------> set up conjecture lists:");
  message(2,"");

  if (tFlags[SEMANTIC_GUIDANCE].val && tFlags[AUTO_SET_CONJECTURE].val) {

    conj_clause_type = auto_set_conjecture();
  
    /*-----------------------------------------------------------.
    | Conditionally move these clasues to the INPUT_FIRST list.  |
    `-----------------------------------------------------------*/

    if (conj_clause_type && tFlags[CONJ_IN_INPUT_FIRST].val) {

      first = true;
      cl=Sos->first_cl; 
      while (cl) {
	next = cl->next_cl;
	if ((*conj_clause_type)(cl)) {
	  if (first) {
	    message(2,"");
	    message(2,
		    "====> Moving the following conjecture clauses "
		    "to the INPUT_FIRST list:");
	    message(2,"");
	    first = false;
	  }
	  if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
	  rem_from_list(cl);
	  append_cl(First,cl);
	}
	cl = next;
      }
      
      first = true;
      cl=First->first_cl; 
      while (cl) {
	next = cl->next_cl;
	if (!(*conj_clause_type)(cl)) {
	  if (first) {
	    message(2,"");
	    message(2,
		    "====> Moving the following non-conjecture clauses "
		    "to the SOS list:");
	    message(2,"");
	    first = false;
	  }
	  if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
	  rem_from_list(cl);
	  append_cl(Sos,cl);
	}
	cl = next;
      }
    }
  }
  else {

    /*--------------------------------------------------------.
    | Otherwise we regard all initial INPUT_FIRST clauses are |
    | then conjectures.                                       |
    `--------------------------------------------------------*/

    first = true;
    cl=First->first_cl; 
    while (cl) {
      next = cl->next_cl;
      if (first) {
	message(2,"");
	message(2,"Adding the following INPUT_FIRST clauses to the conjecture list:");
	message(2,"");
	first = false;
      }
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      cl->stype = Conj;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      tStats[CONJECTURE_SIZE]++;
      cl = next;
    }
  }

  if (tFlags[DEMOD_IN_SEM_BASE].val) {
    first = true;
    for (cl=Demodulators->first_cl; cl; cl = cl->next_cl) {
      if (first) {
	message(2,"");
	message(2,"Adding the following DEMODULATORS to the semantic base:");
	message(2,"");
	first = false;
      }
      cl->stype = Base;
      append_cl_ext(cl,&semantic_base); 
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      tStats[SEM_BASE_SIZE]++;
    }
  }
  else if (tFlags[DEMOD_IN_CONJECTURE].val) {
    first = true;
    for (cl=Demodulators->first_cl; cl; cl = cl->next_cl) {
      if (first) {
	message(2,"");
	message(2,"Adding the following DEMODULATORS to the conjecture list:");
	message(2,"");
	first = false;
      }
      cl->stype = Conj;// special code for conj    
      append_cl_ext(cl,&conjecture); 
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      tStats[CONJECTURE_SIZE]++;
    }
  }

  if (!conjecture.first_cp || !conjecture.first_cp->c) 
    message(1,"WARNING: no conjecture clause has been established.\n");
  
  check_conjecture_list();

}/*--init_conjecture_list()--*/

void init_semantic_base()
  /*------------------------------------------------.
  | Called by init_semantics() after auto_set_sos() |
  `------------------------------------------------*/
{
  struct clause *cl;
  boolean first;

  if (tFlags[USABLE_IN_SEM_BASE].val) {
    first = true;
    for (cl=Usable->first_cl; cl; cl = cl->next_cl) {
//       if (first) {
// 	message(2,"");
// 	message(2,"Adding the following USABLE clauses to the semantic base:");
// 	message(2,"");
// 	first = false;
//       }
      cl->stype = Base;
      append_cl_ext(cl,&semantic_base); 
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      tStats[SEM_BASE_SIZE]++;
    }
  }
}

void check_conjecture_list()
/*-------------------------------------------------------------.
|            Called by init_semantics().                       |
|                                                              |
| Check that the conjecture can be easily model by FINDER.  If |
| not, this conjecture list will probally cause the search to  |
| be guided by poor assessment of satisfiablilty.              |
`-------------------------------------------------------------*/
{
  struct clause *cl;
  struct clause_dptr *cp, *next_cp;
  char finder_cl[CLAUSE_STRING_ARRAY];
  boolean first = true;

  // Use maxiset position 1 for this testing
  // and set maximum time limit for modelling input clauses
  set_time_limit(tParms[MAX_SEARCH_TIME].val);

  // model each conjecture clause
  cp = conjecture.first_cp; 
  while (cp){
    next_cp = cp->next;
    cl = cp->c;
    if (!make_finder_clause(finder_cl, cl)) {
      disable_semantics("Clause is too large to model");
      return;
    }
    if (!model(finder_cl, 1, false, true)){
      if (first) {
	message(2,"\nWARNING:List of conjecture clauses cannot be modelled.");
	first = false;
      }
      message(2,"Removing clause %d from list of conjecture clauses.", 
	      cl->id);
      if (tParms[VERBOSITY].val >=2) print_clause(stdout,cl);
      delete_cl_ext(cl,&conjecture);
      cl->stype = Norm;
      tStats[CONJECTURE_SIZE]--;
    }
    cp = next_cp;
  }
  
  // reinit tempory maxiset position 1
  init_sizes(1);
  init_theory(1);
  
  message(2,"");

}/*--check_conjecture_list()--*/

boolean make_finder_clause(char *finder_clause, struct clause *otter_clause)
/*-----------------------------------------------------------------------*
 | Called by model_list() and add_semantic_wt().                         |
 |                                                                       |
 | Given an OTTER clause, creates a corresponding FINDER clause.  Return |
 | TRUE iff the clause is successfully converted.                        |
 *-----------------------------------------------------------------------*/
{
  struct literal *lit;
  int separator, num_vars;

  *finder_clause = '\0'; /* start with an empty FINDER clause */
  
  num_vars = distinct_vars(otter_clause);
  if (num_vars >= SCOTT_MAX_VARS) {
    // assume the clause is shorter than the current limits
    message(0,"WARNING: clause has too many (%d) variables", 
	    num_vars);
    message(0,"Resetting distinct variable limit to %d\n",
	    num_vars-1);    
    Parms[MAX_DISTINCT_VARS].val = num_vars-1;
    cull_clauses_over(num_vars-1,distinct_vars);
    return false;
  }

  /* convert negative literals */
  for (lit=otter_clause->first_lit,separator=false;lit;lit=lit->next_lit)
    if (!lit->sign) {
      if (separator)
	strcat(finder_clause, ", ");
      if (!add_finder_term(finder_clause, lit->atom)) {
	// assume the clause is shorter than the current limits
	message(0,"WARNING: clause is too long (%d characters)",
		CLAUSE_STRING_ARRAY);
	message(0,"Resetting weight limit to %d\n",otter_clause-1);    
	Parms[MAX_WEIGHT].val = otter_clause->pick_weight-1;
	cull_clauses_over(otter_clause->pick_weight-1,clause_pick_weight);
	return false;
      }
      else separator = true;      
    }
  
  strcat(finder_clause, " -> ");
  
  /* convert positive literals */
  for (lit=otter_clause->first_lit,separator=false;lit;lit=lit->next_lit)
    if (lit->sign) {
      if (separator)
	strcat(finder_clause, ", ");
      if (!add_finder_term(finder_clause, lit->atom)) {
	// assume the clause is shorter than the current limits
	message(0,"WARNING: clause is too long (%d characters)",
		CLAUSE_STRING_ARRAY);
	message(0,"Resetting weight limit to %d\n",otter_clause-1);    
	Parms[MAX_WEIGHT].val = otter_clause->pick_weight-1;
	cull_clauses_over(otter_clause->pick_weight-1,clause_pick_weight);
	return false;
      }
      else separator = true;
    }
  strcat(finder_clause, ".");

  return true;
}/*--make_finder_clause()--*/


boolean add_finder_term(char *finder_clause, struct term *otter_term)
/*---------------------------------------------------------------------*
 | Called by make_finder_clause().                                     |
 |                                                                     |
 | Converts an OTTER term into a FINDER term and appends to a partial  |
 | finder_clause.  Recursively checks that the clause hasn't grown too |
 | large.  Return TRUE iff the clause is successfully converted.       |
 *---------------------------------------------------------------------*/
{
  struct rel *r;
  char variable_index[VARMAX];
  static boolean clause_too_big = false;

  if (clause_too_big) return false;

  /* exit if s is too big to append the next name to */
  if (strlen(finder_clause) >= (CLAUSE_STRING_ARRAY - FUNCTION_NAME_ARRAY)){
    clause_too_big =true;
    return false;
  }
  if (otter_term->type == NAME){
    strcat(finder_clause, sn_to_str(otter_term->sym_num));
  }
  else if (otter_term->type == VARIABLE) {
    strcat(finder_clause, "V");
    // now we need to turn the number into a string
    sprintf(variable_index, "%d", otter_term->varnum);
    strcat(finder_clause, variable_index);
  }
  else {
    strcat(finder_clause, sn_to_str(otter_term->sym_num));
    strcat(finder_clause, "(");
    for (r = otter_term->farg; r; r = r->narg) {
      if(!add_finder_term(finder_clause, r->argval)) return false ;
      if (r->narg)
	strcat(finder_clause, ",");
    }
    strcat(finder_clause, ")");
  }
  return true;

}/*--add_finder_term()--*/


void in_set_to_str(char *in_set, struct clause *cl)
/*----------------------------------------------------------------------*
 | Called by extend_base(), print_clause() and print_in_set().          |
 |                                                                      |
 | Copies the "in_set" field of "cl" across to the string pointer to by |
 | "in_set", transferring appropriate character codes.                  |
 *----------------------------------------------------------------------*/
{
    int maxiset, count;

    *in_set = '(';
    in_set++;
    for (maxiset = 0, count= 0; 
	 maxiset <= last_set_index && count <= num_maxisets; 
	 maxiset++, in_set++) {
      if (!active[maxiset]) {
	*in_set = 'x';
	continue;
      }
      else if (cl->in_set[maxiset]) {
	if (cl->stype == Base) *in_set = 'b';
	else if (cl->stype == Old_Base) *in_set = 'o';
	else if (cl->stype == New_Base) *in_set = 'n';
	else if (cl->stype == Conj) *in_set = 'c';
	else *in_set = '1';
      }
      else *in_set = '0';
      count++;
    }
    *in_set = ')';
    
    if (tFlags[SEM_RES].val) {
      in_set++; *in_set = ' '; in_set++;
      if (cl->in_set[constraint_set]) *in_set = 'T';
      else *in_set = 'F';
    }
    
    *(in_set+1)  = '\0';

}/*--in_set_to_str()--*/


boolean in_set_identical(struct clause *c1, struct clause *c2)
/*----------------------------------------------------------------------*
 | Called by shuffle_ext_list().  Returns true iff in_set fileds of the |
 | two clauses are identical.                                           |
 *----------------------------------------------------------------------*/
{
  int maxiset;

  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
    if (c1->in_set[maxiset] != c2->in_set[maxiset]) 
      return false;
  }
  return true;

}/*in_set_identical()--*/


/*================== DEBUGGING TOOLS =================================*/

void print_sos_scan()
{
  int list;
  
  for (list = 0; list <= num_maxisets; list++) {
    printf("\n%d | ",list);
    print_ext_list(&sos_scan[list]);
  }
  printf("\n\n");
  fflush(stdout);
  
} /* print_sos_scan() ------------------------------------------------*/


void print_failed_clause_nos()
{
  int num_sets;
  
  for (num_sets = 0; num_sets<=num_maxisets; num_sets++){
    printf("\n%d | ",num_sets);
    print_ext_list(&failed_clauses[num_sets]);
  }
  printf("\n\n");
  fflush(stdout);
  
} /* print_failed_clauses() ------------------------------------------*/


void print_conjecture_nos()
{    
  struct clause_dptr *cp;

  for (cp = conjecture.first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
  printf("\n");
  fflush(stdout);
} 

void print_base_nos()
{    
  struct clause_dptr *cp;

  for (cp = semantic_base.first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
  printf("\n");
  fflush(stdout);
}

void print_ext_list(struct external_list *elist)
{
  struct clause_dptr *cp;

  for (cp = elist->first_cp; cp && cp->c ; cp = cp->next){
    printf("%d ",cp->c->id);
  }
} /* print_ext_list() ------------------------------------------------*/ 

void debug_scan(char *event)

{
  int num_sets, maxiset;
  struct clause_dptr *cp;

  if (!develop_semantics) return;

  for (num_sets = 0; num_sets <= num_maxisets; num_sets++){
    for (cp = sos_scan[num_sets].first_cp; cp; cp = cp->next) {
      if (cp->c->num_sets != num_sets) 
	abort_scott("clause %d gets misplaced in sos_scan[]", cp->c->id);
    }
  }
  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
    for (cp = failed_clauses[maxiset].first_cp; cp; cp = cp->next) {
      if (cp->c->in_set[maxiset]) 
	abort_scott("clause %d gets misplaced in failed_clauses[]", cp->c->id);
    }
  }
  
} /*check_scan() -----------------------------------------------------*/


void print_sos()
{  
  struct clause *cl;

  printf("---->sos clauses\n");
  for (cl=Sos->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d id=", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }
} /* print_sos() -----------------------------------------------------*/

void print_demod()
{  
  struct clause *cl;

  printf("---->Demodulators\n");
  for (cl=Demodulators->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d id=", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }
} /* print_sos() -----------------------------------------------------*/

void print_base()
{    
  struct clause_dptr *cp;

  printf("---->base clauses\n");
  for (cp = semantic_base.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
}

void print_new_base()
{    
  struct clause_dptr *cp;

  printf("---->new base clauses\n");
  for (cp = new_base_clauses.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
}

void print_conjecture()
{    
  struct clause_dptr *cp;

  printf("---->conjecture clauses\n");
  for (cp = conjecture.first_cp; cp && cp->c ; cp = cp->next){
    printf("wt=%2d, m=%d id=", cp->c->pick_weight, cp->c->num_sets);
    print_clause(stdout,cp->c);
  }
  fflush(stdout);
} 

void print_usable()

{
  struct clause *cl;

  printf("---->usable clauses\n"); 
  for (cl=Usable->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d id=", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }

} /* print_usable() --------------------------------------------------*/

void print_first()

{
  struct clause *cl;

  printf("---->input_first clauses\n"); 
  for (cl=First->first_cl; cl; cl = cl->next_cl) {
    printf("wt=%2d, m=%d ", cl->pick_weight, cl->num_sets);
    print_clause(stdout,cl);
  }

} /* print_first() ---------------------------------------------------*/

