   /*---------------------------------------------------------.
   | FILE: cost.c: rotines to manage cost of semntic guidance |
   `---------------------------------------------------------*/

#include "scott.h"

static struct data_array cost_of_given;          // in ms
static struct data_array cost_per_failed_clause; // in ms
static struct data_array time_saved;             // in ms
static struct data_array change_in_time_saved;   // in ms
static struct data_array prob_of_success;      // fraction
static struct data_array prop_failed_clauses;   // fraction

void init_costs()
  /*------------------------------------------------------------------.
  | Called by init_semantics(). Initialise data structures associated |
  | with this file.                                                   |
  `------------------------------------------------------------------*/
{
  int point;

  strcpy(cost_of_given.name,"cost_of_given");
  cost_of_given.num_points = 0;
  cost_of_given.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    cost_of_given.list[point] = 0; 
  
  strcpy(cost_per_failed_clause.name,"cost_per_failed_clause");
  cost_per_failed_clause.num_points = 0;
  cost_per_failed_clause.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    cost_per_failed_clause.list[point] = 0;

  strcpy(prop_failed_clauses.name,"prop_failed_clauses");
  prop_failed_clauses.num_points = 0;
  prop_failed_clauses.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    prop_failed_clauses.list[point] = 0;

  strcpy(time_saved.name,"time_saved");
  time_saved.num_points = 0;
  time_saved.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    time_saved.list[point] = 0;

  strcpy(change_in_time_saved.name,"change_in_time_saved");
  change_in_time_saved.num_points = 0;
  change_in_time_saved.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    change_in_time_saved.list[point] = 0;
  
  strcpy(prob_of_success.name,"prob_of_success");
  prob_of_success.num_points = 0;
  prob_of_success.total = 0;
  for (point=0; point < NUM_POINTS; point++)
    prob_of_success.list[point] = 0;
  
}/*init_costs()--*/

void reset_prob_of_success()
  /*----------------------------------------------------------------------.
  | Called by execute_strategy().                                         |
  |                                                                       |
  | Search time limit has been increased so discard previous estimates of |
  | extention success.  Keep the maximum value.                           |
  `----------------------------------------------------------------------*/
{
  int point;
  float maximum = 0;

  prob_of_success.num_points = 1;
  prob_of_success.total = 0;
  for (point=0; point < NUM_POINTS; point++) {
    if (prob_of_success.list[point] > maximum) 
      maximum = prob_of_success.list[point];
    prob_of_success.list[point] = 0;
  }
  prob_of_success.num_points = 1;
  prob_of_success.total = maximum;
  prob_of_success.list[0] = maximum;
}

void set_data_array(struct data_array *array, float value)  
  /*----------------------------------------------------------------------.
  | Called by various functions                                           |
  |                                                                       |
  | Append "value" to a finite list of data points.  If we have reach our |
  | limit on data point (NUM_POINTS), pop off the first value in the list |
  | before appending the new value.  Note we assume all values are        |
  | positive integers.                                                    |
  `----------------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) {
    array->list[array->num_points] = value;
  }
  else { // shuffle along
    for (point = 1; point < NUM_POINTS; point++)
      array->list[point - 1] = array->list[point];
    array->list[point - 1] = value;
  }
  
  array->num_points++;
  array->total += value;
  
}/*set_data_array()--*/


void set_cost_of_given(float value)
  /*----------------------------------------------------------.
  | Called by semantic_selection().  Set next cost_of_given.  |
  `----------------------------------------------------------*/
{
  set_data_array(&cost_of_given, value);

}/*set_cost_of_given()--*/


void set_cost_per_failed_clause(float value)
  /*-----------------------------------------------------------------.
  | Called by execute_strategy().  Set next cost_per_failed_clause.  |
  `-----------------------------------------------------------------*/
{
  set_data_array(&cost_per_failed_clause, value);

}/*set_cost_per_failed_clause()--*/

void set_prob_of_success(float value)
  /*----------------------------------------------------------.
  | Called by execute_strategy().  Set next "prob_of_success" |
  `----------------------------------------------------------*/
{
  set_data_array(&prob_of_success,value);
  
}/*set_prob_of_success()--*/

void set_prop_failed_clauses(float value)
  /*--------------------------------------------------------------.
  | Called by execute_strategy().  Set next "prop_failed_clauses" |
  `--------------------------------------------------------------*/
{
  set_data_array(&prop_failed_clauses,value);
  
}/*set_prob_of_success()--*/

void set_time_saved(int givens_avoided)
  /*------------------------------------------------------------------.
  | Called by semantic_selection().  Set the next "time saved" and if |
  | that array is full, set the next "change_in_time_saved".          |
  `------------------------------------------------------------------*/
{
  static int last_givens_avoided = 0;
  
  if (givens_avoided){
    float value;
    
    value = 
      est_from_data_array(&cost_of_given)*
      (givens_avoided - last_givens_avoided); //*
    //      rate_partition(); // adjust for confidence

    //    message(3, "time_saved = %.0f", value);
    set_data_array(&time_saved,value);
    
    if (time_saved.num_points >= NUM_POINTS) // list is full
      set_change_in_time_saved();
  }
  last_givens_avoided = givens_avoided;

}/*set_time_saved()--*/


void set_change_in_time_saved()
  /*--------------------------------------------------------------.
  | Called by set_time_saved()                                    |
  |                                                               |
  | If a new maxiset has been kept since the last call, reset the |
  | estimated change in time_saving accordingly.                  |
  `--------------------------------------------------------------*/
{
  static int last_kept = 0;
  static int last_time_saved = 0;
  int this_time_saved;

  if (tStats[KEPT] > last_kept) {

    this_time_saved = est_from_data_array(&time_saved);

    set_data_array(&change_in_time_saved, 
		   (this_time_saved - last_time_saved));

    last_time_saved = this_time_saved;
    last_kept = tStats[KEPT];
  }

}/*set_change_in_time_saved()--*/

void data_list_to_str(char *array_str, struct data_array *array)
  /*--------------------------------------------------------------------.
  | Called by persist().                                                |
  |                                                                     |
  | Convert the "list" of the given "array" to a comma delimited string |
  | of its values incased in square brackets.  Places the result in the |
  | string pointed to by "array_str".                                   |
  `--------------------------------------------------------------------*/

{
  int point;
  char tmp[10];

  *array_str = *tmp = '\0';

  if (array->num_points == 0) return;

  snprintf(tmp,10,"[%.2f",array->list[0]); 
  strncat(array_str,tmp,100);
  
  for (point = 0; 
       point < array->num_points && point < NUM_POINTS;
       point++){
    snprintf(tmp,10,", %.3f",array->list[point]); 
    strncat(array_str,tmp,100);
  }
  strncat(array_str,"]",100);
  
}/*data_list_to_str()--*/


float est_from_data_array(struct data_array *array)
  /*-------------------------------------------------------------.
  | Called by various functions.  Return average from data array |
  `-------------------------------------------------------------*/
{  
  float average, estimate;

  if (array->num_points == 0) 
    abort_scott("Attempting to estimate from data array \"%s\" "
		"which contains no data",array->name);
  if (array->num_points < 2) {
    char list_str[100];

    data_list_to_str(list_str,array);
    message(2,"");
    message(2,"WARNING: estimating from partial data array"
	    "\n         --> \"%s\" has only %d data points",
	    array->name,array->num_points);
    message(2,"");
  }

  average = (array->total/(float)array->num_points);  

  if (array->num_points < NUM_POINTS)
    estimate = (array->list[array->num_points-1] + gradient(array));
  else 
    estimate = (array->list[NUM_POINTS-1] + gradient(array));
    
  return (average + estimate)/2.0;

}/*est_from_data_array()--*/


float est_cost_per_failed_clause()
{
  return (est_from_data_array(&cost_per_failed_clause));
}

float est_prob_of_success()
{
  return (est_from_data_array(&prob_of_success));
}

float est_change_in_time_saved()
{
  return (est_from_data_array(&change_in_time_saved));
}

float gradient(struct data_array *array)
  /*---------------------------------------------------------------------.
  | Called by persist().  Returns least squares estimate of the gradient |
  | of a line of best fit.                                               |
  `---------------------------------------------------------------------*/
{
  int n;
  float x, y, sum_y, sum_x, sum_xx, sum_xy;
  
  x = y = n = 0;
  sum_x = sum_y = sum_xy= sum_xx = 0;
  
  while (n < array->num_points 
	 && n < NUM_POINTS ){
    x = n;
    y = array->list[n];
    sum_y += y;
    sum_x += x;
    sum_xy += (x*y);
    sum_xx += (x*x);
    n++;
  }
  
  return (((sum_x*sum_y)-(n*sum_xy)) /
	  ((sum_x*sum_x)-(n*sum_xx))  );
  
}/*gradient()--*/

float intercept(struct data_array *array)
  /*----------------------------------------------------------------------.
  | Called by persist().  Returns least squares estimate of the intercept |
  | of a line of best fit.                                                |
  `----------------------------------------------------------------------*/
{
  int n;
  float x, y, sum_y, sum_x, sum_xx, sum_xy;
  
  x = y = n = 0;
  sum_x = sum_y = sum_xy= sum_xx = 0;
  
  while (n < array->num_points
	 && n < NUM_POINTS){
    x = n;
    y = array->list[n];
    sum_y += y;
    sum_x += x;
    sum_xy += (x*y);
    sum_xx += (x*x);
    n++;
  }
  
  return (((sum_x*sum_xy)-(sum_xx*sum_y)) /
	  ((sum_x*sum_x)-(n*sum_xx))	  );
  
}/*intercept()--*/


boolean strictly_increasing(struct data_array *array)
  /*--------------------------------------------------------------------.
  | Called by persist().  Returns TRUE iff values in the data array are |
  | strictly increasing.                                                |
  `--------------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] >= array->list[point]) return false;
  }

  return true;

}/*strictly_increasing()--*/


boolean strictly_decreasing(struct data_array *array)
  /*--------------------------------------------------------------------.
  | Called by persist().  Returns TRUE iff values in the data array are |
  | strictly decreasing.                                                |
  `--------------------------------------------------------------------*/
{
  int point;
  
  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] <= array->list[point]) return false;
  }

  return true;

}/*strictly_decreasing()--*/

boolean decreasing(struct data_array *array)
  /*--------------------------------------------------------------------.
  | Called by persist().  Returns TRUE iff values in the data array are |
  | strictly increasing.                                                |
  `--------------------------------------------------------------------*/
{
  int point;

  if (array->num_points < NUM_POINTS) return false;
  
  for (point = 1; point < NUM_POINTS; point++){
    if (array->list[point - 1] < array->list[point]) return false;
  }

  return true;

}/*decreasing()--*/

void end_development(char *event);

void end_development(char *event)
{
  // check these at least once!!
  debug_partition(event); 
  debug_scan(event); 
  check_null_set_size(event);
  
  message(2,"");
  message(2,"\n%s",event);
  message(2,"");
  print_partition("at end of development");
  tStats[MAXISETS_AT_END_DEV] = num_maxisets;
  develop_semantics = recording = false;
}

boolean persist()
  /*-------------------------------------------------------------------.
  | Called by execute_strategy()                                       |
  |                                                                    |
  | Determined whether the cost of attempting to form a new maxiset is |
  | likely to out-weigh the expected return in terms of guidance.      |
  `-------------------------------------------------------------------*/
{
  int cost_of_attempt, time_left; // in millisecs
  //  int  accrued_time_saving; // in millisecs
  //  float expected_return;
  float prob;
  char array_str[100];

  if (!develop_semantics) return false;

  if (First->first_cl) return false; // clause in semantic queue.
  if (num_maxisets == tParms[MAX_NUM_MAXISETS].val) return false;

  // SOS is smaller than can be easily split by a new maxiset
  if (num_maxisets * 2 > Stats[SOS_SIZE]) return false;

  if (tParms[SELECTION_BUFFER].val < Stats[SOS_SIZE])  return false;

  /*--------------------------------------------------------------------.
  | cost depends largely on the number of failed clauses so we estimate |
  | the likely number of failed clauses in this attempt and assess the  |
  | probable cost relative to this.                                     |
  `--------------------------------------------------------------------*/

  // this array must be contain something for the remaining tests
  if (prop_failed_clauses.num_points < 3) return true;

  cost_of_attempt = (int)(est_from_data_array(&prop_failed_clauses) *
			  (Stats[SOS_SIZE] + Stats[USABLE_SIZE]) *
			  get_time_limit());

  if (Parms[MAX_SECONDS].val > 0)
    time_left = (Parms[MAX_SECONDS].val*1000) - run_time();
  else 
    time_left = MAX_INT;
  
  // buffer to ensure at least a few maxisets have been kept
  if (NUM_POINTS > tStats[ATTEMPTS] 
      && cost_of_attempt*fParms[COST_MARGIN].val < time_left) {
    return true;
  }
  
  message(3,"");
  message(3,"%-30s %4.0f secs","Estimated cost of an attempt",
	  cost_of_attempt/(float)1000);
  message(3,"%-30s %4.0f secs", "Time remaining",
	  time_left/(float)1000);

  if (cost_of_attempt*fParms[COST_MARGIN].val > time_left){
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_COST_VS_TIME]++;
    record_event("<time limit>");
    end_development("Too close to time_limit");
    return false;
  }

  // this array must be full before the remaining tests
  if (prob_of_success.num_points < NUM_POINTS) return true;
  
  *array_str = '\0';
  data_list_to_str(array_str,&prob_of_success);
  
  // pre-emptive strike
  if (get_time_limit() >= tParms[MAX_SEARCH_TIME].val
      && strictly_decreasing(&prob_of_success)) {

    /*--------------------------------------------------------------------.
    | Note this uses values that are different from the estimates printed |
    | by persist() which are 5 point averages.                            |
    `--------------------------------------------------------------------*/

    char tmp[130];
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_PROB_DEC]++;
    record_event("<prob decr>");
    sprintf(tmp,"Probability that attempt will succeed is "
	    "strictly decreasing\n---->%s", array_str);
    end_development(tmp);
    return false;
  }
   
  prob = est_from_data_array(&prob_of_success);
  message(3,"%-30s    %.4f", "Probability of success",prob);
  
  if (cost_of_attempt > time_left*prob) {
    tStats[FAILS_IN_PERSIST]++;
    tStats[PFAIL_COST_VS_PROB]++;
    record_event("<prob too low>");
    end_development("prob too low");
    return false;
  }
  
//   if (num_maxisets < tParms[MIN_NUM_MAXISETS].val) return true;
// 
//   // this array must be full before the remaining tests
//   if (time_saved.num_points < NUM_POINTS) return true;
//   
//   // semantic is'nt helping: pre-emptive strike
//   if (strictly_decreasing(&time_saved)) {
//     tStats[FAILS_IN_PERSIST]++;
//     tStats[PFAIL_TIME_DEC]++;
//     record_event("<t-dec>");
//     end_development("Timed saved is strictly decreasing");
//     return false;
//   }
//   
//   // have we accumulated enough time savings for a new attempt?
//   accrued_time_saving = time_saved.total-scott_clock_val(SEM_GUIDE_TIME);
// 
//   if (accrued_time_saving < cost_of_attempt*fParms[COST_MARGIN].val){
//     tStats[FAILS_IN_PERSIST]++;
//     tStats[PFAIL_TIME_ACCRUED]++;
//     record_event("<t-acr>");
//     message(3,"%-30s %4.0f secs", "Accrued time saving",
// 	    accrued_time_saving/(float)1000);
//     return false;    
//   }
//   
//   // this array must be full before the remaining tests
//   if (change_in_time_saved.num_points < NUM_POINTS) return true;
//   
//   expected_return =  est_from_data_array(&change_in_time_saved)*prob;
//   
//   if (cost_of_attempt*fParms[COST_MARGIN].val > expected_return) {
//     tStats[FAILS_IN_PERSIST]++;
//     tStats[PFAIL_COST_VS_RETURN]++;
//     message(3,"%-30s %4.0f secs","Expected return",
// 	    expected_return/(float)1000);
//     record_event("<return too low>");
//     end_development("Expected return too small");
//     return false;
//   }  

  return true;
  
}/*--persist()--*/


