/*=========================================================\
|                 FILE: Spartition.c                       |
|                                                          |
| SCOTT functions for maintaining partition in formation.  |
\=========================================================*/

#include "scott.h"
#include <stdio.h>
#include <stdarg.h>

static struct partition set_in_cell; // partition by NMCSs
static struct partition cmp_in_cell; // partition by co-NMCSs

/*-------------------------------------------------------------.
| The following group of functions all return the value of the |
| partition structures that would otherwise be out of scope.   |
`-------------------------------------------------------------*/

int get_usable_partition(int maxiset) { return set_in_cell.usable[0][maxiset];}
int get_sos_partition(int maxiset)    { return set_in_cell.sos[0][maxiset];   }
int get_size_on_sos(int maxiset)      { return set_in_cell.sos[maxiset][0];   }
int get_comp_on_sos(int maxiset)      { return cmp_in_cell.sos[maxiset][0];   }
int get_comp_on_usable(int maxiset)   { return cmp_in_cell.usable[maxiset][0];   }
int get_size_on_usable(int maxiset)   { return set_in_cell.usable[maxiset][0];}

int get_comp_size(int maxiset)        
  /*--------------------------------------------------------.
  | Called by execute_strategy().  Used to set the value of |
  | cost_per_failed_clause                                  |
  `--------------------------------------------------------*/
{ 
  return (cmp_in_cell.usable[maxiset][0] + cmp_in_cell.sos[maxiset][0]);   
}

int num_failed_clauses() 
  /*-----------------------------------------------------------.
  | Called by check_failed_clauses(). Return the sum of the    |
  | sizes of the complements of all the active maxisets.  This |
  | should equal the number of pointers on the failed clause   |
  | lists.  Used to estimate the cost of an attempt            |
  `-----------------------------------------------------------*/
{
  int maxiset;
  int count = 0;

  for (maxiset = 1; maxiset <=last_set_index; maxiset++) {
    if (active[maxiset]) {
      // sum of the complements of each maxiset
      count += cmp_in_cell.sos[maxiset][0];
      count += cmp_in_cell.usable[maxiset][0];
    }
  }
  return count;
}

/*------------------------------------------------------------.
| The remaining functions in this file operate extensively on |
| this partition information.                                 |
`------------------------------------------------------------*/


void print_a_partition(struct partition *part_ptr)
  /*-----------------------------------------------------------.
  | Called by various print_partition().  Prints out the given |
  | partition data appropriately formated.                     |
  `-----------------------------------------------------------*/
{
  int maxiset, cell;
  char str[15];
  
  // print USABLE partition
  printf("\n U ");
  for (cell = 0; cell <= num_maxisets; cell++) 
    printf("%5d",part_ptr->usable[0][cell]);
  printf(" | %5ld\n", Stats[USABLE_SIZE]);
  
  printf("------------");
  for (cell = 0; cell <= num_maxisets; cell++) printf("-----");
  printf("\n");
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      printf("%2d:     ",maxiset);
      for (cell = 1; cell <= num_maxisets; cell++)
	printf("%5d",part_ptr->usable[maxiset][cell]);
      printf(" | %5d\n",part_ptr->usable[maxiset][0]);
    }
  }

  // print SOS partition
  printf("\n S "); 
  for (cell = 0; cell <= num_maxisets; cell++) 
    printf("%5d",part_ptr->sos[0][cell]);
  printf(" | %5ld     rating\n",Stats[SOS_SIZE]);

  printf("------------");
  for (cell = 0; cell <= num_maxisets; cell++) printf("-----");
  printf("\n");
  
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      printf("%2d:     ",maxiset);
      for (cell = 1; cell <= num_maxisets; cell++)
	printf("%5d",part_ptr->sos[maxiset][cell]);
      //printf(" | %5d\n",part_ptr->sos[maxiset][0]);

      sprintf(str,"%+.1f",rate_maxiset(maxiset));
      printf(" | %5d     %6s\n",part_ptr->sos[maxiset][0],str);
    }
  }
  
  fflush(stdout);

} /*--print_a_partition()--*/

void print_partition(char *event)   
  /*--------------------------------------------------------------------*
   | Called by various functions                                        |
   |                                                                    |
   | Prints a table describing the current state of the partition. The  |
   | string "event" indicates what imediately preceeded the call.  This |
   | also doubles as a usefull debugging tool.                          |
   *--------------------------------------------------------------------*/
{
  if (tParms[VERBOSITY].val <= 1) return;

  if (tParms[VERBOSITY].val >= 4){
    printf("\nMaxiset partititon %s\n", event); 
    print_a_partition(&set_in_cell);
  }
  
  printf("\nComplement partititon %s\n", event); 
  print_a_partition(&cmp_in_cell);
  printf("\n");
  message(2,"Average size of choice space: %0.1f", size_choice_space());
  message(2,"");
  
} /*--print_partition()--*/


boolean subset(int set1, int set2)
  /*---------------------------------------------------------*
   | Called by already_exists() and discard_subsets().       |
   |                                                         |
   | Detemines whether "set1" is a (weak) subset of "set2".  |
   *---------------------------------------------------------*/
{
  int cell;
  struct clause *cl;

  // compare partitions
  for (cell=0;cell <= last_set_index; cell++){
    if (set_in_cell.sos[set1][cell] > 
	set_in_cell.sos[set2][cell]) return false; 
    if (set_in_cell.usable[set1][cell] > 
	set_in_cell.usable[set2][cell]) return false; 
  }

  // search for a cl in set1 that is not in set2 
  for (cl = Sos->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl =First->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl = Usable->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  
  return true;

} /* subset()--*/

boolean sos_subset(int set1, int set2)
  /*---------------------------------------------------------------------*
   | Called by already_exists() and discard_subsets().                   |
   |                                                                     |
   | Detemines whether "set1" is a (weak) subset of "set2", but checking |
   | for disparity only on the SOS or INPUT_FIRST list.                  |
   *---------------------------------------------------------------------*/
{
  int cell;
  struct clause *cl;

  // compare partitions
  for (cell=0;cell <= last_set_index; cell++){
    if (set_in_cell.sos[set1][cell] > 
	set_in_cell.sos[set2][cell]) return false; 
  }

  // search for a cl in set1 that is not in set2 
  for (cl =Sos->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  for (cl =First->first_cl; cl; cl = cl->next_cl){
    if (cl->in_set[set1] && !cl->in_set[set2]) return false;   
  }
  
  return true;

} /* sos_subset()--*/

boolean already_exists(int new_maxiset) 
  /*----------------------------------------------------------------------*
   | Called by keep().                                                    |
   |                                                                      |
   | Returns TRUE iff new_maxiset is a subset of an existing maxiset --   |
   | this property is tested with respect to the SOS only.  Note: returns |
   | FALSE if the maxiset contains no SOS clauses (and hence is an SOS    |
   | subset of all other maxisets).                                       |
   *----------------------------------------------------------------------*/
{
  int old_maxiset;

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] && old_maxiset !=  new_maxiset) {
      if (subset(new_maxiset,old_maxiset) // real_subset
	  //if (sos_subset(new_maxiset,old_maxiset) &&
	  && get_size_on_sos(old_maxiset) < Stats[SOS_SIZE]) {
	
	message(2,"Maxiset %d is a subset of the maxiset %d", 
		new_maxiset, old_maxiset);
	return true;
      }
    }
  }
  
  return false;
  
} /*--already_exists()--*/

int superset_of(int new_maxiset) 
  /*--------------------------------------------------------------------*
   | Called by keep().                                                  |
   |                                                                    |
   | Returns the index of any maxiset that is a superset of the         |
   | new_maxiset.  Returns 0 if there is no such maxiset.  Since Usable |
   | list is not rigourously tested, set membership may be excessively  |
   | inacurrate, hence we only check the SOS and INPUT_FIRST lists.     |
   *--------------------------------------------------------------------*/
{
  int old_maxiset;

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] && old_maxiset !=  new_maxiset) {
      if (sos_subset(new_maxiset,old_maxiset)
	  && get_size_on_sos(old_maxiset) < Stats[SOS_SIZE]) {
	return old_maxiset;
      }
    }
  }
  
  return 0;
  
} /*--superset_of()--*/

boolean verify_model_in_list(int maxiset, struct list *lptr, char list_name[])
/*---------------------------------------------------------------------*
 | Called by check_for_model().                                        |
 |                                                                     |
 | Testing each clause in the given list against the model associated  |
 | with the given maxiset.                                             |
 |                                                                     |
 | Program exits with an error if an empty clause is passed as this is |
 | definately a bug.                                                   |
 *---------------------------------------------------------------------*/
{
  struct clause *cl;
  boolean result = true;
  char finder_cl[CLAUSE_STRING_ARRAY];

  for (cl = lptr->first_cl; cl ; cl = cl->next_cl) {
    if (!cl->first_lit) {
      abort_scott(0,"clause space contains an empty clause -- this is a bug");
    }
    if (!make_finder_clause(finder_cl, cl)) {
      message(0,"%s clause %d is too large to model", cl->id, list_name);
      result = false;
    }
    if (!model(finder_cl, maxiset, true, false)) {
      message(0,"%s clause %d fails in propossed model", 
	      list_name, cl->id);
      print_clause(stdout, cl);
      result = false;
    }
  }
  return result;

}/*--verify_model_in_list()--*/

void check_for_model(int new_maxiset)
  /*----------------------------------------------------------------------*
   | Called by execute_strategy().  Checks if a model of the whole clause |
   | space has been generated.c Either returns TRUE or does not return at |
   | all!!                                                                |
   *----------------------------------------------------------------------*/
{
  // check if we have a model of the whole clause space
  if (set_in_cell.sos[new_maxiset][0] == Stats[SOS_SIZE] &&
      set_in_cell.usable[new_maxiset][0] == Stats[USABLE_SIZE]) {
    
    record_event(VERIFY_MODEL);
    
    message(0,"Clause space appears to be consistent.\n"
	    "----> Verifying model");        
    if (!verify_model_in_list(new_maxiset,Usable,"usable") ||
	!verify_model_in_list(new_maxiset,Sos,"sos") ||
	!verify_model_in_list(new_maxiset,First,"input_first") ||
	!verify_model_in_list(new_maxiset,Demodulators,"demodulator")) {
#ifdef SCOTT_DEBUG
      // something really odd has happend
      debug_partition("Proposed model fails!!!");
      debug_scan("Proposed model fails!!!");
#endif
      abort_scott("<---- Proposed model fails!!!");
    }

    if (Stats[CL_WT_DELETE] > 0)
      message(0,"<---- Clause space rendered satisfiable by incomplete strategy");
    else {
      message(0,"<---- Model checks out!!!");
      output_the_model(new_maxiset);
    }
    cleanup();
    exit(POSSIBLE_MODEL_EXIT);
  }  
}/*--check_for_model()--*/
  
float size_choice_space()
  /*---------------------------------------------------------------------*
   | Called by rate_partition().                                         |
   |                                                                     |
   | Returns the average size of the current choice space.  This costing |
   | provides a measure of the quality of choice: smaller ratings are    |
   | better.                                                             |
   *---------------------------------------------------------------------*/
{
  int maxiset, whole;

  /* 
     average choice space of whole partition

       |C1|+|C2|+...+|Cn|+|M1^M2^...^Mn|
     = ---------------------------------
                     n+1
		    
     = whole/(n+1)

     Now |M1^M2^...^Mn| = |{cl:cl is in "n" maxisets}|
  */

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  return (whole/(float)(num_maxisets+1));
  
} /*--size_choice_space()--*/

float rate_partition()
  /*--------------------------------------------------------------------*
   | Called by keep() and refine_choice().                              |
   |                                                                    |
   | Returns the ratio of the choice space size with semantic selection |
   | versus the choice space size with lightest first selection.        |
   *--------------------------------------------------------------------*/
{
  return (size_choice_space()/(float)Stats[SOS_SIZE]);

} /*--rate_partition()--*/

float rate_maxiset(int skip_set)
  /*-----------------------------------------------------------------------*
   | Called by rate_clause() and keep().                                   |
   |                                                                       |
   | Returns the effect that keep "skip_set" has on the partition rating.  |
   | This costing provides a measure of the quality of choice: (larger)    |
   | negative ratings are better because the maxiset decreases the size    |
   | choice space most.                                                    |
   *-----------------------------------------------------------------------*/
{
  int maxiset;
  float whole, part = 0;
  float rating_with_skip_set, rating_without_skip_set;

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  part = whole - cmp_in_cell.sos[skip_set][0] 
    + set_in_cell.sos[skip_set][num_maxisets-1];
  
  rating_with_skip_set = (whole/(float)(num_maxisets+1));
  rating_without_skip_set = (part/(float)(num_maxisets));
  
  return (rating_with_skip_set - rating_without_skip_set);
  
} /*--rate_maxiset()--*/

//
float rate_clause(struct clause *cl)
  /*-----------------------------------------------------------------------*
   | Called by extend_base().                                              |
   |                                                                       |
   | Returns the expected change in the size of the choice space if all    |
   | the maxisets that the given clause does not belong to are discarded.  |
   |                                                                       |
   | Since we dont back cull in this way any more the rating is            |
   | meaningless                                                           |
   |                                                                       |
   | This costing provides a measure of the quality of choice.             |
   *-----------------------------------------------------------------------*/
{
  int maxiset, cmp, num_cmps;
  float whole, part = 0;
  float rating_with_skip_sets, rating_without_skip_sets;
  
  if (num_maxisets < 2) 
    abort_scott("rate_clause() called when num_maxisets < 2");

  whole = set_in_cell.sos[0][num_maxisets];
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) whole += cmp_in_cell.sos[maxiset][0];
  
  part = whole;
  num_cmps = num_maxisets - cl->num_sets;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++){
    if (active[maxiset] && !cl->in_set[maxiset]) {
      whole -= cmp_in_cell.sos[maxiset][0];
      for (cmp = 1; cmp <= num_cmps; cmp++) 
	whole += cmp_in_cell.sos[maxiset][num_maxisets-cmp];
    }
  }
  
  rating_with_skip_sets = (whole/(float)(num_maxisets+1));
  rating_without_skip_sets = (part/(float)(num_maxisets));
  
  return (rating_with_skip_sets - rating_without_skip_sets);
  
} /*--rate_clause()--*/


/*==================================================================\
| The following functions specify the principle semantic startegies |
\==================================================================*/

void set_sem_queue(struct list *lptr)
  /*----------------------------------------------------------.
  | Called by semantic_selection() and keep()                 |
  |                                                           |
  | If performing a semantic queue, moves appropriate clauses |
  | to the "input_fist" list. This function assumes the       |
  | INPUT_FIRST list is empty before it is called             |
  `----------------------------------------------------------*/
{
  struct clause *cl, *next_cl;
  int queue_limit, maxiset, count, wt_limit;
  boolean first = true;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) {
    // move all SOS to INPUT_FIRST
    record_event(SELECT_QUEUE);
    cl = lptr->first_cl;
    while (cl) {
      next_cl = cl->next_cl;
      rem_from_list(cl);
      append_cl(First, cl);
      cl = next_cl;
      tStats[CL_SEM_QUEUED]++;
    }
    message(2,"====> Entire SOS placed in the semantic queue");
    return;
  }
  
  wt_limit = weight_percentile(50);
  
  count = 0;
  queue_limit = tParms[SEM_QUEUE_LIMIT].val;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {

    if (num_maxisets > 1 && maxiset == last_queue_theory) continue;
    if (!active[maxiset] ) continue; 
    if (cmp_in_cell.sos[maxiset][1] == 0) continue; 
    if ((cmp_in_cell.sos[maxiset][1] + 
	 cmp_in_cell.usable[maxiset][1])
	> queue_limit) continue; 
    count++;
    
    cl = lptr->first_cl;
    while (cl) {
      next_cl = cl->next_cl;
      if (!cl->in_set[maxiset]   // clause is in the complement
	  && cl->num_sets == num_maxisets-1 // only one complement
	  && cl->pick_weight < wt_limit // avoid really big clauses
	  ) { 
	rem_from_list(cl);
	append_cl(First, cl);    // move to the input_first list
	tStats[CL_SEM_QUEUED]++;
	last_queue_theory = maxiset; // break cycles
	if (first) {
	  message(2,"");
	  message(2,"====> Following clauses placed in the semantic queue");
	  first = false;
	}
	if (tParms[VERBOSITY].val > 1) print_clause(stdout,cl);
      }
      cl = next_cl;
    }
  }
  if (count > 1) last_queue_theory = 0;

  if (!first) message(2,"");


} /*--set_sem_queue()--*/

int next_set_to_skip()
  /*---------------------------------------------------------------------*
   | Called by find_best_cl().                                           |
   |                                                                     |
   | The next given clause will belong to the complement of the next     |
   | skip_set.  If there is more than one maxiset, then skip set is set  |
   | to the next highest index, cycling through the availiable theories, |
   | and skipping empty theories.  If no clear skip set can be           |
   | established -1 is returned.  (This may never actually happen)       |
   *---------------------------------------------------------------------*/
{
  int skip_set, count;
  static int last_skip_set = -1;

  // dont cycle if we have semantic queue
  if (First->first_cl) return -1;
  
  skip_set = last_skip_set;
  for (count = 0; count <= last_set_index; count++) {
    
    if (skip_set == last_set_index) skip_set = 0;
    else skip_set++;
    
    if (skip_set == last_skip_set) break; 
    
    if (active[skip_set] 
	// &&
	//	size_on_sos[skip_set] > 0 &&
	//      size_on_sos[skip_set] < Stats[SOS_SIZE]
	// ...  these guys covered by previous call to discard_bad_maxisets()
	){
      last_skip_set = skip_set;
      
      return skip_set;
    }
  }
  
  last_skip_set = -1;
  return -1;

} /*--next_skip_set()--*/

int variance(struct clause *cl)
/*---------------------------------------------------------------------*
 |              Called by find_best_sem_cl().                          |
 |                                                                     |
 | Return the resulting variance of the distribution of complements on |
 | the USABLE list that would be expected if cl were introduced. Note: |
 | clauses with no models belong to every complement.                  |
 *---------------------------------------------------------------------*/
{
  int maxiset;
  float diff = 0;
  float average = 0;
  float variance = 0;
  
  // get the new distribution
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      average += (cmp_in_cell.usable[maxiset][0]
		  +(int)(!cl->in_set[maxiset]));
    }
  }
  average /= num_maxisets;
  
  // get the new variance after cl is chosen
  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]) {
      diff = (cmp_in_cell.usable[maxiset][0] 
	      + (int)(!cl->in_set[maxiset]) - average);
      // probably fewer operations this way
      variance += diff*diff;
    }
  }
  
  return (variance/num_maxisets);
  
} /*--variance()--*/


boolean in_key_set(int key_set, struct clause *cl)
{
  if (num_maxisets > 0 && key_set)
    return (cl->in_set[key_set]);
  else
    return false;

} /*--in_key_set()--*/


/*====================================================================\
| The following group of functions maintain the partition information |
\====================================================================*/

void init_a_partition(struct partition *part_ptr)
  /*-----------------------------------------------------------------*
   | Called by init_semantics().  Initialise a partition structure.  |
   *-----------------------------------------------------------------*/
{
  int cell, maxiset = 0;

  // initialize the partition
  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++){
      part_ptr->sos[maxiset][cell]    = 0;
      part_ptr->usable[maxiset][cell] = 0;
    }
  }
  
} /*--init_a_partition()--*/

void init_partitions()
  /*-----------------------------------------------------------------*
   | Called by init_semantics().  Initialise a partition structure.  |
   *-----------------------------------------------------------------*/
{

  init_a_partition(&set_in_cell);
  set_in_cell.sos[0][0]    = Stats[SOS_SIZE];
  set_in_cell.usable[0][0] = Stats[USABLE_SIZE];

  init_a_partition(&cmp_in_cell);
  cmp_in_cell.sos[0][0]    = Stats[SOS_SIZE];
  cmp_in_cell.usable[0][0] = Stats[USABLE_SIZE];

} /*--init_partitions()--*/


void init_sizes(int maxiset)
  /*-----------------------------------------------------------------*
   | Called by SCOTT fcn discard(). Resets the sizes of the recently |
   | discarded maxiset.                                              |
   *-----------------------------------------------------------------*/
{
  set_in_cell.sos[maxiset][0] = 0;
  cmp_in_cell.sos[maxiset][0] = 0;
  set_in_cell.usable[maxiset][0] = 0;
  cmp_in_cell.usable[maxiset][0] = 0;

} /*--init_sizes()--*/

void usable_part_insert_all(struct clause *cl)
  /*--------------------------------------------------------------------*
   | Called by scott_insert().  Incorporates the new clause into in the |
   | usable partition data structures.                                  |
   *--------------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.usable[0][num_sets]++;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]++;
	set_in_cell.usable[maxiset][0]++;
      }
    }
  }

  cmp_in_cell.usable[0][num_cmps]++;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]++;
	cmp_in_cell.usable[maxiset][0]++;
      }
    }
  }
} /*--usable_part_insert_all()--*/

void sos_part_insert_all(struct clause *cl)
  /*----------------------------------------------------------------------*
   | Called by scott_insert() Incorporates the new clause into in the sos |
   | partition data structures.                                           |
   *----------------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.sos[0][num_sets]++;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]++;
	set_in_cell.sos[maxiset][0]++;
      }
    }
  }
  
  cmp_in_cell.sos[0][num_cmps]++;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]++;
	cmp_in_cell.sos[maxiset][0]++;
      }
    }
  }
} /*--sos_part_insert_all()--*/

void usable_part_delete_all(struct clause *cl)
  /*------------------------------------------------------------------*
   | Called by scott_delete(). Removes the old clause from the usable |
   | partition data structures.                                       |
   *------------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.usable[0][num_sets]--;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]--;
	set_in_cell.usable[maxiset][0]--;
      }
    }
  }
  
  cmp_in_cell.usable[0][num_cmps]--;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]--;
	cmp_in_cell.usable[maxiset][0]--;
      }
    }
  }
} /*--usable_part_delete_all()--*/

void sos_part_delete_all(struct clause *cl)
  /*---------------------------------------------------------------*
   | Called by scott_delete(). Removes the old clause from the sos |
   | partition data structures.                                    |
   *---------------------------------------------------------------*/
{
  int maxiset, num_sets, num_cmps;

  num_sets = cl->num_sets;
  num_cmps = num_maxisets - num_sets;

  set_in_cell.sos[0][num_sets]--;
  if (num_sets > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]--;
	set_in_cell.sos[maxiset][0]--;
      }
    }
  }
  
  cmp_in_cell.sos[0][num_cmps]--;
  if (num_cmps > 0) {
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]--;
	cmp_in_cell.sos[maxiset][0]--;
      }
    }
  }
} /*--sos_part_delete_all()--*/

void scott_delete_from_base(struct clause *cl)
  /*---------------------------------------------------------------------*
   | Called by scott_delete(), pre_process() and post_process().  If     |
   | stype marks this as any type of Base clause, remove from either the |
   | conjecture listing or the semantic_base and change its stype.       |
   *---------------------------------------------------------------------*/
{
  // this could be buggy
  if (cl->stype == Conj) {
    delete_cl_ext(cl,&conjecture);
    tStats[CONJECTURE_SIZE]--;
  }
  if (cl->stype == Base) {
    delete_cl_ext(cl,&semantic_base);
    tStats[SEM_BASE_SIZE]--;
    cl->stype = Old_Base;
  }
  
  cl->stype = Norm;

} /*--scott_delete_from_base()--*/

void scott_insert(struct list *lptr, struct clause *cl)
  /*---------------------------------------------------------------------*
   | Called by append_cl() etc...  Incorporate the given clause into the |
   | semantic partition data structures.                                 |
   *---------------------------------------------------------------------*/
{
  if (lptr == Usable) {
     if (tFlags[SEMANTIC_GUIDANCE].val) {
      usable_part_insert_all(cl);
    }
  }

  else if (lptr == Sos) { 
    /* New clauses have in_set[0] set to false; however we need to do
     * this if we are moving clauses about */
    
    if (cl->in_set[0])
      null_set_size++;
    if (tFlags[SEMANTIC_GUIDANCE].val) {
      if (develop_semantics) sos_scan_insert(cl);
      sos_part_insert_all(cl);
    }
  }
 
  else if (lptr == First) {
    Stats[SOS_SIZE]++;
    if (tFlags[SEMANTIC_GUIDANCE].val) {
      sos_part_insert_all(cl);
    }
  }  

  else if (lptr == Demodulators) {
    if (tFlags[SEMANTIC_GUIDANCE].val) {
      if (tFlags[DEMOD_IN_SEM_BASE].val) {
	if (cl->stype != Base) {
	  append_cl_ext(cl,&semantic_base);
	  cl->stype = Base;
	  tStats[SEM_BASE_SIZE]++;
	}
      }
      else if (tFlags[DEMOD_IN_CONJECTURE].val) {
	if (cl->stype != Conj) {
	  append_cl_ext(cl,&conjecture);
	  cl->stype = Conj;
	  tStats[CONJECTURE_SIZE]++;
	}
      }
    }
  }
  
} /*--scott_insert()--*/


void scott_delete(struct list *lptr, struct clause *cl) 
  /*----------------------------------------------------------------------*
   | Called by rem_from_list(). Delete the given clause form the semantic |
   | data structures.                                                     |
   *----------------------------------------------------------------------*/
{
  if (lptr == Usable) {
     if (tFlags[SEMANTIC_GUIDANCE].val) {
      usable_part_delete_all(cl);
      scott_delete_from_base(cl);
    }
  }

  else if (lptr == Sos) { 
    if (tFlags[SEMANTIC_GUIDANCE].val) {
      if (cl->in_set[0]) 
	null_set_size--;
      if (null_set_size < 0)
	abort_scott("null set size gets set to a negative value");
      if (develop_semantics) sos_scan_delete(cl);
      sos_part_delete_all(cl);
    }
  }
 
  else if (lptr == First) {
    Stats[SOS_SIZE]--;
    if (tFlags[SEMANTIC_GUIDANCE].val) {
      sos_part_delete_all(cl);
    }
  }
  
  else if (lptr == Demodulators) {
    if (tFlags[SEMANTIC_GUIDANCE].val &&
	(tFlags[DEMOD_IN_SEM_BASE].val ||
	 tFlags[DEMOD_IN_CONJECTURE].val)) {
      scott_delete_from_base(cl);
    }
  }

} /*--scott_delete()--*/


void usable_part_insert(struct clause *cl, int new_maxiset) 
  /*-------------------------------------------------------------------*
   | Called by update_partition().  Incorporates the given clause into |
   | the new_maxiset in the usable partition data structures..         |
   *-------------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[new_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    set_in_cell.usable[0][num_sets-1]--;
    set_in_cell.usable[0][num_sets]++;
    set_in_cell.usable[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]++;
	if (maxiset != new_maxiset)
	  set_in_cell.usable[maxiset][num_sets-1]--;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    cmp_in_cell.usable[0][num_cmps-1]--;
    cmp_in_cell.usable[0][num_cmps]++;
    cmp_in_cell.usable[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]++;
	if (maxiset != new_maxiset)
	  cmp_in_cell.usable[maxiset][num_cmps-1]--;
      }
    }
  }
} /*--usable_part_insert()--*/

void sos_part_insert(struct clause *cl, int new_maxiset) 
  /*-----------------------------------------------------------------------*
   | Called by update_partition(). Incorporates the given clause into the  |
   | sos partition data structures after the addition of the new maxiset.  |
   *-----------------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[new_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    set_in_cell.sos[0][num_sets-1]--;
    set_in_cell.sos[0][num_sets]++;
    set_in_cell.sos[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]++;
	if (maxiset != new_maxiset)
	  set_in_cell.sos[maxiset][num_sets-1]--;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    cmp_in_cell.sos[0][num_cmps-1]--;
    cmp_in_cell.sos[0][num_cmps]++;
    cmp_in_cell.sos[new_maxiset][0]++;
    
    // adjust the table
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]++;
	if (maxiset != new_maxiset)
	  cmp_in_cell.sos[maxiset][num_cmps-1]--;
      }
    }
  }
} /*--sos_part_insert()--*/


void usable_part_delete(struct clause *cl, int old_maxiset) 
  /*-----------------------------------------------------------------*
   | Called by reset_partition().  Removes the given clause from the |
   | old_maxiset in the usable partition.                            |
   *-----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[old_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    
    set_in_cell.usable[0][num_sets]--;
    set_in_cell.usable[0][num_sets-1]++;
    //    set_in_cell.usable[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.usable[maxiset][num_sets]--;
	if (maxiset != old_maxiset)
	  set_in_cell.usable[maxiset][num_sets-1]++;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    
    cmp_in_cell.usable[0][num_cmps]--;
    cmp_in_cell.usable[0][num_cmps-1]++;
    //    cmp_in_cell.usable[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.usable[maxiset][num_cmps]--;
	if (maxiset != old_maxiset)
	  cmp_in_cell.usable[maxiset][num_cmps-1]++;
      }
    }
  }
} /*--usable_part_delete()--*/


void sos_part_delete(struct clause *cl, int old_maxiset) 
  /*-----------------------------------------------------------------*
   | Called by reset_partition().  Removes the given clause from the |
   | old_maxiset in the sos partition.                               |
   *-----------------------------------------------------------------*/
{
  int maxiset;

  if (cl->in_set[old_maxiset]) {
    int num_sets;
    
    num_sets = cl->num_sets;
    
    set_in_cell.sos[0][num_sets]--;
    set_in_cell.sos[0][num_sets-1]++;
    //    set_in_cell.sos[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && cl->in_set[maxiset]) {
	set_in_cell.sos[maxiset][num_sets]--;
	if (maxiset != old_maxiset)
	  set_in_cell.sos[maxiset][num_sets-1]++;
      }
    }
  }
  else {
    int num_cmps;
    
    num_cmps = num_maxisets - cl->num_sets;
    
    cmp_in_cell.sos[0][num_cmps]--;
    cmp_in_cell.sos[0][num_cmps-1]++;
    //    cmp_in_cell.sos[old_maxiset][0]--;  This gets set to zero later on.
    
    for (maxiset=1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset] && !cl->in_set[maxiset]) {
	cmp_in_cell.sos[maxiset][num_cmps]--;
	if (maxiset != old_maxiset)
	  cmp_in_cell.sos[maxiset][num_cmps-1]++;
      }
    }
  }
} /* sos_part_delete()--*/


void sos_part_mv_to_set(struct clause *cl, int old_maxiset)
  /*-------------------------------------------------------------*
   | Called by retest_failed_clauses().  Clause that was in the  |
   | complement of "old_maxiset" is moved to the "old_maxiset".  |
   *-------------------------------------------------------------*/
{
  int num_cmps, num_sets, maxiset;
  
  num_sets = cl->num_sets;
  num_cmps = num_maxisets - cl->num_sets;

  // move to new set 
  set_in_cell.sos[0][num_sets-1]--;
  set_in_cell.sos[0][num_sets]++;
  set_in_cell.sos[old_maxiset][0]++;

  // remove from previous cmp
  cmp_in_cell.sos[0][num_cmps+1]--;
  cmp_in_cell.sos[0][num_cmps]++;
  cmp_in_cell.sos[old_maxiset][0]--;

  // was once in cmp of old_maxiset
  cmp_in_cell.sos[old_maxiset][num_cmps+1]--;

  // adjust table
  for (maxiset=1; maxiset <= last_set_index; maxiset++){
    if (!active[maxiset]) continue;

    if (cl->in_set[maxiset]) {
      if (num_sets>1 && maxiset != old_maxiset) 
	set_in_cell.sos[maxiset][num_sets-1]--;
      set_in_cell.sos[maxiset][num_sets]++;
    }
    else {
      cmp_in_cell.sos[maxiset][num_cmps+1]--;
      if (num_cmps>0) cmp_in_cell.sos[maxiset][num_cmps]++;
    }
  }
  
} /*--sos_part_mv_to_set()--*/

void usable_part_mv_to_set(struct clause *cl, int old_maxiset)
  /*-------------------------------------------------------------*
   | Called by retest_failed_clauses().  Clause that was in the  |
   | complement of "old_maxiset" is moved to the "old_maxiset".  |
   *-------------------------------------------------------------*/
{
  int num_cmps, num_sets, maxiset;
  
  num_sets = cl->num_sets;
  num_cmps = num_maxisets - cl->num_sets;

  // move to new set 
  set_in_cell.usable[0][num_sets-1]--;
  set_in_cell.usable[0][num_sets]++;
  set_in_cell.usable[old_maxiset][0]++;

  // remove from previous cmp
  cmp_in_cell.usable[0][num_cmps+1]--;
  cmp_in_cell.usable[0][num_cmps]++;
  cmp_in_cell.usable[old_maxiset][0]--;

  // was once in cmp of old_maxiset
  cmp_in_cell.usable[old_maxiset][num_cmps+1]--;

  // adjust table
  for (maxiset=1; maxiset <= last_set_index; maxiset++){
    if (!active[maxiset]) continue;

    if (cl->in_set[maxiset]) {
      if (num_sets>1 && maxiset != old_maxiset) 
	set_in_cell.usable[maxiset][num_sets-1]--;
      set_in_cell.usable[maxiset][num_sets]++;
    }
    else {
      cmp_in_cell.usable[maxiset][num_cmps+1]--;
      if (num_cmps>0) cmp_in_cell.usable[maxiset][num_cmps]++;
    }
  }
  
} /*--usable_part_mv_to_set()--*/

/*================== DEBUGGING TOOLS =================================*/


void init_partition_box(int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY])
{
  int cell, maxiset;

  // Initialize the list partition copy
  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      set_in_list_cell[maxiset][cell] = 0;
    }
  }
  
} /*--init_partition_box()--*/

int get_list_part(char *list_name, struct list *lptr,
		  int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		  int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY])
  
{
  int maxiset, num_sets, num_cmps;
  struct clause *cl;
  int errors = 0;
  
  for (cl=lptr->first_cl;cl;cl=cl->next_cl) {
    
    num_sets = 0;
    for (maxiset = 1; maxiset <= last_set_index; maxiset++)
      if (active[maxiset] && cl->in_set[maxiset])
	num_sets++;
    
    if (cl->num_sets != num_sets) {
      printf("ERROR: Clause %d on %s gets wrong num_sets.\n",cl->id,list_name);
      errors++;
    }
    
    if (cl->pick_weight != (semantic_wt(cl))) {
      printf("ERROR: Clause %d on %s gets bad weight.\n",cl->id,list_name);
      errors++;
    }    

    num_cmps = num_maxisets - num_sets;

    set_in_list_cell[0][num_sets]++;
    cmp_in_list_cell[0][num_cmps]++;

    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset]) {
	if (cl->in_set[maxiset]) {
	  set_in_list_cell[maxiset][num_sets]++;
	  set_in_list_cell[maxiset][0]++;
	}
	else {
	  cmp_in_list_cell[maxiset][num_cmps]++;
	  cmp_in_list_cell[maxiset][0]++;
	}
      }
    }
  }

  return errors;

} /*--get_list_part()--*/


int compare_partitions(char *list_name,
		       int set_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int set_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell[MAXISET_ARRAY][MAXISET_ARRAY],
		       int cmp_in_list_cell_copy[MAXISET_ARRAY][MAXISET_ARRAY])

{
  int cell, maxiset;
  int errors = 0;

  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && 
	  set_in_list_cell_copy[maxiset][cell] != 
	  set_in_list_cell[maxiset][cell]) {
	printf("ERROR: set_in_cell.%s[%d][%d]) = %d but should be %d.\n",
	       list_name, 
	       maxiset,
	       cell,
	       set_in_list_cell[maxiset][cell],
	       set_in_list_cell_copy[maxiset][cell]);
	errors++;
      }
    }
  }

  for (cell = 0; cell <= last_set_index; cell++) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++) {
      if (active[maxiset] && 
	  cmp_in_list_cell_copy[maxiset][cell] != 
	  cmp_in_list_cell[maxiset][cell]) {
	printf("ERROR: cmp_in_cell.%s[%d][%d]) = %d but should be %d.\n",
	       list_name, 
	       maxiset,
	       cell,
	       cmp_in_list_cell[maxiset][cell],
	       cmp_in_list_cell_copy[maxiset][cell]);
	errors++;
      }
    }
  }

  
  return errors;

} /*--compare_partitions()--*/


void debug_partition(char *event)
  /*----------------------------------------------------------------------*
   | Called by OTTER fcn main() and various other functions.              |
   |                                                                      |
   | At end of search (while clause space is still intact) or while       |
   | debugging, checks that weights have been maintained properly and the |
   | partition information is correct.                                    |
   |                                                                      |
   | Exits iff a bad value has been detected.                             |
   *----------------------------------------------------------------------*/
{
  int errors = 0;
  int count, maxiset;
  struct partition set_in_cell_cp;
  struct partition cmp_in_cell_cp;
  
  init_a_partition(&set_in_cell_cp);
  init_a_partition(&cmp_in_cell_cp);

  count = 0;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    if (active[maxiset]) count++;

  if (num_maxisets != count) {
    printf("ERROR: num_maxisets = %d but should be %d.\n",
	   num_maxisets, count);
    errors++;
  }

  // check the SOS partitions
  errors += get_list_part("sos", Sos, 
			  set_in_cell_cp.sos,
			  cmp_in_cell_cp.sos);

  errors += get_list_part("input_first", First, 
			  set_in_cell_cp.sos,
			  cmp_in_cell_cp.sos);
  
  errors += compare_partitions("sos",
			       set_in_cell.sos, set_in_cell_cp.sos,
			       cmp_in_cell.sos, cmp_in_cell_cp.sos);

  // check the USABLE partitions
  errors += get_list_part("usable", Usable, 
			  set_in_cell_cp.usable,
			  cmp_in_cell_cp.usable);

  errors += compare_partitions("usable",
			       set_in_cell.usable, set_in_cell_cp.usable,
			       cmp_in_cell.usable, cmp_in_cell_cp.usable);

  if (errors) {
    message(0,"");
    abort_scott("%d errors found in the partition record %s",
		errors,event);
  }
} /*--debug_partition()--*/

