   	/*========================================================\
   	| FILE: scan.c: functions for scanning the clause space.  |
   	\========================================================*/
   
#include "scott.h"

void start_scan(int new_maxiset)
/*------------------------------------------------------------------------*
 | Called by execute_strategy()                                           |
 |                                                                        |
 | Some newly generated clauses may belong to none of our maxisets.       |
 | Since we have no semantic information about these clauses yet, we      |
 | don't really know how to treat them.  Hence we place these clauses     |
 | near the beginning of the scan so that the next maxiset will contain   |
 | at least one of them.                                                  |
 |                                                                        |
 | We start our scan by modelling clauses as follows:                     |
 |                                                                        |
 | a) clauses newly added to the semantic base, b) existing conjecture    |
 | clauses, c) remaining clauses form the seamntic base, d) (if possible) |
 | at least one clause that does not belong to any maxiset                |
 |                                                                        |
 | Various retesting occurs to ensure as much of this holds as possible   |
 *------------------------------------------------------------------------*/
{
  static int prev_time_limit = 0;
  struct clause_dptr *cp;
  int last_time_limit, variables, weight;
  
  // has time limit changed since last call?
  last_time_limit = get_time_limit();
  if (last_time_limit != prev_time_limit) {
    reset_base_markers();
    prev_time_limit = last_time_limit;
  }
  // allow for extended search time here
  set_time_limit(tParms[MAX_SEARCH_TIME].val);
  
  scan_new_base_clauses(new_maxiset); 
  if (!tFlags[SEMANTIC_GUIDANCE].val) return;
  scan_conjecture(new_maxiset);       

  if (!tFlags[SEMANTIC_GUIDANCE].val) return;

  // run thru' sos_scan[0] until at least one is modelled
  cp = sos_scan[0].first_cp;
  while (cp) {
    if (!cp->c->in_set[0]  &&   // skip clauses already in the null set
	cp->c->stype  == Norm) { // or already scanned

      // model the psuedo null clause
      if (model_clause(cp->c,new_maxiset)) {
	break;
      }	
      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
      
      // a genuine null set clause has been generated 
      cp->c->in_set[0] = true;
      cp->c->pick_weight = semantic_wt(cp->c);
      null_set_size++;

      message(2,"Could not model clause %d with the conjecture",cp->c->id);
      if (tParms[VERBOSITY].val >= 3) print_clause(stdout, cp->c);

      // Note: extension with this clause will still be attempted in scan_sos()

     /*------------------------------------------------------------------*
      | Clauses actually contradicting the conjecture will normally be   |
      | picked up by the unit conflick check; if we fail to model the    |
      | clause, then it is probably because the clause is too big or has |
      | too many distinct variables.                                     |
      *------------------------------------------------------------------*/

      variables = distinct_vars(cp->c);
      weight = cp->c->pick_weight;

      if (variables < Parms[MAX_DISTINCT_VARS].val &&
	  variables > variable_percentile(90)) {
	output_variable_distn();
	Parms[MAX_DISTINCT_VARS].val = variables - 1 ;
	message(3,"Resetting distinct variable limit to %d"
		" (search is now incomplete)\n",
		Parms[MAX_DISTINCT_VARS].val);
	fflush(stdout);
      }
      if (weight < Parms[MAX_WEIGHT].val &&
	       weight > weight_percentile(90)) {
	output_weight_distn();
	Parms[MAX_WEIGHT].val = weight - 1 ;
	message(3,"Resetting weight limit to %d"
		" (search is now incomplete)\n",
		Parms[MAX_WEIGHT].val);
	fflush(stdout);
      }
    }
    cp = cp->next;
  }

  if (!tFlags[SEMANTIC_GUIDANCE].val) return;
  scan_base(new_maxiset);             

  // if extending time limit was not needed, reset to old value
  if (get_max_time_in(new_maxiset) < last_time_limit){
    set_time_limit(last_time_limit);
  }

}/*--start_scan()--*/

void scan_new_base_clauses(int new_maxiset)
/*-----------------------------------------------------------------------*
 | Called by start_scan()                                                |
 |                                                                       |
 | Scan the clauses in the external list "new_base_clauses".  Remove     |
 | clauses from that list if we fail to model them or if their inclusion |
 | would cause the clause count to exceed MAX_NEW_BASE_CL.               |
 *-----------------------------------------------------------------------*/
{
  int count;
  struct clause_dptr *cp, *next;
  struct clause *cl;

  count = 0;
  cp = new_base_clauses.first_cp;
  while (cp) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    // clause already scanned, remove duplicate copies
    if (cl->in_set[new_maxiset]) 
      delete_cl_ext(cl,&new_base_clauses);
    else {
      count++;
      if (count > tParms[MAX_NEW_BASE_CL].val 
	  || !model_clause(cl, new_maxiset)){

	/*-------------------------------------------------------------*
	 | Clause retains its New_Base marker to prevent cycles. Don't |
	 | mark as Old_Base?                                           |
	 *-------------------------------------------------------------*/

	delete_cl_ext(cl,&new_base_clauses);
      }
    }
    cp = next;
  }

} /*--scan_new_base_clauses()--*/


void scan_conjecture(int new_maxiset)
/*-------------------------------------------------------------------------.
| Called by start_scan() and auto_set_sos().                               |
|                                                                          |
| Scan the clauses in external list "conjecture", moving those that        |
| cannot be modelled with the new base clauses to the old_conjeture list.  |
`-------------------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next_cp;
  struct clause *cl;

  // model each conjecture clause
  cp = conjecture.first_cp; 
  while (cp){
    next_cp = cp->next;
    cl = cp->c;
    if (cl->in_set[new_maxiset]) {
      delete_cl_ext(cl,&conjecture);
    }
    else if (!model_clause(cl,new_maxiset)) {
      delete_cl_ext(cl,&conjecture);
      append_cl_ext(cl,&old_conjecture);
    }
    cp = next_cp;
  }
}/*--scan_conjecture()--*/


void scan_base(int new_maxiset)
/*----------------------------------------------------------------------*
 | Called by start_scan()                                               |
 |                                                                      |
 | We continue our scan with clauses from the semantic base.  All these |
 | clauses are on either the USABLE or INPUT_FIRST list, and hence will |
 | not be affected by clause selection.  Note that, since all our scans |
 | have the semantic base clauses, our maxisets will tend to have a     |
 | common core. This places a strong restriction on the number of       |
 | maxisets that we will have to consider.                              |
 |                                                                      |
 | At various points we may choose to prepend new base clauses to our   |
 | list.  This may cause some later base clause to be omited.  There    |
 | stype field is set to false so that they may rejoin the base at some |
 | later point.                                                         |
 *----------------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next_cp;
  struct external_list  base_failures;

  base_failures.first_cp = NULL;
  base_failures.last_cp = NULL;

  cp = semantic_base.first_cp;
  while (cp) {
    next_cp = cp->next; // incase cp is deleted

    if (cp->c->in_set[new_maxiset]) { 
      // may be a new base clause, this cuts out the extra copy
      delete_cl_ext(cp->c,&semantic_base);      
    }
    else if (model_clause(cp->c,new_maxiset)) {
      // Need this in case Base markers are reset by start_scan()
      cp->c->stype = Base;
    }
    else {
      message(3,"Skipping clause %d from the semantic base", cp->c->id);
      delete_cl_ext(cp->c,&semantic_base);
      
      /*--------------------------------------------------------.
      |                   cp->c->stype = Norm;                  |
      |                                                         |
      | We don't do this becuase of the potential for cycles    |
      | --- reset_base_markers() takes care of this in the rare |
      | case that it is appropriate. Perhaps set to Old_Base?   |
      `--------------------------------------------------------*/
      
      cp->c->stype = Old_Base; //???

      prepend_cl_ext(cp->c,&base_failures);
      tStats[BASE_CL_SKIPPED]++;
      tStats[SEM_BASE_SIZE]--;
    }

    if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    cp = next_cp;
  }
  
  // second try at base clauses that failed
  cp = base_failures.first_cp;
  while (cp) {
    next_cp = cp->next; // before cp is deleted
    delete_cl_ext(cp->c,&base_failures);  // garbage collection

    if (model_clause(cp->c,new_maxiset)) {

      message(3,"Skipped clause %d is still covered", cp->c->id);
      // note clause is hard to model so place last in base
      append_cl_ext(cp->c,&semantic_base);
      cp->c->stype = Base;
      tStats[BASE_CL_SKIPPED]--;
      tStats[SEM_BASE_SIZE]++;
    }
    if (!tFlags[SEMANTIC_GUIDANCE].val) return;  
    
    cp = next_cp;
  }

  fflush(stdout);

}/*--scan_base()--*/


void scan_sos(int new_maxiset)
/*------------------------------------------------------------------------*
 | Called by execute_strategy()                                           |
 |                                                                        |
 | Extend the new maxiset by scanning the SOS in an order determined by   |
 | the previous attempt at a model.  This is essentially cell order, or   |
 | the increasing order of the number of maxisets that each clause        |
 | belongs to.                                                            |
 |                                                                        |
 | While performing this scan the order of the next scan altered so that  |
 | it will be significantly different from this one.                      |
 |                                                                        |
 | Clauses that are omitted from the new maxiset in the course of the     |
 | scan are tested against the final model, in case we get lucky and this |
 | model bears witness to its consistency (a cheap test).                 |
 *------------------------------------------------------------------------*/
{
  int cell, scanned = 0;
  struct clause *cl;
  struct clause_dptr *cp, *next;
  struct external_list new_sos_scan[MAXISET_ARRAY];

  // init new_sos_scan[]
  for (cell = 0; cell <= last_set_index; cell++) {
    new_sos_scan[cell].first_cp = NULL;
    new_sos_scan[cell].last_cp = NULL;
  }

  for (cell = 0 ; cell <= num_maxisets; cell++){  
    cp = sos_scan[cell].first_cp; 
    while (cp) {
      scanned++;
      cl = cp->c;
      delete_cl_ext(cl,&sos_scan[cell]);// pop off the top element

      if (cl->in_set[new_maxiset]) {      
	// already scanned by start_scan() or scan_base(), so copy to
	// the NEW sos_scan[] and alternate to deprecate similar scans
	if (scanned % 2)  append_cl_ext(cl,&new_sos_scan[cell+1]);
	else prepend_cl_ext(cl,&new_sos_scan[cell+1]);
      }
      else if (model_clause(cl,new_maxiset)) {
	append_cl_ext(cl,&new_sos_scan[cell+1]); // deprecate similar scans
      }
      else {
	if (!tFlags[SEMANTIC_GUIDANCE].val) return;
	// alternate to deprecate similar scans 
	if (scanned % 2)  append_cl_ext(cl,&new_sos_scan[cell]);
	else prepend_cl_ext(cl,&new_sos_scan[cell]);

	// prepend so that new failures occur first
	prepend_cl_ext(cl,&failed_clauses[new_maxiset]); 
      }
      cp = sos_scan[cell].first_cp;
    }
  } 

  // note all the ptrs in sos_scan[] have been removed
  for (cell = 0; cell <= last_set_index; cell++) 
    sos_scan[cell] = new_sos_scan[cell];
  
  // retest failed clauses
  cp = failed_clauses[new_maxiset].first_cp;
  while (cp) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[new_maxiset]) 
      abort_scott("Retesting failed clause already true in model");
    
    if (test_clause(cl,new_maxiset)) {

      if (cl->container == Sos) {
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
      }

      // this list should be quite short
      delete_cl_ext(cl,&failed_clauses[new_maxiset]);
      tStats[SUCCESSFUL_RETESTS]++;
    }

    if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }

}/*--scan_sos()--*/


void scan_usable(int new_maxiset)
/*-----------------------------------------------------------------------*
 | Called by execute_strategy().                                         |
 |                                                                       |
 | Extends maxiset with the clauses from the Usable list.  This ensures  |
 | partition data is correct and that selection of base extensions is    |
 | optimal.  This may be expensive since the USABLE list will eventually |
 | contain many hard to model clauses, so after a while we resort to     |
 | simply testing.                                                       |
 *-----------------------------------------------------------------------*/
{
  struct clause *cl;
  
  if (tFlags[MODEL_USABLE_LIST].val) {
    for (cl= Usable->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) model_clause(cl,new_maxiset);
      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  else {
    for (cl= Usable->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) test_clause(cl,new_maxiset);
      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  
}/*--scan_usable()--*/


void scan_first(int new_maxiset)
/*------------------------------------------------------------------------*
 | Called by execute_strategy().                                          |
 |                                                                        |
 | Extends maxiset with the clauses from the INPUT FIRST list.  Note that |
 | these will soon be moved to the USABLE LIST, and may be hard to model  |
 | clauses.                                                               |
 *------------------------------------------------------------------------*/
{
  struct clause *cl;
  
  if (tFlags[MODEL_USABLE_LIST].val) {
    for (cl= First->first_cl; cl; cl = cl->next_cl) {
      // clauses in the semantic base have already been scanned
      if (!cl->in_set[new_maxiset]) model_clause(cl,new_maxiset);
      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    }
  }
  for (cl= First->first_cl; cl; cl = cl->next_cl) {
    // clauses in the semantic base have already been scanned
    if (!cl->in_set[new_maxiset]) test_clause(cl,new_maxiset);
    if (!tFlags[SEMANTIC_GUIDANCE].val) return;
  }
  
}/*--scan_usable()--*/


void invert_existing_base()
/*----------------------------------------------------------------------*
 | Called by keep().                                                    |
 |                                                                      |
 | Append non-base clauses on USABLE to the new base clauses. This is a |
 | shuffle that helps to avoid SOS covers.                              |
 *----------------------------------------------------------------------*/
{
  struct clause *cl;
  char in_set[MAXISET_ARRAY+3];
  
  message(3,"");
  message(3,"----> Inverting base");
  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    if(cl->stype == Norm) {
      in_set_to_str(in_set,cl);
      prepend_cl_ext(cl,&semantic_base);
      message(3,"<---- Extending the semantic base with clause %d %s", 
	      cl->id, in_set);	
    }

  // so that next call to scan_base() can set base markers appropriately
  reset_base_markers(); 
  message(3,"");
  
}/*invert_existing_base()--*/


void print_scott_sets(FILE *fp, struct clause *cl)
/*-------------------------------------------*
 | Called by OTTER function print_clause().  |
 *-------------------------------------------*/
{
  if (num_maxisets > 0) {
    char in_set[MAXISET_ARRAY+6];
    
    in_set_to_str(in_set,cl);
    fprintf(fp, " %s ",in_set);
  }
}/*print_scott_sets()--*/


void print_in_set(struct external_list *elist)
/*--------------------------------------------------*
 | Debugging routine called by shuffle_ext_list().  |
 *--------------------------------------------------*/
{
  struct clause_dptr *cp;
  char in_set[MAXISET_ARRAY+3];
  
  for (cp = elist->first_cp;cp && cp->c; cp = cp->next){
    in_set_to_str(in_set,cp->c);
    printf("%3d %s\n", cp->c->id, in_set);
  }
  fflush(stdout);

}/* print_in_set()--*/


void shuffle_ext_list(struct external_list *elist)
/*-----------------------------------------------------------------------*
 | Called by extend_base() and execute_strategy()                        |
 |                                                                       |
 | Attempts to shuffle the clauses in the given list so the the in_set   |
 | fields of adjacent clauses are different.  This alters the scan so as |
 | to maximise the likelihood that the next attempt at an MCT will be    |
 | different to any that already exist.                                  |
 *-----------------------------------------------------------------------*/
{      
  struct clause_dptr *cp;
  struct clause *prev_cl, *next_cl;
  struct external_list new_elist, failed;
  int count_before = 0;
  int count_mid = 0;
  int count_after = 0;

  /*
  printf("\nexternal list before shuffle\n");
  print_in_set(elist);
  */

  // debugging code
  for (cp = elist->first_cp;cp && cp->c; cp = cp->next)
    count_before++;
  
  new_elist.first_cp = new_elist.last_cp = NULL;
  failed.first_cp = failed.last_cp = NULL;
  
  cp = elist->first_cp;
  
  if (cp && cp->c) { // there must be at least one clause

    count_mid++;    
    prev_cl = cp->c;
    delete_cl_ext(prev_cl,elist);
    append_cl_ext(prev_cl,&new_elist);
    cp = elist->first_cp;
    while (cp && cp->c) {
      
      count_mid++;
      next_cl = cp->c;
      delete_cl_ext(next_cl,elist);
      
      if (!in_set_identical(next_cl,prev_cl)) {
	append_cl_ext(next_cl,&new_elist);
	prev_cl = next_cl;
      }
      else if (!in_set_identical(next_cl,new_elist.first_cp->c))
	prepend_cl_ext(next_cl,&new_elist);
      else if (count_mid % 3 ||
	       count_mid % 5) // shuffle identical guys
      	append_cl_ext(next_cl,&failed);
      else 
      	prepend_cl_ext(next_cl,&failed);

      cp = elist->first_cp;
    }
    
    // debugging code
    if (count_before != count_mid) abort_scott("bad shuffle");
    /*
      printf("\nexternal list mid shuffle\n");
      print_in_set(&new_elist);
      printf("\n");
      print_in_set(&failed);
    */

    // copy the shuffled clauses across to the elist
    // which should be empty by now.
    cp =  new_elist.first_cp;
    while (cp && cp->c) {
      count_after++;
      delete_cl_ext(cp->c,&new_elist);
      append_cl_ext(cp->c,elist);
      cp =  new_elist.first_cp;
    }

    // append matching and prepend non-matching failed clause to the
    // elist this should preserve the shuffe at least at the beginning
    cp = failed.first_cp;
    while (cp && cp->c) {
      count_after++;
      next_cl = cp->c;
      delete_cl_ext(next_cl,&failed);
      
      if (!in_set_identical(next_cl,elist->first_cp->c))
	prepend_cl_ext(next_cl,elist);
      else
	append_cl_ext(next_cl,elist);
      
      cp = failed.first_cp;
    }
  }

  // debugging code
  if (count_after != count_mid) abort_scott("bad shuffle");
  /*
    printf("\nexternal list after shuffle\n");
    print_in_set(elist);
  */
} /*shuffle_ext_list()--*/


void reset_base_markers()
  /*----------------------------------------------------------------------*
   | Called by scan_base() and invert_base().                             |
   |                                                                      |
   | Resets all stype fields to false prior to a scan_base() which resets |
   | them appropriately.  Have to do this if we invert the base of if we  |
   | change the timelimit on our model search.                            |
   *----------------------------------------------------------------------*/
{
  struct clause *cl;

  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    if (cl->stype <= Base) cl->stype = Norm;

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    if (cl->stype <= Base) cl->stype = Norm;
  
  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    if (cl->stype <= Base) cl->stype = Norm;

} /*--reset_base_markers()--*/


struct clause *search_list(struct list *lptr, 
			   int skip_maxiset,
			   scan_type skip_type)
  /*-----------------------------------------------------------------------*
   |                 Called by extend_base().                              |
   |                                                                       |
   | Search given list for a clause with the following conditions:         |
   |                                                                       |
   | a). clause doesnt not have an stype that is greater than 'skip_type'  |
   | b). clause does not belong to the skip_maxiset (if given),            |
   | c). clause is consistent with the existing base,                      |
   | d). and clause does not belong to every maxiset,                      |
   | e). preference is given to clauses with more maxisets.                |
   |                                                                       |
   | We search backwards so that new clauses get a chance to participate.  |
   *-----------------------------------------------------------------------*/

{
  struct clause *cl, *new_cl, *new_base_first;

  if (new_base_clauses.first_cp)
    new_base_first = new_base_clauses.first_cp->c;
  else new_base_first = NULL;

  for (new_cl = NULL, cl = lptr->last_cl; cl; cl = cl->prev_cl) {

    // skip clauses of above the given type
    if (cl->stype > skip_type) continue;

    // skip the key set
    if (num_maxisets > 0 &&  
	skip_maxiset && 
	cl->in_set[skip_maxiset]) 
      continue;

    if (skip_type < Conj && num_maxisets > 2) {
      if (cl->num_sets == 0) continue; // maybe inconsistent with base?
      if (cl->num_sets == num_maxisets) continue; // skip_set is 0 ?
      if (new_base_first && in_set_identical(cl,new_base_first))
	continue;
    }

    if (!new_cl	
	// || (cl->num_sets < new_cl->num_sets)
	|| (cl->num_sets > new_cl->num_sets)
	) new_cl = cl;
  }
  return new_cl;

} /*--search_list()--*/


boolean extend_base(int key_maxiset)
  /*------------------------------------------------------------------*
   | Called by execute_strategy() and discard_bad_maxisets().  Search |
   | lists for a suitable clause to add to the existing base          |
   *------------------------------------------------------------------*/
{
  struct clause *new_base_cl = NULL;

  message(3,"");
  message(3, "----> Attempting to extend the semantic base");

  new_base_cl = search_list(Usable,key_maxiset,Norm);
  
  if (!new_base_cl) {
    message(3,"No suitable clause on USABLE list...searching SOS");
    new_base_cl = search_list(Sos,key_maxiset,Norm);
  }
  
  if (!new_base_cl) {
    message(3,"No suitable clause on SOS, consider old base clauses");
    new_base_cl = search_list(Usable,key_maxiset,Old_Base);
  }
  
  if (!new_base_cl) {
    message(3,"No suitable clause on USABLE list...searching SOS");
    new_base_cl = search_list(Sos,key_maxiset,Old_Base);
  }

  //???
  if (!new_base_cl) {
    message(3,"No suitable clause on SOS, consider all clauses");
    new_base_cl = search_list(Usable,key_maxiset,All);
  }
  if (!new_base_cl) {
    message(3,"No suitable clause on USABLE list...searching SOS");
    new_base_cl = search_list(Sos,key_maxiset,All);
  }

  if (!new_base_cl) {  
    record_event(STRENGTHEN_BASE_FAILURE);
    tStats[BASE_EXTENSION_FAILS]++;
    message(3,"<---- Could not find a suitable clause to add to the semantic base");
    shuffle_ext_list(&new_base_clauses);
    message(3,"");
    return false;
  }
  else {
    char in_set[MAXISET_ARRAY+3];

    in_set_to_str(in_set,new_base_cl);
    message(3,"<---- Extending the semantic base with clause %d %s", 
	    new_base_cl->id, in_set);
    //  print_clause(stdout,*new_base_cl);
    new_base_cl->stype = New_Base;  // to avoid cycles
    prepend_cl_ext(new_base_cl,&new_base_clauses);
    message(3,"");
    return true;
  }

} /*--extend_base()--*/


boolean model_clause(struct clause *cl, int maxiset)
  /*---------------------------------------------------------------------*
   | Called by start_scan(), scan_sos(), scan_first(), scan_usable(),    |
   | scan_base(), scan_new_base_clauses().                               |
   |                                                                     |
   | Attempts to extend maxiset with the given clause.  Returns TRUE iff |
   | the extension was sucessful.  Returns FALSE if semantics are        |
   | disabled or if the clause is not modelled                           |
   *---------------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];

  if (!cl->first_lit) return false;  // can't model empty clause!

  if (!make_finder_clause(finder_cl, cl)) {
    return false;
  }

  if (cl->container &&      // cl has been subsummed
      cl->in_set[maxiset])  // there's probably a bug somewhere
    abort_scott("model_clause() gets a clause (%d) that is "
		"already true in the new maxiset", cl->id);

  if (model(finder_cl, maxiset, false, true)) {
    cl->in_set[maxiset] = true;
    if (cl->in_set[0] && cl->container == Sos) 
      null_set_size--;
    cl->in_set[0] = false;
    cl->num_sets++;
    message(4,"%s %%passed",finder_cl);
  }
  else {
    //    cl->in_set[maxiset] = false;
    message(4,"%s %%failed",finder_cl);
  }

  // this is a cross check
  cl->pick_weight = semantic_wt(cl);

  return cl->in_set[maxiset];
  
}/*--model_clause()--*/


boolean test_clause(struct clause *cl, int maxiset)
/*-----------------------------------------------------------------------*
 | Called by scan_sos(), scan_usable() and retest_failed_clauses().      |
 |                                                                       |
 | Tests the given clause in the current model of the given maxiset.  If |
 | successful, the clause is added to that maxiset and the fcn returns   |
 | TRUE.  Otherwise, returns FALSE.                                      |
 *-----------------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];

  if (!cl->first_lit) return false;  // can't model empty clause!

  if (!make_finder_clause(finder_cl, cl)) {
    return false;
  }

  if (cl->container &&      // cl has been subsummed
      cl->in_set[maxiset])  // probably a bug somewhere
    abort_scott("test_clause() gets a clause (%d) that is "
		"already true in the new maxiset", cl->id);

  if (model(finder_cl, maxiset, true, true)) {  
    cl->in_set[maxiset] = true;
    if (cl->in_set[0] && cl->container == Sos) 
      null_set_size--;
    cl->in_set[0] = false;
    cl->num_sets++;
    message(4,"%s %%passed",finder_cl);
  } 
  else {
    //cl->in_set[maxiset] = false;
    message(4,"%s %%failed",finder_cl);
  }
  
  // this is a cross check!!
  cl->pick_weight = semantic_wt(cl);

  return cl->in_set[maxiset];
  
}  /*--test_clause()--*/


void check_null_set_size(char *event)
  /*--------------------------------------------------------------------*
   |      Debugging routine called via the SCOTT_DEBUG macro.           |
   |                                                                    |
   | Check that the value of null_set_size aggrees with the pseudo null |
   | clauses on the SOS.                                                |
   *--------------------------------------------------------------------*/
{
  struct clause *cl;
  int null_clause_count = 0;
  
  for (cl=Sos->first_cl; cl; cl = cl->next_cl)
    if (cl->in_set[0]) null_clause_count++;

  if (null_set_size != null_clause_count) {
    message(0,"");
    abort_scott("null_set_size is %d by SOS contains %d psuedo null clauses!",
		null_set_size, null_clause_count, event);
  }
}  /*--check_null_set_size()--*/




