/*------------------------------------------------------------------------.
|                              SCOTT.c                                    |
|                                                                         |
| This file contains the functions called by OTTER when compiled with     |
| SCOTT defined. They set up the data structures specific to SCOTT and    |
| call the requisite FINDER.                                              |
|                                                                         |
| SCOTT uses the FINDER subsystem to determine approximately maximal      |
| consistent sets (maxisets) of the evolving clause space.  These         |
| maxisets are used to implement a guidance strategy.                     |
|                                                                         |
| Suppose M1, M2, M3, ..., Mn are maximally consistent set on our clause  |
| space.  Then (by completeness) every proof must contain at least one    |
| clause from each of the complements M1', M2, ..., Mn'.  There may, of   |
| course, be some necessary infernce that will belong to the initerection |
| M1 ^ M2 ^ M3 ^...^ Mn, but since proofs tend to be short and maxisets   |
| plenetiful, they will be in the minorirty.  In order to increase the    |
| probability that these necessary inferences are made, we restrict our   |
| selection to each complement in turn, and then the intersection.        |
|                                                                         |
| Main contact points between OTTER to SCOTT:                             |
|                                                                         |
| (1) semantic selection() selects best SOS clause based on semantic      |
| criterion.                                                              |
|                                                                         |
| (2) add_semantic_wt() incorporates new clauses into the existing        |
| collection of maxisets.                                                 |
|                                                                         |
| (4) scott_insert(), scott_delete() utilities that add/remove clauses    |
| to/form the partition record and the scanning order.                    |
|                                                                         |
| (5) automatic_3_settings() provides SCOTTs on autonomous mode           |
|                                                                         |
| (6) scott_control_memory() provides more sophisticated memory control   |
|                                                                         |
| Main contact points between SCOTT to FINDER:                            |
|                                                                         |
| (1) model() attempts to extend the current theory with a given clause.  |
|                                                                         |
| (2) declare_otter_symbols() generates function declarations for the     |
| signature of the given problem.                                         |
`------------------------------------------------------------------------*/

#define OTTER_VERSION  "3.1-b1"
#define VERSION_DATE   "Feb 2001"

#define SCOTT_VERSION  "5.3.9 or 6beta"
#define SCOTT_DATE     "5 June 2001"

#define FINDER_VERSION "4.0.5"
#define FINDER_DATE    "9 March 2001"

#include "scott.h"
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>

#define MAX_HISTORY 100000
static char History[MAX_HISTORY];

static int givens_avoided;
static int givens_avoided_by[MAXISET_ARRAY];
static int last_num_updates[MAXISET_ARRAY];
static int giv_cl_from_comp[MAXISET_ARRAY];

int semantic_wt(struct clause *cl)
  /*------------------------------------------------------------------------.
  | Called by add_semantic_wt(), model_clause(), test_clause(), and         |
  | decendents of check_partition(), init_clause_space(), start_scan(), and |
  | discard().                                                              |
  |                                                                         |
  | Returns the semantic weight of the given clause. Clauses that belong to |
  | more than one maxisets have their weight altered, according to the      |
  | parameter TRUE_WEIGHT.                                                  |
  `------------------------------------------------------------------------*/
{
  if (cl->num_sets < 2) return (cl->length);
  else {
    char str[5];
    int weight = 0;
    strcpy(str,"");

    // a very chunky round function
    sprintf(str,"%.0f",
	    (cl->num_sets -1)
	    *tParms[TRUE_WEIGHT].val
	    /(float)num_maxisets);

    sscanf(str,"%d",&weight);

    return (cl->length + weight);
  }
}/*--semantic_wt()--*/

void add_semantic_wt(struct clause *cl)
  /*----------------------------------------------------------------------.
  | Called by OTTER fcn pre_process().                                    |
  |                                                                       |
  | Fattens new clause according to the number of sets (maxisets) that    |
  | contain the clause. Also sets semantic information about the clause.  |
  |                                                                       |
  | If in the course of this call the search time limit changes, the      |
  | global variable time_limit_has_changed is set.                        |
  `----------------------------------------------------------------------*/
{
  char finder_cl[CLAUSE_STRING_ARRAY];
  int maxiset;

  SCOTT_CLOCK_START(SEM_GUIDE_TIME);
  if (cl->first_lit) {
    if(!make_finder_clause(finder_cl, cl)) {
      SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
      return;
    }
    for (maxiset = 1; maxiset <= last_set_index; maxiset++){
      if (active[maxiset]) {
	if (model(finder_cl,maxiset,false, true) ){
	  cl->in_set[maxiset] = true;
	  cl->num_sets++;
	}
      }
    }
    cl->length = cl->pick_weight;
    cl->pick_weight = semantic_wt(cl);
  }
  SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);

}/*--add_semantic_wt()--*/

void retest_failed_clauses(int maxiset)
  /*----------------------------------------------------------------------.
  | Called by check_failed_clauses().                                     |
  |                                                                       |
  | The model of given maxiset has just been updated so all the clauses   |
  | external to that maxiset are retested in the new model.  Because      |
  | these clause have already failed once, a new search is still quite    |
  | likely to fail with the given resource limits, hence we only test the |
  | clause in the new model.                                              |
  `----------------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next;
  struct clause *cl;

  message(3,"Retesting failed clauses in model of maxiset %d", maxiset);

  cp = failed_clauses[maxiset].first_cp;
  while (cp  && cp->c) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[maxiset])
      abort_scott("Function retest_failed_clauses()"
		  "called with clause already true in model");

    if (test_clause(cl,maxiset)) {

      if (cl->container == Sos ||
	  cl->container == First){
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
	sos_part_mv_to_set(cl,maxiset);
      }
      else if (cl->container == Usable) {
	usable_part_mv_to_set(cl,maxiset);
      }
      // could still be a demodualtor or deleted clause
      delete_cl_ext(cl,&failed_clauses[maxiset]);
      tStats[SUCCESSFUL_RETESTS]++;
    }

    if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }

} /*--retest_failed_clauses()--*/

void remodel_failed_sos_clauses(int maxiset)
  /*----------------------------------------------------------------------.
  | Called by check_failed_clauses().                                     |
  |                                                                       |
  | Our bound on search time has just changed, so we re-model SOS clauses |
  | to improve the quality of guidance.                                   |
  `----------------------------------------------------------------------*/
{
  struct clause_dptr *cp, *next;
  struct clause *cl;

  cp = failed_clauses[maxiset].first_cp;
  while (cp && cp->c) {
    next = cp->next; // before cp is deleted
    cl = cp->c;

    if (cl->in_set[maxiset])
      abort_scott("Function remodel_failed_clauses() called"
		  "with clause already true in model");

    if (cl->container == Sos ||
	cl->container == First) { // only remodel SOS
      if (model_clause(cl,maxiset)) {
	delete_cl_ext(cl,&sos_scan[cl->num_sets-1]);
	append_cl_ext(cl,&sos_scan[cl->num_sets]);
	sos_part_mv_to_set(cl,maxiset);
	delete_cl_ext(cl,&failed_clauses[maxiset]);
	tStats[SUCCESSFUL_RETESTS]++;
      }
    }
    if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    cp = next;
  }


} /*--remodel_failed_sos_clauses()--*/


void check_failed_clauses()
  /*--------------------------------------------------------------------.
  | Called by semantic_selection().                                     |
  |                                                                     |
  | Performs two checks in the following order:                         |
  |                                                                     |
  | A) If our time limit has changed between calls to this function,    |
  | extension of each maxiset by each of its failed SOS clauses is      |
  | attempted.                                                          |
  |                                                                     |
  | B) If any maxiset has a new model (witnessing its consistency) each |
  | of its failed clauses (all lists) are retested.                     |
  `--------------------------------------------------------------------*/
{
  static int prev_time_limit = 0;
  float est_cost;
  int time_left, maxiset;

  if (num_maxisets > 0
      && prev_time_limit
      && get_time_limit() > prev_time_limit) {

    //first do we want to do this??
    est_cost = num_failed_clauses() * get_time_limit();

    if (Parms[MAX_SECONDS].val > 0)
      time_left = (Parms[MAX_SECONDS].val*1000) - run_time();
    else time_left = MAX_INT;

    if (est_cost*fParms[COST_MARGIN].val < time_left) {
      int start;  
      est_cost /= 1000;

      message(3,"");
      message(3,"Remodelling failed clauses (estimated cost %.2f secs)",
	      est_cost);

      // cross check our conservative estimate 
      start = run_time();
      for (maxiset = 1; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) remodel_failed_sos_clauses(maxiset);
      message(3,"                           (actual cost    %.2f secs)",
	      (float)(run_time()-start)/1000);

      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
      print_partition("after failed clauses remodelled");
    }
  }
  prev_time_limit = get_time_limit();

  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset]
	&& last_num_updates[maxiset]
	&& get_updates(maxiset) > last_num_updates[maxiset]){
      message(3,"");
      retest_failed_clauses(maxiset);
      if (!tFlags[SEMANTIC_GUIDANCE].val) return;
    }
    last_num_updates[maxiset] = get_updates(maxiset);
  }

}/*check_failed_clauses()--*/


struct clause *semantic_selection(struct list *lptr)
  /*---------------------------------------------------------------------.
  | Called by OTTER function find_given_cl() if and only if SCOTT has    |
  | been defined.                                                        |
  |                                                                      |
  | Select the next given clause according to the following semantic     |
  | criterion.                                                           |
  |                                                                      |
  | 1. Select the first clause on the "input_first" list if one exists,  |
  | else                                                                 |
  |                                                                      |
  | 2. Move appropriate clause to the "input_first" list if we are       |
  | performing a semantic queue, and select the first clause form that   |
  | list, else                                                           |
  |                                                                      |
  | 3. Select the first, lightest SOS clause that does not belong to the |
  | current set_to_skip.                                                 |
  |                                                                      |
  | Plus various contortions in order to record the event history.       |
  `---------------------------------------------------------------------*/
{

  static boolean first = false;
  static boolean semantic = false;
  static boolean standard = false;
  static int count = 0;
  static int ext_fails = 0;
  static int skip_revisions = 0;
  int set_to_skip = 0;
  //  static float last_time = 0;
  struct clause *giv_cl;

  SCOTT_DEBUG("before semantic_selection()");

  SCOTT_CLOCK_START(SEM_GUIDE_TIME);

//   if (tFlags[SEMANTIC_GUIDANCE].val) {
//     set_cost_of_given(run_time() - last_time);
//     set_time_saved(givens_avoided);
//   }

  if (!First->first_cl) set_sem_queue(lptr);

  if (First->first_cl) {
    // catch remaining semantic queue clauses
    if (!first) {
      record_event("\n%dm->f...",num_maxisets);
      first = true;
      standard = semantic = false;
      count = 0;
    }
    count++;
    tStats[GIVENS_FROM_FIRST]++;
    message(3,"");
    message(3,"selecting from semantic queue",set_to_skip);

    givens_avoided += (Stats[SOS_SIZE]-1);
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    //    last_time = run_time();
    return First->first_cl;
  }

  if (!lptr->first_cl) {
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    return NULL; // henceforth we may assume list is non-empty
  }

  //--revise semantic information--------------------------------------//
  if (tFlags[SEMANTIC_GUIDANCE].val) {

    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    SCOTT_CLOCK_START(SEM_DEV_TIME);

    check_failed_clauses();
    if (discard_bad_maxisets()){
      record_event("(%d)...",count);
      count = 0;
    }
    if (develop_semantics) {
      if (skip_revisions > 0) {
	skip_revisions--;
      }
      else {
	int last_num_attempts;
	int int_fails = 0;
	last_num_attempts = tStats[ATTEMPTS];
	
	while (tFlags[SEMANTIC_GUIDANCE].val
	       && !First->first_cl
	       && (tParms[MAX_INT_FAILS].val == 0 ||
		   (int_fails >= 0  
		    && int_fails < tParms[MAX_INT_FAILS].val))
	       && cover_new_clauses(TEST)) {
	  if (!persist()) {
	    int_fails = -MAX_INT; // avoids sequential calls to persist
	    break;
	  };
	  int_fails += execute_strategy(cover_new_clauses);
	}
	
	while (tFlags[SEMANTIC_GUIDANCE].val
	       && !First->first_cl
	       && (tParms[MAX_INT_FAILS].val == 0 ||
		   (int_fails >= 0  
		    && int_fails < tParms[MAX_INT_FAILS].val))
	       && refine_choice(TEST)) {
	  if (!persist()) {
	    int_fails = -MAX_INT; // avoids sequential calls to persist
	    break;
	  };
	  int_fails += execute_strategy(refine_choice);
	}

	if (tParms[REVS_TO_SKIP_ON_FAIL].val > 0 &&
	    tParms[MAX_EXT_FAILS].val > 0 ) {
	  if (int_fails >= tParms[MAX_INT_FAILS].val) ext_fails++;
	  if (ext_fails >= tParms[MAX_EXT_FAILS].val) {
	    skip_revisions = tParms[REVS_TO_SKIP_ON_FAIL].val;
	    message(2,"====> Skipping the next %d revisons.",  
		    tParms[REVS_TO_SKIP_ON_FAIL].val);
	    message(2,""),
	      ext_fails = 0;
	  }
	}

	if (tStats[ATTEMPTS] > last_num_attempts) {
	  first = standard = semantic = false;
	  record_event(" (%d)",count);
	}
      }
    }
//     else if (num_maxisets < tParms[MIN_NUM_MAXISETS].val) {
//       //  else if (num_maxisets < 2) {
//       // principle point at which semantics is disabled
//       disable_semantics("Semantic guidance too expensive to continue");
//     }

    check_failed_clauses();
    if (discard_bad_maxisets()){
      record_event("(%d)...",count);
      count = 0;
    }

    SCOTT_CLOCK_STOP(SEM_DEV_TIME);
    SCOTT_CLOCK_START(SEM_GUIDE_TIME);

  } //--end revise semantic information-------------------------------//

  if (!First->first_cl) set_sem_queue(lptr);

  // catch any new input first clauses
  if (First->first_cl) {
    if (!first) {
      record_event("\n%dm->b...",num_maxisets);
      first = true;
      standard = semantic = false;
      count = 0;
    }
    count++;
    tStats[GIVENS_FROM_FIRST]++;
    message(3,"");
    message(3,"selecting from semantic queue",set_to_skip);
    givens_avoided += (Stats[SOS_SIZE]-1);
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    //    last_time = run_time();
    return First->first_cl;
  }
  if (first) record_event(" (%d)",count);
  first = false;

  // standard selection
  if (!tFlags[SEMANTIC_GUIDANCE].val ||
      num_maxisets == 0 || set_to_skip == -1) {
    if (!standard) {
      record_event("\n%dm->l...\n",num_maxisets);
      standard = true;
      first = semantic = false;
    }
    count++;

    message(3,"");
    message(3,"selecting lightest clause",set_to_skip);
    giv_cl = find_lightest_cl(lptr);
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    //    last_time = run_time();
    return giv_cl;
  }
  standard = false;

  // else semantic selection
  if (!semantic) {
    record_event("\n%dm->s...",num_maxisets);
    semantic = true;
    first =  false;
    count = 0;
  }

  count++;
  set_to_skip = next_set_to_skip(); // SEMANTIC_GUIDANCE is on
  
  if (Parms[PICK_GIVEN_RATIO].val != -1 &&
      giv_cl_from_comp[set_to_skip] % (Parms[PICK_GIVEN_RATIO].val + 1) == 0){

    giv_cl = find_first_sem_cl(lptr,set_to_skip);
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    //    last_time = run_time();
    giv_cl_from_comp[set_to_skip]++;
    return giv_cl;
  }
  else {
    giv_cl = find_best_sem_cl(lptr,set_to_skip);
    SCOTT_CLOCK_STOP(SEM_GUIDE_TIME);
    //    last_time = run_time();
    giv_cl_from_comp[set_to_skip]++;
    return giv_cl;
  }

}/*--semantic_selection()--*/


struct clause *find_first_sem_cl(struct list *lptr, int set_to_skip)
  /*----------------------------------------------------------------------.
  | Called by semantic_selection().                                       |
  |                                                                       |
  | Returns the "first" clause on the list belonging to the complement of |
  | the current set_to_skip.                                              |
  `----------------------------------------------------------------------*/
{
  struct clause *cl, *first = NULL;
  int skipped = 0;

  message(3,"");
  message(3,"skip maxiset number %d",set_to_skip);

  if (set_to_skip == 0) {
    // get first clause belonging to the most maxisets
    first = lptr->first_cl;
    for (cl = first->next_cl; cl; cl = cl->next_cl){
      if (cl->num_sets > first->num_sets) first = cl;
      if (cl->num_sets == num_maxisets) break;
    }
  }
  else {
    // get first clause not in the skip_theory
    for (cl = lptr->first_cl; cl; cl = cl->next_cl){
      if (cl->in_set[set_to_skip]) {
	// clauses that we would have otherwise chosen
	skipped++;
	continue;
      }
      first = cl;
      break;
    }
  }
  
  if (first) tStats[SEMANTIC_SELECTIONS]++;
  givens_avoided_by[set_to_skip] += (skipped/(num_maxisets+1));
  givens_avoided += (skipped/(num_maxisets+1));  
  return first;
  
}/*--find_first_sem_cl()--*/


struct clause *find_best_sem_cl(struct list *lptr, int set_to_skip)
  /*----------------------------------------------------------------------.
  | Called by semantic_selection().                                       |
  |                                                                       |
  | Returns the "best" semantic clause on the list.  That is, the         |
  | lightest clause in the given list, belonging to the complement of the |
  | current set_to_skip, and such that its selection will help to         |
  | minimise the variance in the distribution of complements on the       |
  | usable list.                                                          |
  `----------------------------------------------------------------------*/
{
  struct clause *cl, *best = NULL;
  int skipped = 0;

  message(3,"");
  message(3,"skip maxiset number %d",set_to_skip);

  if (set_to_skip == 0) {
    // get lightest clause belonging to the most maxisets
    best = lptr->first_cl;
    for (cl = best->next_cl; cl; cl = cl->next_cl){
      if (cl->num_sets > best->num_sets ||
	  cl->pick_weight < best->pick_weight) {
	best = cl;
      }
      else if (cl->pick_weight < best->pick_weight){
	// otherwise we would have picked this guy
	skipped++;
      }
    }
  }
  else if (tFlags[SMOOTH_CMP_DIST].val) {

    float best_variance = 0;
    float next_variance = 0;

    for (cl = lptr->first_cl; cl; cl = cl->next_cl) {
      // skip clauses in the set_to_skip
      if (cl->in_set[set_to_skip]) {
	// clauses that we would have otherwise chosen
	if (best && cl->pick_weight < best->pick_weight) skipped++;
	continue;
      }

      if (!best) {
	best = cl;
	best_variance = variance(best);
	continue;
      }

      // should this include equality?
      // the SG strategy is stronger than SF so no
      if (cl->pick_weight > best->pick_weight) continue;

      if (cl->num_sets < best->num_sets) continue;

      next_variance = variance(cl);
      if (abs(next_variance-best_variance)/best_variance >
	  fParms[SMOOTHING_RATIO].val) {
	skipped++;
	continue;
      }
      best = cl;
      best_variance = next_variance;
    }
  }
  else {
    // get first lightest clause not in the skip_theory
    for (cl = lptr->first_cl; cl; cl = cl->next_cl){
      if (cl->in_set[set_to_skip]) {
	// clauses that we would have otherwise chosen
	if (best && cl->pick_weight < best->pick_weight) skipped++;
	continue;
      }
      if (!best) { // ensure best is not in skip_set
	best = cl;
	continue;
      }
      if (cl->pick_weight < best->pick_weight ) best = cl;
    }
  }

  if (best) tStats[SEMANTIC_SELECTIONS]++;

  //  givens_avoided_by[set_to_skip] += (skipped/(num_maxisets+1));
  //  givens_avoided += (skipped/(num_maxisets+1)); 

  givens_avoided_by[set_to_skip] += skipped;
  givens_avoided += skipped;

  return best;

}/*--find_best_sem_cl()--*/


int execute_strategy(int_function strategy)
  /*----------------------------------------------------------------------.
  | Called by semantic_selection().                                       |
  |                                                                       |
  | Attempt to form a new maxisets under the given strategy.  Persists so |
  | long as the expected improvement in guidance is greater than the      |
  | expected cost of an attempt.  Discards resultant maxisets that fail   |
  | the retention test.  Give up if the number of consecutive failures    |
  | exceeds the parameter MAX_FAILS.                                      |
  |                                                                       |
  | Returns prematurely if semantics are disabled at any point.           |
  `----------------------------------------------------------------------*/
{
  int new_maxiset, fails;
  int  last_time;
  int last_time_limit;
  struct clause_dptr *cp;

  fails = 0;
  last_time = run_time();
  last_time_limit = get_time_limit();

  (*strategy)(RECORD_EVENT);

  /* find the next available maxiset (there must be at least one --
     we have a temporary position set aside)*/
  for (new_maxiset = 1;
       new_maxiset < last_set_index && active[new_maxiset];
       new_maxiset++);
  if (active[new_maxiset]) abort_scott("all indices are active");

  num_maxisets++;
  active[new_maxiset] = true;

  // set probability of success for before this attempt
  if (tParms[MIN_MAXISET_ATTEMPTS].val < tStats[ATTEMPTS])
    set_prob_of_success(tStats[KEPT]/(float)(tStats[ATTEMPTS]));

  tStats[ATTEMPTS]++;

  start_scan(new_maxiset);  if (!tFlags[SEMANTIC_GUIDANCE].val) return 0;
  scan_sos(new_maxiset);    if (!tFlags[SEMANTIC_GUIDANCE].val) return 0;
  scan_first(new_maxiset);  if (!tFlags[SEMANTIC_GUIDANCE].val) return 0;
  scan_usable(new_maxiset); if (!tFlags[SEMANTIC_GUIDANCE].val) return 0;

  update_partition(new_maxiset); // this must go before next function

//   set_cost_per_failed_clause((run_time() - last_time)/
// 			     (float)(get_comp_size(new_maxiset)));
  
  set_prop_failed_clauses(get_comp_size(new_maxiset)/(float)
			  (Stats[SOS_SIZE]+Stats[USABLE_SIZE]))

  SCOTT_DEBUG("after update_partition()");

  // allow reasons for failure to accumulate?
  //  fails = (*strategy)(FAILS);
  
  if (keep(new_maxiset,"")) {
    if (old_conjecture.first_cp) {

      message(2, "");
      message(2,"Removing the following clause from the conjecture list:");
      
      // discard the existing old conjecture pointers
      for (cp = old_conjecture.first_cp; cp;
	   cp = old_conjecture.first_cp) {
	delete_cl_ext(cp->c,&old_conjecture);
	cp->c->stype = Old_Base; // degrade its status
	if (tFlags[VERBOSITY].val >= 2) print_clause(stdout, cp->c);
      }

      message(2, "");
      message(2,"Adding the following clauses to the conjecture list:");
      
      // move new base clauses to the conjecture list
      for (cp = new_base_clauses.first_cp; cp; cp = cp->next) {
	if (cp->c->in_set[new_maxiset]) {
	  cp->c->stype = Conj;
	  prepend_cl_ext(cp->c,&conjecture);
	  tStats[CONJECTURE_SIZE]++;
	  if (tFlags[VERBOSITY].val >= 2) print_clause(stdout, cp->c);
	  if (cp->c->container == Sos &&
	      tFlags[CONJ_IN_INPUT_FIRST].val){
	    rem_from_list(cp->c);
	    append_cl(First,cp->c);
	    SCOTT_DEBUG("after moving SOS to INPUT_FIRST");
	  }
	}
	else cp->c->stype = Norm;
      }
    }
    else {      
      // move new base clauses to the semantic base
      for (cp = new_base_clauses.first_cp; cp; cp = cp->next) {
	if (cp->c->in_set[new_maxiset]) {
	  cp->c->stype = Base;
	  prepend_cl_ext(cp->c,&semantic_base);
	  tStats[SEM_BASE_SIZE]++;
	  tStats[BASE_CL_ADDED]++;
	}
	else cp->c->stype = Norm;
      }
    }
    
    // discard the existing new_base_clause pointers
    for (cp = new_base_clauses.first_cp;cp;
	 cp = new_base_clauses.first_cp)
      delete_cl_ext(cp->c,&new_base_clauses);

    tStats[KEPT]++;
    (*strategy)(RECORD_KEPT);
    SCOTT_DEBUG("after new maxiset kept");

    discard_subsets(new_maxiset);
    print_partition("after a new maxiset added");

    if (get_time_limit() > last_time_limit){
      // reset_prob_of_success();
      // reset_optimal();
      // hhmmm
    }

    SCOTT_DEBUG("after subsets discarded");
  }
  else {
    if (old_conjecture.first_cp) {
      // reinstate pointers from the old conjecture list
      for (cp = old_conjecture.first_cp; cp;
	   cp = old_conjecture.first_cp) {
	delete_cl_ext(cp->c,&old_conjecture);
	append_cl_ext(cp->c,&conjecture);
      }
    }
      
    check_for_model(new_maxiset); // may not return
    if (!tFlags[SEMANTIC_GUIDANCE].val) return false;
    reset_partition(new_maxiset);    // must go BEFORE discard()
    message(2,"Discarding new maxiset %d",new_maxiset);
    message(2,"");
    discard(new_maxiset);
    shuffle_ext_list(&sos_scan[0]);
    shuffle_ext_list(&sos_scan[1]);
    shuffle_ext_list(&sos_scan[2]);
    SCOTT_DEBUG("after new maxiset discarded");
    fails = 1;
  }

  if ((*strategy)(FAILS)) fails = 1;
  message(3,"%-30s %4.0f secs","Actual cost of the attempt",
	  (run_time() - last_time)/(float)(1000));
  message(3,"");
  
  return fails;
  
}/*--execute_strategy()--*/

void update_partition(int new_maxiset)
  /*---------------------------------------------------------------------.
  | Called by execute_strategy() after scan.  Updates the partition data |
  | structures to include the new maxiset.                               |
  `---------------------------------------------------------------------*/
{
  struct clause *cl;

  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    sos_part_insert(cl,new_maxiset);

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    sos_part_insert(cl,new_maxiset);

  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    usable_part_insert(cl,new_maxiset);

}/*--update_partition()--*/


void reset_partition(int bad_maxiset)
  /*---------------------------------------------------------------------.
  | Called by execute_strategy() after scan.  Removes the bad_maxiset    |
  | from the partition data structures.  NB must be called BEFORE clause |
  | membership data is reset.                                            |
  |                                                                      |
  | Moved form discard() at some point for reasons that are no longer    |
  | seem relevant.  May merge them agian some time in the future.        |
  `---------------------------------------------------------------------*/
{
  struct clause *cl;

  // reset the SOS
  for (cl=Sos->first_cl; cl; cl=cl->next_cl)
    sos_part_delete(cl,bad_maxiset);

  for (cl=First->first_cl; cl; cl=cl->next_cl)
    sos_part_delete(cl,bad_maxiset);

  for (cl=Usable->first_cl; cl; cl=cl->next_cl)
    usable_part_delete(cl,bad_maxiset);

}/*--reset_partition()--*/


boolean keep(int new_maxiset, char *arrow)
  /*------------------------------------------------------------------.
  | Called by execute_strategy().                                     |
  |                                                                   |
  | Retension test for newly generated maxiset.                       |
  |                                                                   |
  | Note: discarding this maxiset will reorder the whole scan so that |
  | new_maxiset clauses come last, reducing the likelihood that a     |
  | duplicate maxiset set will be formed next time.                   |
  `------------------------------------------------------------------*/
{
  float rating = 0;
  int superset = 0;
  int size_on_sos, size_on_usable;

  size_on_sos = get_size_on_sos(new_maxiset);
  size_on_usable = get_size_on_usable(new_maxiset);

  // check if we have covered the whole SOS (hence offers no distinction)
  if (Stats[SOS_SIZE] > tParms[SEM_QUEUE_LIMIT].val
      && size_on_sos >= Stats[SOS_SIZE]) {

    if (Stats[SOS_SIZE] > tParms[SEM_QUEUE_LIMIT].val
	&& (size_on_usable + 1 == Stats[USABLE_SIZE])) {

      struct clause *odd_clause;

      if ((odd_clause = search_list(Usable,new_maxiset,All))) {
	message(2,"%sMaxiset %d covers all but clause %d on the USABLE list",
		arrow,
		new_maxiset,
		odd_clause->id);
      }
      else
	abort_scott("Maxiset is a \"near cover\" but could not find odd clause");
    }
    else {
      message(2,"%sMaxiset %d covers whole SOS",arrow,new_maxiset);
    }
    if (develop_semantics) {
	extend_base(new_maxiset);
	//invert_existing_base();
    }

    record_event(COVERS_SOS);
    tStats[SOS_COVERS]++;
    return false;
  }

  // check if a subset of any existing theory (and hence not maximal)
  if ((superset = superset_of(new_maxiset))) {

    message(2,"%sMaxiset %d is a subset of maxiset %d",
	    arrow, new_maxiset, superset);

    if (develop_semantics) extend_base(superset);
    record_event(FOR_SUB_MAXISET);
    tStats[FOR_SUBS]++;
    return false;
  }

  // keep maxisets that give us a queue clauses
  // but don't continue development
  set_sem_queue(Sos);
  if (First->first_cl) return true;

  // Ensure new maxiset decreases the size of the choice space.  This
  // also ensures the maxiset has a non-empty SOS.
  if ((rating = rate_maxiset(new_maxiset)
       + tParms[MIN_REFINEMENT].val) > 0 ) {

    message(2,"%sMaxiset %d does not refine our choice: rating = %0.1f",
	    arrow, new_maxiset,rating);

    inc_time_limit();
    record_event(FAILS_TO_REFINE);
    tStats[REFINE_FAILS]++;
    return false;
  }

  return true;

}/*--keep()--*/

int cover_new_clauses(int key)
  /*-----------------------------------------------------------------------.
  | Called by execute_strategy(): provides a strategy.                     |
  |                                                                        |
  | Provides control to the development of a new maxisets so that they     |
  | will (hopefully) cover all SOS clauses that currently belong to none.  |
  `-----------------------------------------------------------------------*/
{
  static int clauses_unmarked_before = 0;
  static int clauses_unmarked_after = 0;

  switch(key) {
    
  case RECORD_EVENT:
    SCOTT_DEBUG("before cover_new_clauses()");
    record_event(COVER_NEW_CLAUSES);
    tStats[COVER_ATTEMPTS]++;
    message(2,"");
    message(2,"====> Attempting to cover new clauses");

    return 1; // return value not important

  case TEST: // are there any new null clauses?

    clauses_unmarked_before =  get_sos_partition(0) - null_set_size;

    if (clauses_unmarked_before > 0) return 1;
    return 0;

  case FAILS: // have we failed to eliminated any null clauses?
    
    clauses_unmarked_after = get_sos_partition(0) - null_set_size;
    
    if (clauses_unmarked_before <= clauses_unmarked_after) {
      message(2,"<==== Failed to cover any new clauses");
      
      inc_time_limit();
      if (null_set_size > tParms[MAX_NULL_SET_SIZE].val) inc_cardinality();

      return 1;
    }
    SCOTT_DEBUG("after cover_new_clauses()");

    return 0;
    
  case RECORD_KEPT:
    
    tStats[COVER_KEPTS]++;
    return 0; // return value not important
  }
  
  abort_scott("cover_new_clauses() called with bad key %d.",key);
  
  return 0; // to quiet -Wall

}/*--cover_new_clauses()--*/

int refine_choice(int key)
  /*---------------------------------------------------------------------.
  | Called by execute_strategy(): provides a strategy.                   |
  |                                                                      |
  | Provide control to the development of new maxisets so that they will |
  | (hopefully) decrease the size of the choice space.                   |
  `---------------------------------------------------------------------*/
{
  static float last_rating = 0;
  float this_rating;

  switch(key) {

  case RECORD_EVENT:
    SCOTT_DEBUG("before refine_choice()");
    record_event(REFINE_CHOICE);
    tStats[REFINE_ATTEMPTS]++;
    message(2,"");
    message(2,"====> Attempting to refine choice (%.2f)",last_rating);
    return 0; // return value not important

  case TEST:  // compare average choice space size and size of SOS 

    last_rating = rate_partition();
    if (last_rating*fParms[GUIDANCE_FACTOR].val > 1)
      return 1;
    else
      return 0;

  case FAILS:    // previous rating was better?

    this_rating = rate_partition();
    if (last_rating > this_rating) return 0;

    message(2,"<==== Previous rating was better (%.2f <= %.2f)",
	    last_rating, this_rating);
    SCOTT_DEBUG("after refine_choice()");
    return 1;

  case RECORD_KEPT:

    tStats[REFINE_KEPTS]++;
    return 1; // return value not important

  }

  abort_scott("refine_choice() called with bad key %d",key);

  return 1; // to quiet -Wall

}/*--refine_choice()--*/

boolean discard_bad_maxisets()
  /*--------------------------------------------------------------------.
  | Called by semantic_selection().                                     |
  |                                                                     |
  | Discards any existing maxisets that no longer provide adequate      |
  | guidance.  Properties required for a "good" maxiset are the same as |
  | for the function keep().                                            |
  `--------------------------------------------------------------------*/
{
  int maxiset;
  boolean bad_sets = false;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val) return false;

  for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
    if (active[maxiset] && !keep(maxiset,"====> ")) {
      bad_sets = true;
      message(2,"<==== Discarding existing maxiset %d", maxiset);
      reset_partition(maxiset);
      discard(maxiset);
      tStats[CULLED]++;
    }
  }

  return bad_sets;

} /*--discard_bad_maxisets()--*/


void discard_subsets(int new_maxiset)
  /*--------------------------------------------------------------.
  | Called by execute_strategy().  Discards any theories that are |
  | sos_subsets of the new_maxiset.                               |
  `--------------------------------------------------------------*/
{
  int old_maxiset;

  if (Stats[SOS_SIZE] <= tParms[SEM_QUEUE_LIMIT].val){
    return;
  }

  for (old_maxiset = 1; old_maxiset <= last_set_index; old_maxiset++){
    if (active[old_maxiset] &&
	old_maxiset !=  new_maxiset &&
	//subset(old_maxiset,new_maxiset)) { // for real subsets
      sos_subset(old_maxiset,new_maxiset)) {

      message(2,"Maxiset %d is a subset of new maxiset. Discarding maxiset %d",
	      old_maxiset, old_maxiset);

      record_event(BACK_SUB_MAXISET);

      reset_partition(old_maxiset);      // must go BEFORE discard
      discard(old_maxiset);

      tStats[BACK_SUBS]++;
      if (old_maxiset == constraint_set)
	constraint_set = new_maxiset;
    }
  }
}/*--discard_subsets()--*/

void discard(int bad_maxiset)
  /*---------------------------------------------------------------------.
  | Called by discard_bad_maxisets(), discard_subsets(),                 |
  | execute_strategy(), and cull_maxisets().                             |
  |                                                                      |
  | Removes the "bad_maxiset" from the scan, resets the weights of any   |
  | clause on the SOS true in it, and make its index avaliable for a new |
  | maxiset.  Does not change the partition data structures.  These are  |
  | managed by update_partition() and reset_partition().                 |
  `---------------------------------------------------------------------*/
{
  struct clause *cl;
  struct clause_dptr *cp, *next;
  int num_models, scanned;

  fflush(stdout);

  // reset the external clause space lists
  cp = failed_clauses[bad_maxiset].first_cp;
  while (cp) {
    delete_cl_ext(cp->c,&failed_clauses[bad_maxiset]);
    cp = failed_clauses[bad_maxiset].first_cp;
  }

  scanned = 0;
  for (num_models = 1 ; num_models <= num_maxisets; num_models++){
    cp = sos_scan[num_models].first_cp;
    while (cp) {
      next = cp->next;
      scanned++;
      // this deletion peels of the top element
      if (cp->c->in_set[bad_maxiset]){
	delete_cl_ext(cp->c,&sos_scan[num_models]);
	// deprecates similar scans
	if (scanned % 3) prepend_cl_ext(cp->c,&sos_scan[num_models-1]);
	else append_cl_ext(cp->c,&sos_scan[num_models-1]);
      }
      cp = next;
    }
  }

  // make old maxiset position available for next maxiset.
  num_maxisets--;
  active[bad_maxiset] = false;
  if (bad_maxiset == last_queue_theory) last_queue_theory = 0;
  init_sizes(bad_maxiset);
  init_theory(bad_maxiset);

  if (tFlags[DEMOD_IN_SEM_BASE].val ||
      tFlags[DEMOD_IN_CONJECTURE].val ) {
    // reset the DEMODULATORS (all base clauses)
    for (cl=Demodulators->first_cl; cl; cl=cl->next_cl){
      cl->in_set[bad_maxiset] = false;
    }
  }
  
  // reset the SOS
  for (cl=Sos->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }

  // reset the INPUT FIRST list
  for (cl=First->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }

  // reset the USABLE list
  for (cl=Usable->first_cl; cl; cl=cl->next_cl){
    if (cl->in_set[bad_maxiset]){
      cl->num_sets--;
      cl->in_set[bad_maxiset] = false;
    }
    cl->pick_weight = semantic_wt(cl);
  }

}/*--discard()--*/

void disable_semantics(char *event)
  /*-----------------------------------------------------------------------.
  | Called by various functions if a clause is too big to parse or if      |
  | development becomes too expensive while there are less than 2 models.  |
  |                                                                        |
  | Semantic guidance is now meaningless.  Behaviour defaults to standard  |
  | OTTER.                                                                 |
  `-----------------------------------------------------------------------*/
{
  record_event("\ndisable semantics after %d given clauses",
	       Stats[CL_GIVEN]);

  message(1,"");
  message(1,"WARNING: %s.\nDisabling semantic guidance "
	  "after %d given clauses\n", event, Stats[CL_GIVEN]);

  auto_change_scott_flag(stdout,SEMANTIC_GUIDANCE,0);
  //  auto_change_parm(stdout,PICK_GIVEN_RATIO,4);

  if (num_maxisets > 0) {// reset the SOS wts
    struct clause *cl;
    int maxiset;

    for (cl=Sos->first_cl; cl; cl=cl->next_cl)
      cl->pick_weight = cl->length;

    for (maxiset = 0; maxiset <= last_set_index; maxiset++)
      active[maxiset] = false;
    num_maxisets = 0;
  }

  fflush(stdout);

}/*--disable_semantics()--*/


void init_semantics(int argc, char *argv[])
  /*-------------------------------------------------------------------.
  | Called by OTTER fcn main() after input file read.                  |
  |                                                                    |
  | Initialises global variables and partition data structures; Checks |
  | the command line to see if a findefile has been provided.  Calls   |
  | init_finder() to prepare for the first model.  Finally, builds an  |
  | approriate semantic base.                                          |
  `-------------------------------------------------------------------*/
{
  int option, stat, maxiset;
  char filename[FILE_NAME_ARRAY];
  extern char *optarg;

  // check for loops
  if (develop_semantics) return;

  *History = '\0';
  recording = true;

  // initialise principle global variables
  num_maxisets = 0;
  null_set_size = 0;
  develop_semantics = true;
  new_base_clauses.first_cp = NULL;
  new_base_clauses.last_cp = NULL;
  last_queue_theory = 0;

  // localise the following parameters
  last_set_index = tParms[MAX_NUM_MAXISETS].val+1; // one spare for replacement
  constraint_set = last_set_index + 1 ;

  // check to see if a finder file is provided
  strcpy(filename,"");
  while ((option = getopt (argc, argv, "f:")) != -1){
    switch( option ) {
    case 'f':
      if (strlen(optarg) <= MAX_FILE_NAME)
	strcpy(filename,optarg);
      else abort_scott("filename longer than MAX_FILE_NAME characters.");
      break;
    case '?':
      fprintf(stderr,"\nusage: scott [-f finder_file] < stdin\n");
      fprintf(stderr,"%c",Bell);
      exit(1);
    }
  }

  // initialise partition data structures
  for (maxiset = 0; maxiset <= last_set_index; maxiset++){
      givens_avoided_by[maxiset] = 0;
      last_num_updates[maxiset] = 0;
      giv_cl_from_comp[maxiset] = 0;
    }

  active[0] = true;
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)
    active[maxiset] = false;

  for (stat = 0; stat < MAX_TSTATS; stat++) tStats[stat] = 0;
  tStats[MAXISETS_AT_END_DEV] = -1;

  // Generate an initial model and set the base maxiset
  // This also checks the integrity of model constraints

  // First pass these parameters to the FINDER fragment
  set_min_prob_extn((float)fParms[MIN_PROB_EXTN].val);

  set_min_time_limit((int)(tParms[MIN_SEARCH_TIME].val/fParms[TOLERANCE].val));
  set_max_time_limit(tParms[MAX_SEARCH_TIME].val);
  set_tolerance(fParms[TOLERANCE].val);

  // maximum time limit for modelling finder file;
  set_time_limit(tParms[MAX_SEARCH_TIME].val);

  SCOTT_CLOCK_START(SEM_DEV_TIME);
  if (!init_finder(filename)){

    SCOTT_CLOCK_STOP(SEM_DEV_TIME);

    message(1,"WARNING: can't model the given semantics.\n"
	      "SCOTT is now running without semantic guidance");

    tFlags[SEMANTIC_GUIDANCE].val = false;
    return;
  }

  set_time_limit(tParms[MIN_SEARCH_TIME].val);

  init_costs();

  /*-----------------------------------------------------------------.
  | put these here since init_clause_space() etc may be called twice |
  `-----------------------------------------------------------------*/
  semantic_base.first_cp =
    semantic_base.last_cp =
    new_base_clauses.first_cp =
    new_base_clauses.last_cp =
    conjecture.first_cp =
    conjecture.last_cp = 
    old_conjecture.first_cp = 
    old_conjecture.last_cp = NULL;

  /*------------------------------------------------------------------.
  | these must be init'ed before we move clauses around despite being |
  | reinit'ed afterwards                                              |
  `------------------------------------------------------------------*/

  init_clause_space();
  init_partitions();
  init_conjecture_list();
  
  if (tFlags[SEMANTIC_GUIDANCE].val && tFlags[AUTO_SET_SOS].val) auto_set_sos();
  if (tFlags[SEM_RES].val) set_up_model();
    
  init_semantic_base(); // must go after auto_set_sos()

  SCOTT_CLOCK_STOP(SEM_DEV_TIME);

}/*--init_semantics()--*/


/*-------------------------------------------.
| Special macros for output_finder_stats().  |
`-------------------------------------------*/
   
#define p_1d_stat(name, value)  fprintf(fp,"\n%-36s%5d",name,value)
#define p_2d_stat(name, value)\
{\
  fprintf(fp,"\n%-21s",name);\
  for (maxiset = 1; maxiset <= last_set_index; maxiset++)\
    if (active[maxiset]) fprintf(fp,"%8d",value);\
}\


void output_scott_stats(FILE *fp, int stats_level)
  /*----------------------------------------------------------------------.
  | Called by OTTER fcn output_stats().                                   |
  |                                                                       |
  | Whenever OTTER outputs it's stats, appends appropriate FINDER stats.  |
  `----------------------------------------------------------------------*/
{
  int maxiset;

  // use negative settings for some weird alternatives
  if (stats_level == -1) {
    print_stats(fp);
    print_times_brief(fp);
  }
    
  if (stats_level == -2) { // special hack for demo
    // print out a list of integers for ANIM
    print_special_stats(fp);
  }    

  if (stats_level < 1 ) return;

  fprintf(fp, "\n------------ semantic stats ------------\n");

  p_1d_stat("Semantic selections",                 tStats[SEMANTIC_SELECTIONS]);
  p_1d_stat("Clauses given from input_first list", tStats[GIVENS_FROM_FIRST]);
  p_1d_stat("Clause placed in semantic queue",     tStats[CL_SEM_QUEUED]);
  p_1d_stat("Given clauses avoided",               givens_avoided);
  p_1d_stat("Number of pseudo null clauses",       null_set_size);
  fprintf(fp,"\n");

  p_1d_stat("Final number of maxisets",            num_maxisets);

  if (tStats[MAXISETS_AT_END_DEV] > -1)
    p_1d_stat("Maxisets at end of development",      tStats[MAXISETS_AT_END_DEV]);

  p_1d_stat("Attempts at a new maxiset",           tStats[ATTEMPTS]);
  p_1d_stat("  to cover new null clauses",           tStats[COVER_ATTEMPTS]);
  p_1d_stat("  to refine choice",                    tStats[REFINE_ATTEMPTS]);

  p_1d_stat("New maxisets kept",                   tStats[KEPT]);
  p_1d_stat("  that covered a new null clause",      tStats[COVER_KEPTS]);
  p_1d_stat("  that refined choice",                 tStats[REFINE_KEPTS]);

  p_1d_stat("Maxisets forward subsumed",           tStats[FOR_SUBS]);
  p_1d_stat("Maxisets covering whole SOS",         tStats[SOS_COVERS]);
  p_1d_stat("Maxisets failing refinement",         tStats[REFINE_FAILS]);
  p_1d_stat("Existing maxisets back subsumed",     tStats[BACK_SUBS]);
  p_1d_stat("Existing maxisets culled",            tStats[CULLED]);
  p_1d_stat("Maxisets clashing with new base",     tStats[CLASHES]);

  fprintf(fp,"\n");
  p_1d_stat("Size of final conjecture list",       tStats[CONJECTURE_SIZE]);
  p_1d_stat("Size of final semantic base",         tStats[SEM_BASE_SIZE]);
  p_1d_stat("Base clauses skipped",                tStats[BASE_CL_SKIPPED]);
  p_1d_stat("Base clauses added",                  tStats[BASE_CL_ADDED]);
  p_1d_stat("Base extension failures",             tStats[BASE_EXTENSION_FAILS]);

  fprintf(fp,"\n");
  p_1d_stat("Clauses marked false by time-limit",  get_time_outs());
  p_1d_stat("  while still in test",               get_time_outs_in_test());
  p_1d_stat("Time limit at end of search",         get_time_limit());
  p_1d_stat("Clauses modelled by retesting",       tStats[SUCCESSFUL_RETESTS]);
  p_1d_stat("Clauses moved to refine SOS strategy",tStats[SOS_STRATEGY_REFINEMENTS]);

  fprintf(fp,"\n");
  p_1d_stat("Fails in persist()",                  tStats[FAILS_IN_PERSIST]);
  p_1d_stat("   time saved strictly increasing",   tStats[PFAIL_TIME_INC]);
  p_1d_stat("   time saved strictly decreasing",   tStats[PFAIL_TIME_DEC]);
  p_1d_stat("   cost vs time left",                tStats[PFAIL_COST_VS_TIME]);
  p_1d_stat("   cost vs time accrued",             tStats[PFAIL_TIME_ACCRUED]);
  p_1d_stat("   probability strictly decreasing",  tStats[PFAIL_PROB_DEC]);
  p_1d_stat("   cost vs probability",              tStats[PFAIL_COST_VS_PROB]);
  p_1d_stat("   cost vs expected return",          tStats[PFAIL_COST_VS_RETURN]);
  fprintf(fp,"\n");

  if (tFlags[SEMANTIC_GUIDANCE].val && num_maxisets > 0) {
    p_2d_stat("Maxiset index",     maxiset);
    fprintf(fp,"\n---------------------");
    for (maxiset = 1; maxiset <= last_set_index; maxiset++)
      if (active[maxiset]) fprintf(fp,"--------");
    p_2d_stat("Last maxiset size",    get_clauses_modelled(maxiset));
    p_2d_stat("Givens avoided by",    givens_avoided_by[maxiset]);
    p_2d_stat("Models searched for",  get_searches(maxiset));
    p_2d_stat("Models found",         get_successful_searches(maxiset));
    p_2d_stat("Max backtracks needed",get_max_backtracks_in(maxiset));
    p_2d_stat("Max time needed",      get_max_time_in(maxiset));
    p_2d_stat("Final domain size",    get_last_default_size(maxiset));
    fprintf(fp,"\n");

    print_partition("at end of search");

    debug_partition("at end of search");
    debug_scan("at end of search");

    message(2,"");
  }
  message(2,"");
  output_weight_distn();
  output_variable_distn();
  message(2,""); 
  output_history();

}/*--output_finder_stats()--*/

void print_special_stats(FILE *fp)
  /*-------------------------------------------------------------.
  | Print Stats for ANIM program. Used only if stats_level == -2 |
  `-------------------------------------------------------------*/
{
  int theory;
  int total_searches = 0; 
  int total_sucessful_searches = 0;
  
  for (theory = 1; theory < MAXISET_ARRAY; theory++){
    //    if (active[theory]) {
    total_searches += get_searches(theory);
    total_sucessful_searches += get_successful_searches(theory);
    //    }
  }
  
  // $1 size of SOS
  // $2 size of virtual SOS
  // $3 clause genertated
  // $4 semantic time
  // $5 total time
  // $6 number of searches for a model
  // $7 number of searches that where successful

  fprintf(fp,"\nANIM DATA %d %d %d %d %d %d %d\n",
	  (int)Stats[SOS_SIZE],
	  (tFlags[SEMANTIC_GUIDANCE].val 
	   ? (int)size_choice_space(): (int)Stats[SOS_SIZE]),

	  (int)Stats[CL_GENERATED],
	  
	  // proportion of total time spent in semantics
	  (int)(clock_val(SEM_GUIDE_TIME) + clock_val(SEM_DEV_TIME)),
	  (int)run_time(),
	  
	  // prob of new model
	  total_searches, 
	  total_sucessful_searches
	  );
  fflush(fp);
}

void scott_clock_init()
  /*-------------------------------.
  | Called by scott_clock_init().  |
  |                                |
  | Initialise all scott clocks.   |
  `-------------------------------*/
{
    int i;

#ifdef THINK_C  /* kludge for mac: see run_time */
    long l;
    l = run_time();
#endif

    for (i=0; i<MAX_TCLOCKS; i++)
	scott_clock_reset(i);

}  /*--clock_init()--*/

void scott_clock_reset(int c)
  /*-----------------------------------------.
  |      Called by scott_clock_init().       |
  |                                          |
  | Clocks must be reset before being used.  |
  `-----------------------------------------*/
{
    tClocks[c].accum_sec = tClocks[c].accum_usec = 0;
    tClocks[c].curr_sec = tClocks[c].curr_usec = -1;

}  /*--clock_reset()--*/


long scott_clock_val(int c)
  /*----------------------------------------------------------------------.
  | Called by cost_per_attempt() and cost_per_given().                    |
  |                                                                       |
  | Returns accumulated time (in milliseconds) of the given clock.  Clock |
  | need not be stopped.                                                  |
  `----------------------------------------------------------------------*/
{
    long sec, usec, i, j;

    i = (tClocks[c].accum_sec * 1000) + (tClocks[c].accum_usec / 1000);
    if (tClocks[c].curr_sec == -1)
	return(i);
    else {
	SCOTT_CPU_TIME(sec, usec)
	j = ((sec - tClocks[c].curr_sec) * 1000) +
	    ((usec - tClocks[c].curr_usec) / 1000);
	return(i+j);
	}
}/*--scott_clock_val()--*/


void message(int priority, char *msg, ...)
  /*--------------------------------------------------------------------.
  | Called by various functions...                                      |
  |                                                                     |
  | Prints the given message to the screen iff priority is greater than |
  | or equal to the VERBOSITY parameter.  -- msg may be a std format    |
  | string, and in this case may be followed by arbitrary arguments.    |
  |                                                                     |
  | If the priority is less than 2 then the terminal bell is rung.      |
  `--------------------------------------------------------------------*/
{
  static boolean newline = false;

  if (priority <= tParms[VERBOSITY].val) {
    va_list args;

    if (strcmp(msg, "") == 0) {
      newline = true;
      return;
    }

    if (newline) printf("\n");

    va_start(args, msg);
    vprintf(msg, args);
    printf("\n");
    fflush(stdout);

#ifndef SCOTT_DEBUG
    if (priority < 2) {

      if (newline) fprintf(stderr,"\n");

      fprintf(stderr,"%c",Bell);
      vfprintf(stderr, msg, args);
      fprintf(stderr,"\n");
    }
#endif
    newline = false;

    fflush(stdout);
    va_end(args);
  }
} /*--message()--*/

void abort_scott(char *msg, ...)
  /*----------------------------------------------------------------------.
  | Called by various functions...                                        |
  |                                                                       |
  | An error has been detected.  the terminal bell is rung, the given     |
  | message is printed to the screen, and the program exits -- msg may be |
  | a std format string, and in this case may be followed by arbitrary    |
  | arguments.                                                            |
  `----------------------------------------------------------------------*/
{
  va_list args;

  va_start(args, msg);

  fprintf(stderr,"\nERROR: ");
  fprintf(stderr,"%c", Bell);
  vfprintf(stderr, msg, args);
  fprintf(stderr, "\n");
#ifndef SCOTT_DEBUG
  fprintf(stdout,"\nERROR: ");
  vfprintf(stdout, msg, args);
  fprintf(stdout, "\n");
#endif
  va_end(args);

  fprintf(stdout,"\nEvents preceding the error:\n\n");
  output_history();
  exit(1);

}/*--abort_scott()--*/

void record_event(char *event, ...)
{

  if (recording) {
    va_list args;
    char tmp[20];

    *tmp = '\0';

    va_start(args, event);
    vsprintf(tmp,event,args);
    strncat(History,tmp,MAX_HISTORY);
    va_end(args);
  }
}/*--record_event()--*/

void output_history()
{
  message(2,"----> Start Development history:\n%s\n\n"
	 "<---- End development history\n",History);

}/*--output_history()--*/

void print_scott_banner(int argc, char **argv)
{
    int i;
    int pid = my_process_id();
    printf("\n---------------------------------------------------------------");
    printf("\n                 SCOTT-v%-5s (%s) \n", SCOTT_VERSION, SCOTT_DATE);
    printf("\n                         based on \n\n");
    printf(  "    OTTER-v%-5s (%s) and FINDER-v%-5s (%s)", 
	     OTTER_VERSION, VERSION_DATE, FINDER_VERSION, FINDER_DATE);
    printf("\n---------------------------------------------------------------\n\n");
    printf("The process was started by %s on %s, %s",
	   username(), hostname(), get_time());

    printf("The command was \"");
    for(i = 0; i < argc; i++)
        printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
    printf("\".");
    if (pid != 0)
      printf("  The process ID is %d.\n\n", pid);
    else
      printf("\n\n");
    
}/*--print_scott_banner()--*/


/*-----------------------------------------------------------------------.
| The following group of functions simply return the value of a variable |
| that would otherwise be out of scope.                                  |
`-----------------------------------------------------------------------*/

// Called Foutput.c:fprint_model()
int get_num_maxisets() { return num_maxisets; }

// Called by clause.c:print_clause() and process.c:pre_process()
boolean in_constraint_set(struct clause *cl) {  return cl->in_set[constraint_set];}

/*--------------------------------------------------------------------*/
